/* $Id: erash.c,v 1.20 2004/03/16 20:58:42 jhayes Exp $ */
/*
 * Copyright  2004 The Regents of the University of California. 
 * All Rights Reserved. 
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes, without
 * fee, and without a written agreement is hereby granted, provided that the
 * above copyright notice, this paragraph and the following three paragraphs
 * appear in all copies. 
 *
 * Permission to incorporate this software into commercial products may be
 * obtained by contacting
 * Technology Transfer Office 
 * 9500 Gilman Drive 
 * 411 University Center 
 * University of California 
 * La Jolla, CA 92093-0093
 * (858) 534-5815
 * invent@ucsd.edu
 *
 * This software program and documentation are copyrighted by The Regents of
 * the University of California. The software program and documentation are
 * supplied "as is", without any accompanying services from The Regents. The
 * Regents does not warrant that the operation of the program will be
 * uninterrupted or error-free. The end-user understands that the program was
 * developed for research purposes and is advised not to rely exclusively on
 * the program for any reason. 
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
 * EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED
 * HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO
 * OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS. 
 */

 /*
  * This program sets up a rash proxy and associated agents.  Usage:
  *   erash [--cpus n] [--nodes n] [--port n] [--queue name] [--scheduler type]
  *         [--wd path] [type#]server
  */

#include <signal.h> /* signal */
#include <stdio.h>  /* printf fprintf */
#include <stdlib.h> /* atoi */
#include <string.h> /* strchr strncmp */
#define ASENV_SHORT_NAMES
#include <envseed.h>
#include <oshseed.h>
#define ASSTR_SHORT_NAMES
#include <strseed.h>
#include "elbase.h"
#include "rashgrp.h"


static RASHGRP_GroupInfo group = NULL;


static void
SignalHandler(int signal) {
  if(group != NULL)
    RASHGRP_GroupFree(group);
  RASHGRP_Shutdown();
  exit(0);
}


int
main(int argc,
     char **argv) {

  const char *VALID =
    "--param     name=value batch submission parameter\n"
    "--port      int rash proxy port\n"
    "--scheduler shell|condor|dqs|loadleveler|lsf|pbs|sge type of batch scheduler\n"
    "--storage   string [type#]server\n"
    "--wd        string proxy/agent working directory\n"
    "            string [type#]server";
  const char *COMPUTE_TYPE_NAMES[ELBASE_COMPUTE_SERVICE_COUNT + 1] = {
    "fork", "gram", "netsolve", "rash", "ssh", NULL
  };
  const ELBASE_StorageServiceTypes
    COMPUTE_STORAGE_TYPES[ELBASE_COMPUTE_SERVICE_COUNT] = {
    ELBASE_CP, ELBASE_GASS, ELBASE_SCP, ELBASE_SCP, ELBASE_SCP
  };
  const char *SCHEDULER_TYPE_NAMES[ELBASE_SCHEDULER_SERVICE_COUNT + 1] = {
    "shell", "condor", "dqs", "loadleveler", "lsf", "pbs", "sge", NULL
  };
  const char *STORAGE_TYPE_NAMES[ELBASE_STORAGE_SERVICE_COUNT + 1] = {
    "cp", "ftp", "gass", "scp", "sftp", "srb", NULL
  };

  const char **attributes = NULL;
  unsigned attributeCount = 0;
  const char *arg;
  const char *computeServer;
  ELBASE_ComputeServiceTypes computeServerType;
  unsigned long i;
  ParsedArgv pa;
  unsigned port;
  int status;
  ELBASE_SchedulerServiceTypes schedulerType;
  const char *storageServer;
  ELBASE_StorageServiceTypes storageServerType;
  const char *wd;

  if((pa = ParseArgv((const char **)argv, VALID, NULL)) == NULL ||
     !SwitchPresent(pa, "")) {
    fprintf(stderr, "usage: erash\n%s\n", VALID);
    return 1;
  }

  arg = SwitchValue(pa, "", NULL);
  if((computeServer = strchr(arg, '#')) == NULL) {
    computeServerType = ELBASE_SSH;
    computeServer = arg;
  }
  else if(StrArrayNFind(COMPUTE_TYPE_NAMES, StrArrayLen(COMPUTE_TYPE_NAMES),
                        arg, computeServer - arg, &i)) {
    computeServerType = (ELBASE_ComputeServiceTypes)i;
    computeServer++;
  }
  else {
    fprintf(stderr, "Unknown server type %s\n", arg);
    return 1;
  }

  arg = SwitchValue(pa, "--storage", NULL);
  if(arg == NULL || (storageServer = strchr(arg, '#')) == NULL) {
    storageServerType = COMPUTE_STORAGE_TYPES[computeServerType];
    storageServer = arg == NULL ? computeServer : arg;
  }
  else if(StrArrayNFind(STORAGE_TYPE_NAMES, StrArrayLen(STORAGE_TYPE_NAMES),
                        arg, storageServer - arg, &i)) {
    storageServerType = (ELBASE_StorageServiceTypes)i;
    storageServer++;
  }
  else {
    fprintf(stderr, "Unknown server type %s\n", arg);
    return 1;
  }

  arg = SwitchValue(pa, "--scheduler", NULL);
  if(arg == NULL)
    schedulerType = ELBASE_PBS;
  else if(StrArrayFind(SCHEDULER_TYPE_NAMES, arg, &i))
    schedulerType = (ELBASE_SchedulerServiceTypes)i;
  else {
    fprintf(stderr, "Unknown scheduler type %s\n", arg);
    return 1;
  }

  wd = SwitchValue(pa, "--wd", ".");
  arg = SwitchValue(pa, "--port", NULL);
  port = arg == NULL ? (unsigned)getpid() % 4000 + 4000 : atoi(arg);
  for(i = 1; (arg = SwitchValueN(pa, "--param", i, NULL)) != NULL; i++) {
    attributes = (const char **)realloc
      ((void *)attributes, (attributeCount + 2) * sizeof(char *));
    attributes[attributeCount++] = arg;
  }
  if(attributes != NULL)
    attributes[attributeCount] = NULL;

  RASHGRP_Init();
  signal(SIGINT, &SignalHandler);
  signal(SIGTERM, &SignalHandler);
  group = RASHGRP_GroupNew(storageServerType, storageServer, computeServerType,
                           computeServer, port, wd);
  if(group == NULL) {
    fprintf(stderr, "Script copying failed\n");
    return 1;
  }
  if(!RASHGRP_StartProxy(group)) {
    RASHGRP_GroupFree(group);
    fprintf(stderr, "Proxy start failed\n");
    return 1;
  }
  if(!RASHGRP_SubmitAgents(group, schedulerType, attributes)) {
    RASHGRP_GroupFree(group);
    fprintf(stderr, "Agent submission failed\n");
    return 1;
  }
  if(attributes != NULL)
    free((void *)attributes);
  printf("Waiting for agents to start...\n");
  while(RASHGRP_AgentsPending(group, &status))
    sleep(15);
  if(status != 0) {
    RASHGRP_GroupFree(group);
    RASHGRP_Shutdown();
    fprintf(stderr, "Agents failed to start; exit code %d\n", status);
    return 1;
  }
  printf("Group running; proxy is listening on %s:%u\n", computeServer, port);
  while(RASHGRP_AgentsRunning(group))
    sleep(15);
  RASHGRP_GroupFree(group);
  RASHGRP_Shutdown();
  ParsedArgvFree(pa);
  return 0;

}
