/* $Id: esh.c,v 1.14 2004/03/15 16:55:10 jhayes Exp $ */
/*
 * Copyright  2004 The Regents of the University of California. 
 * All Rights Reserved. 
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes, without
 * fee, and without a written agreement is hereby granted, provided that the
 * above copyright notice, this paragraph and the following three paragraphs
 * appear in all copies. 
 *
 * Permission to incorporate this software into commercial products may be
 * obtained by contacting
 * Technology Transfer Office 
 * 9500 Gilman Drive 
 * 411 University Center 
 * University of California 
 * La Jolla, CA 92093-0093
 * (858) 534-5815
 * invent@ucsd.edu
 *
 * This software program and documentation are copyrighted by The Regents of
 * the University of California. The software program and documentation are
 * supplied "as is", without any accompanying services from The Regents. The
 * Regents does not warrant that the operation of the program will be
 * uninterrupted or error-free. The end-user understands that the program was
 * developed for research purposes and is advised not to rely exclusively on
 * the program for any reason. 
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
 * EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED
 * HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO
 * OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS. 
 */

 /*
  * This program provides a command-line interface to Elagi's process spawning
  * services.  Usage:
  *
  *   esh [--env string][--scheduler type][--wd path]
  *       [type#]server command [arg ...]
  *
  * The type defaults to ssh.
  */

#include <signal.h> /* signal */
#include <stdio.h>  /* fprintf */
#include <stdlib.h> /* free */
#include <string.h> /* strchr strdup */
#define ASENV_SHORT_NAMES
#include <envseed.h>
#include <fileseed.h>
#include <oshseed.h>
#define ASSTR_SHORT_NAMES
#include <strseed.h>
#include "elbase.h"


static ELBASE_Process pid = NULL;


static void
SignalHandler(int sig) {
  if(pid != NULL)
    ELBASE_Kill(pid);
  exit(1);
}


int
main(int argc,
     char **argv) {

  const char *VALID =
    "--env       string name=value[;...]\n"
    "--scheduler shell|condor|dqs|loadleveler|lsf|pbs|sge\n"
    "--scriptgen yes|no\n"
    "--stderr    path\n"
    "--stdin     path\n"
    "--stdout    path\n"
    "--wd        path\n"
    "            string  [type#]server command [arg ...]";
  const char *SCHEDULER_TYPE_NAMES[ELBASE_SCHEDULER_SERVICE_COUNT + 1] = {
    "shell", "condor", "dqs", "loadleveler", "lsf", "pbs", "sge", NULL
  };
  const char *SERVER_TYPE_NAMES[ELBASE_COMPUTE_SERVICE_COUNT + 1] = {
    "fork", "gram", "netsolve", "rash", "ssh", NULL
  };

  char *arg;
  char **args;
  const char **attributes = NULL; /* TODO */
  char **env = NULL;
  unsigned long i;
  ParsedArgv pa;
  ELBASE_SchedulerServiceTypes scheduler;
  char *server;
  const char *stderrPath;
  const char *stdinPath;
  const char *stdoutPath;
  int status;
  ELBASE_ComputeServiceTypes type;
  const char *wd;

  if((pa = ParseArgv((const char **)argv, VALID, NULL)) == NULL ||
     SwitchValueN(pa, "", 2, NULL) == NULL) {
    fprintf(stderr, "usage: esh\n%s\n", VALID);
    return 1;
  }

  arg = strdup(SwitchValueN(pa, "", 2, NULL));
  for(i = 3; SwitchValueN(pa, "", i, NULL) != NULL; i++)
    arg =
      StrAppend(arg, " ", SwitchValueN(pa, "", i, NULL), NULL);
  args = StrSplit(arg, NULL);
  free(arg);

  arg = SwitchValueN(pa, "", 1, NULL);
  if((server = strchr(arg, '#')) == NULL) {
    type = ELBASE_SSH;
    server = arg;
  }
  else if(StrArrayNFind(SERVER_TYPE_NAMES, StrArrayLen(SERVER_TYPE_NAMES),
                        arg, server - arg, &i)) {
    type = (ELBASE_ComputeServiceTypes)i;
    server++;
  }
  else {
    fprintf(stderr, "Unknown server type %s\n", arg);
    return 1;
  }
  arg = SwitchValue(pa, "--scheduler", "shell");
  if(StrArrayFind(SCHEDULER_TYPE_NAMES, arg, &i))
    scheduler = (ELBASE_SchedulerServiceTypes)i;
  else {
    fprintf(stderr, "Unknown scheduler type %s\n", arg);
    return 1;
  }
  if(SwitchPresent(pa, "--env"))
    env = StrSplit(SwitchValue(pa, "--env", NULL), ";");
  stdinPath = SwitchValue(pa, "--stdin", NULL);
  stderrPath = SwitchValue(pa, "--stderr", NULL);
  stdoutPath = SwitchValue(pa, "--stdout", NULL);
  wd = SwitchValue(pa, "--wd", NULL);

  signal(SIGINT, &SignalHandler);
  signal(SIGTERM, &SignalHandler);
  if(strcmp(SwitchValue(pa, "--scriptgen", "no"), "yes") == 0) {
    char *script = ELBASE_ScriptForSubmit
      (type, server, scheduler, attributes, args[0], wd,
       (const char **)&args[1], stdinPath, stdoutPath, stderrPath,
       (const char **)env);
    if(script == NULL) {
      fprintf(stderr, "Script generation failed\n");
      return 1;
    }
    fprintf(stdout, "%s\n", script);
    status = 0;
  }
  else if(!ELBASE_Submit(type, server, scheduler, attributes, args[0], wd,
                         (const char **)&args[1], stdinPath, stdoutPath,
                         stderrPath, (const char **)env, &pid))  {
    fprintf(stderr, "Spawn failed\n");
    return 1;
  }
  else if(!ELBASE_Poll(pid, 1, &status)) {
    fprintf(stderr, "Poll failed\n");
    return 1;
  }

  StrArrayFree(args);
  if(env != NULL)
    StrArrayFree(env);
  ParsedArgvFree(pa);
  return status;

}
