/* $Id: elmetric.c,v 1.14 2004/03/05 19:39:44 jhayes Exp $ */
/*
 * Copyright  2004 The Regents of the University of California. 
 * All Rights Reserved. 
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes, without
 * fee, and without a written agreement is hereby granted, provided that the
 * above copyright notice, this paragraph and the following three paragraphs
 * appear in all copies. 
 *
 * Permission to incorporate this software into commercial products may be
 * obtained by contacting
 * Technology Transfer Office 
 * 9500 Gilman Drive 
 * 411 University Center 
 * University of California 
 * La Jolla, CA 92093-0093
 * (858) 534-5815
 * invent@ucsd.edu
 *
 * This software program and documentation are copyrighted by The Regents of
 * the University of California. The software program and documentation are
 * supplied "as is", without any accompanying services from The Regents. The
 * Regents does not warrant that the operation of the program will be
 * uninterrupted or error-free. The end-user understands that the program was
 * developed for research purposes and is advised not to rely exclusively on
 * the program for any reason. 
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
 * EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED
 * HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO
 * OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS. 
 */


#include "config.h"
#include <oshseed.h>
#include <stdio.h>  /* file functions */
#include <stdlib.h> /* free malloc */
#include <string.h> /* strdup */
#include <time.h>   /* time */
#include <fileseed.h>
#define ASIP_SHORT_NAMES
#include <ipseed.h>
#define ASSTR_SHORT_NAMES
#include <strseed.h>
#define ASXML_SHORT_NAMES
#include <xmlseed.h>
#include "nwsclient.h"
#include "ghsclient.h"
#include "elmetric.h"

#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>

#ifndef NULL
#  define NULL 0
#endif


struct _ELMETRIC_KeyStruct {
  ELMETRIC_MetricServiceTypes service;
  char *server;
  char *entity;
  ELMETRIC_MetricTypes metric;
  void *serviceKeyInfo;
};
struct _ELMETRIC_MetricStruct {
  unsigned long expiration;
  ELMETRIC_Key key;
  char *value;
};

#define SECONDS_IN_A_MINUTE 60
#define SECONDS_IN_AN_HOUR (SECONDS_IN_A_MINUTE * 60)
#define SECONDS_IN_A_DAY (SECONDS_IN_AN_HOUR * 24)


/* Allocates and returns a new key containing the parameters. */
static ELMETRIC_Key
NewKey(ELMETRIC_MetricServiceTypes service,
       char *server,
       char *entity,
       ELMETRIC_MetricTypes metric,
       void *serviceKeyInfo) {
  ELMETRIC_Key k = (ELMETRIC_Key)malloc(sizeof(struct _ELMETRIC_KeyStruct));
  k->service = service;
  k->server = server;
  k->entity = entity;
  k->metric = metric;
  k->serviceKeyInfo = serviceKeyInfo;
  return k;
}


/*
 * Returns the components of #server# (format host[:port]) in #port# and the
 * #hostSize#-long string #host#.  Sets #port# to #defaultPort# if #server#
 * contains no port.  Returns 1 if successful, else 0.
 */
static int
ParseServer(const char *server,
            Port defaultPort,
            char *host,
            unsigned hostSize,
            Port *port) {
  const char *c;
  if((c = strchr(server, ':')) == NULL) {
    if(strlen(server) >= hostSize)
      return 0;
    strcpy(host, server);
    *port = defaultPort;
  }
  else if(c - server >= (int)hostSize)
    return 0;
  else {
    strncpy(host, server, c - server);
    host[c - server] = '\0';
    *port = (Port)strtol(c + 1, NULL, 10);
  }
  return 1;
}


/*
 * Returns a connection to #server#; NO_SOCKET on error.  #defaultPort# is used
 * if #server# contains no port.
 */
static Socket
ConnectToServer(const char *server,
                Port defaultPort) {
  Address addr;
  char host[256];
  Port port;
  if(!ParseServer(server, defaultPort, host, sizeof(host), &port) ||
     !AddressValue(host, &addr))
    return NO_SOCKET;
  return ConnectToPort(addr, port);
}


/* === GANGLIA === */


#define GANGLIA_DEFAULT_GMETAD_PORT 8651
#define GANGLIA_DEFAULT_GMOND_PORT 8649
static const char *GANGLIA_METRIC_NAMES[ELMETRIC_METRIC_COUNT + 1] = {
  "cpu_num", "load_one", "", "", NULL
};


/*
 * Returns an XML tree built from XML returned by #server#; NULL on error.  The
 * value returned should be freed after use.
 */
static XmlNode
GangliaGetXml(const char *server) {

  char buffer[256];
  XmlNode doc;
  int n;
  Socket sock;
  char *xml = NULL;

  if((sock=ConnectToServer(server,GANGLIA_DEFAULT_GMETAD_PORT)) == NO_SOCKET &&
     (sock=ConnectToServer(server,GANGLIA_DEFAULT_GMOND_PORT)) == NO_SOCKET)
    return NULL;
  while((n = Receive(sock, buffer, sizeof(buffer) - 1)) > 0) {
    buffer[n] = '\0';
    xml = StrAppend(xml, buffer, NULL);
  }
  Disconnect(&sock);
  if(xml == NULL)
    return NULL;
  doc = XmlFromString(xml, NULL, 1, 1, NULL, NULL, NULL);
  free(xml);
  return doc;

}


/*
 * Returns in an allocated, NULL-terminated array keys for all metrics that are
 * available from the Ganglia server #server#, or NULL on error.
 */
static ELMETRIC_Key *
GangliaGetAllKeys(const char *server) {

  XmlNode doc;
  unsigned i;
  unsigned long idx;
  unsigned j;
  unsigned len;
  XmlNodeList hostList;
  XmlNode hostNode;
  XmlNodeList metricList;
  XmlNode metricNode;
  ELMETRIC_Key *result;

  if((doc = GangliaGetXml(server)) == NULL)
    return NULL;
  hostList = DocumentGetElementsByTagName(doc, "HOST");
  len = 0;
  result = NULL;
  for(i = 0; (hostNode = NodeListItem(hostList, i)) != NULL; i++) {
    const char *hostName = ElementGetAttribute(hostNode, "NAME");
    metricList = ElementGetElementsByTagName(hostNode, "METRIC");
    for(j = 0; (metricNode = NodeListItem(metricList, j)) != NULL; j++) {
      const char *metricName = ElementGetAttribute(metricNode, "NAME");
      if(!StrArrayFind(GANGLIA_METRIC_NAMES, metricName, &idx))
        continue;
      result = realloc(result, (len + 1) * sizeof(ELMETRIC_Key));
      result[len++] = NewKey
        (ELMETRIC_GANGLIA, strdup(server), strdup(hostName),
         (ELMETRIC_MetricTypes)idx, NULL);
    }
    NodeListFree(metricList);
  }
  NodeListFree(hostList);
  result = realloc(result, (len + 1) * sizeof(ELMETRIC_Key));
  result[len] = NULL;
  return result;

}


/*
 * Returns the #metric# value for #entity# retrieved from the Ganglia server
 * #server#; NULL on error.  The returned value should be freed after use.
 */
static char *
GangliaGetMetricValue(const char *server,
                      const char *entity,
                      ELMETRIC_MetricTypes metric) {
  
  XmlNode doc;
  XmlNodeList list;
  XmlNode node;
  unsigned i;
  const char *metricName = GANGLIA_METRIC_NAMES[metric];
  char *result;
  
  if(*metricName == '\0' || (doc = GangliaGetXml(server)) == NULL)
    return NULL;
  list = DocumentGetElementsByTagName(doc, "HOST");
  for(i = 0;
      (node = NodeListItem(list, i)) != NULL && 
      strcmp(ElementGetAttribute(node, "NAME"), entity) != 0;
      i++)
    ; /* empty */
  NodeListFree(list);
  if(node != NULL) {
    list = ElementGetElementsByTagName(node, "METRIC");
    for(i = 0;
        (node = NodeListItem(list, i)) != NULL && 
        strcmp(ElementGetAttribute(node, "NAME"), metricName) != 0;
        i++)
      ; /* empty */
    NodeListFree(list);
    if(node != NULL)
      node = ElementGetAttributeNode(node, "VAL");
  }
  result = node == NULL ? NULL : strdup(NodeValue(node));
  NodeFree(doc);
  return result;

}


/*
 * Returns 1 and sets the serviceKeyInfo field of #k# appropriately if the
 * Ganglia server #server# reports #entity#'s #metric# value; else returns 0.
 */
static int
GangliaFillKey(const char *server,
               const char *entity,
               ELMETRIC_MetricTypes metric,
               ELMETRIC_Key k) {
  char *value = GangliaGetMetricValue(server, entity, metric);
  if(value == NULL)
    return 0;
  free(value);
  return 1;
}


/*
 * Contacts the Ganglia server specified by #k#, then sets the value and
 * expiration fields of #m# for #k#'s metric.  Returns 1 if successful, else 0.
 */
static int
GangliaFillMetric(ELMETRIC_Key k,
                  ELMETRIC_Metric m) {
  if((m->value = GangliaGetMetricValue(k->server,k->entity,k->metric)) == NULL)
    return 0;
  m->expiration =
    time(NULL) + (k->metric == ELMETRIC_CPU_LOAD ?
                  SECONDS_IN_A_MINUTE : SECONDS_IN_A_DAY);
  return 1;
}


/* Frees Ganglia info stored in the serviceKeyInfo field of #k#. */
#define GangliaFreeKey(k) /* NOTE: k->serviceKeyInfo not used by Ganglia. */


/* Returns 1 if #server# specifies a ganglia server, else 0. */
static int
GangliaValidServer(const char *server) {
  XmlNode doc = GangliaGetXml(server);
  if(doc == NULL)
    return 0;
  NodeFree(doc);
  return 1;
}


/* === LOCAL === */


#if !defined(ELMETRIC_HAVE_SYSCONF)
#  undef SYSCONF_NPROC
#elif defined(_SC_NPROC_CONF)
#  define SYSCONF_NPROC _SC_NPROC_CONF
#elif defined(_SC_NPROCESSORS_CONF)
#  define SYSCONF_NPROC _SC_NPROCESSORS_CONF
#elif defined(_SC_CRAY_NCPU)
#  define SYSCONF_NPROC _SC_CRAY_NCPU
#endif


/*
 * Returns 1 and sets the serviceKeyInfo field of #k# appropriately if the
 * local server #server# reports #entity#'s #metric# value; else returns 0.
 */
static int
LocalFillKey(const char *server,
             const char *entity,
             ELMETRIC_MetricTypes metric,
             ELMETRIC_Key k) {
    char *myName = MyMachineName();
    int result = (
#if defined(SYSCONF_NPROC)
    metric == ELMETRIC_CPU_COUNT ||
#endif
    metric == ELMETRIC_CPU_LOAD) && strcmp(entity, myName) == 0;
    free(myName);
    return result;
}


/*
 * Contacts the local server specified by #k#, then sets the value and
 * expiration fields of #m# for #k#'s metric.  Returns 1 if successful, else 0.
 */
static int
LocalFillMetric(ELMETRIC_Key k,
                ELMETRIC_Metric m) {
  char buffer[256];
  char *c;
  FILE *f;
  if(k->metric == ELMETRIC_CPU_COUNT) {
#ifdef SYSCONF_NPROC
    int count = sysconf(SYSCONF_NPROC);
    sprintf(buffer, "%d", count == 0 ? 1 : count);
    m->value = strdup(buffer);
    m->expiration = time(NULL) + SECONDS_IN_A_DAY;
    return 1;
#else
    return 0;
#endif
  }
  else if(k->metric == ELMETRIC_CPU_LOAD) {
    if((f = popen("uptime", "r")) == NULL)
      return 0;
    c = fgets(buffer, sizeof(buffer), f);
    pclose(f);
    if(c == NULL || (c = strstr(buffer, "average:")) == NULL)
      return 0;
    sprintf(buffer, "%.2f", strtod(c + 8, NULL));
    m->value = strdup(buffer);
    m->expiration = time(NULL) + SECONDS_IN_A_MINUTE;
    return 1;
  }
  else
    return 0;
}


/* Frees local info stored in the serviceKeyInfo field of #k#. */
#define LocalFreeKey(k) /* NOTE: k->serviceKeyInfo not used by local. */


/*
 * Returns in an allocated, NULL-terminated array keys for all metrics that are
 * available from the local server #server#, or NULL on error.
 */
static ELMETRIC_Key *
LocalGetAllKeys(const char *server) {
  ELMETRIC_Key *result;
  result = (ELMETRIC_Key *)malloc(3 * sizeof(ELMETRIC_Key));
  result[0] = NewKey
   (ELMETRIC_LOCAL, strdup(server), MyMachineName(), ELMETRIC_CPU_LOAD, NULL);
  result[1] = NULL;
#ifdef SYSCONF_NPROC
  result[1] = NewKey
    (ELMETRIC_LOCAL,strdup(server),MyMachineName(),ELMETRIC_CPU_COUNT,NULL);
#endif
  result[2] = NULL;
  return result;

}
/* for GHS -START*/
static ELMETRIC_Key *
GhsGetAllKeys(const char *server) {
  ELMETRIC_Key *result;
  result = (ELMETRIC_Key *)malloc(4 * sizeof(ELMETRIC_Key));
  result[0] = NewKey
    (ELMETRIC_GHS, strdup(server), MyMachineName(), ELMETRIC_CPU_LOAD, NULL);
  result[1] = NewKey
    (ELMETRIC_GHS, strdup(server), MyMachineName(), ELMETRIC_JOB_ARRIVAL_RATE, NULL);
  result[2] = NewKey
    (ELMETRIC_GHS, strdup(server), MyMachineName(), ELMETRIC_SERVICE_TIME_DEVIATION, NULL);
  result[3] = NewKey
    (ELMETRIC_GHS, strdup(server), MyMachineName(), ELMETRIC_COMPUTING_CAPACITY, NULL);
  return result;

}
/* for GHS -END*/

/* === MDS === */


#define MDS_DEFAULT_GIIS_PORT 2135
static const char *MDS_METRIC_NAMES[ELMETRIC_METRIC_COUNT + 1] = {
  "Mds-Cpu-Total-count", "Mds-Cpu-Free-1minX100", "", "", NULL
};


/*
 * Returns the #metric# value for #entity# retrieved from the MDS server
 * #server#; NULL on error.  The returned value should be freed after use.
 */
static char *
MdsGetMetricValue(const char *server,
                  const char *entity,
                  ELMETRIC_MetricTypes metric) {

  const char *attr = MDS_METRIC_NAMES[metric];
  unsigned attrLen;
  char buffer[256];
  char *c;
  char *cmd;
  char *dn;
  FILE *f;
  char *filter;
  char host[256];
  char *mdsPath;
  Port port;

  if(*attr == '\0' ||
     !ParseServer(server, MDS_DEFAULT_GIIS_PORT, host, sizeof(host), &port))
    return NULL;
  if((mdsPath = ASFILE_FindFile(getenv("PATH"), "ldapsearch")) == NULL &&
     (mdsPath = ASFILE_FindFile(getenv("PATH"), "mdssearch")) == NULL)
    return NULL;

  sprintf(buffer, "%d", port);
  dn = StrAppend
    (NULL, "mds-host-hn=", entity, ",", "mds-vo-name=local,o=grid", NULL);
  filter = StrAppend(NULL, "(", attr, "=*)", NULL);
  cmd = StrAppend
    (NULL, mdsPath, " -h ", host, " -p ", buffer, " -x -b '", dn, "' -l 10 '",
     filter, "' '", attr, "'", NULL);
  free(dn);
  free(filter);
  free(mdsPath);

  f = popen(cmd, "r");
  free(cmd);
  if(f == NULL)
    return NULL;

  attrLen = strlen(attr);
  while((c = fgets(buffer, sizeof(buffer), f)) != NULL &&
        (buffer[attrLen] != ':' || strncmp(buffer, attr, attrLen) != 0))
    ; /* empty */
  pclose(f);
  if(c == NULL)
    return NULL;

  for(c = strchr(buffer, '\n') - 1; *c == ' '; c--)
    ; /* empty */
  *(c + 1) = '\0';
  for(c = strchr(buffer, ':') + 1; *c == ' '; c++)
    ; /* empty */
  if(metric == ELMETRIC_CPU_LOAD)
    sprintf(c, "%.2f", 1.0 - strtol(c, NULL, 10) / 100.0);
  return strdup(c);

}


/*
 * Returns 1 and sets the serviceKeyInfo field of #k# appropriately if the MDS
 * server #server# reports #entity#'s #metric# value; else returns 0.
 */
static int
MdsFillKey(const char *server,
           const char *entity,
           ELMETRIC_MetricTypes metric,
           ELMETRIC_Key k) {
  char *value = MdsGetMetricValue(server, entity, metric);
  if(value == NULL)
    return 0;
  free(value);
  return 1;
}


/*
 * Contacts the MDS server specified by #k#, then sets the value and expiration
 * fields of #m# for #k#'s metric.  Returns 1 if successful, else 0.
 */
static int
MdsFillMetric(ELMETRIC_Key k,
              ELMETRIC_Metric m) {
  if((m->value = MdsGetMetricValue(k->server, k->entity, k->metric)) == NULL)
    return 0;
  m->expiration =
    time(NULL) + (k->metric == ELMETRIC_CPU_LOAD ?
                  SECONDS_IN_A_MINUTE : SECONDS_IN_A_DAY);
  return 1;
}


/* Frees MDS info stored in the serviceKeyInfo field of #k#. */
#define MdsFreeKey(k) /* NOTE: k->serviceKeyInfo not used by MDS. */


/*
 * Returns in an allocated, NULL-terminated array keys for all metrics that are
 * available from the MDS server #server#, or NULL on error.
 */
static ELMETRIC_Key *
MdsGetAllKeys(const char *server) {

  char buffer[256];
  char *c;
  char *cmd;
  FILE *f;
  char host[256];
  unsigned long idx;
  unsigned len = 0;
  char *mdsPath;
  Port port;
  ELMETRIC_Key *result = NULL;
  char *value;

  if(!ParseServer(server, MDS_DEFAULT_GIIS_PORT, host, sizeof(host), &port))
    return NULL;
  if((mdsPath = ASFILE_FindFile(getenv("PATH"), "ldapsearch")) == NULL &&
     (mdsPath = ASFILE_FindFile(getenv("PATH"), "mdssearch")) == NULL)
    return NULL;
  sprintf(buffer, "%d", port);
  cmd = StrAppend
    (NULL, mdsPath, " -h ", host, " -p ", buffer,
     " -x -b 'mds-host-hn=*,mds-vo-name=local,o=grid' -l 10", NULL);
  f = popen(cmd, "r");
  free(cmd);
  free(mdsPath);
  if(f == NULL)
    return NULL;

  while(fgets(buffer, sizeof(buffer), f) != NULL) {
    if(buffer[0] == '#' || (c = strchr(buffer, ':')) == NULL)
      continue;
    *c = '\0';
    for(value = c + 1; *value == ' '; value++)
      ; /* empty */
    if(strcmp(buffer, "dn") == 0 && strncmp(value, "Mds-Host-hn=", 12) == 0) {
      if((c = strchr(value, ',')) != NULL)
        *c = '\0';
      strcpy(host, value + 12);
      continue;
    }
    if(!StrArrayFind(MDS_METRIC_NAMES, buffer, &idx))
      continue;
    result = realloc(result, (len + 1) * sizeof(ELMETRIC_Key));
    result[len++] = NewKey
      (ELMETRIC_MDS, strdup(server), strdup(host), (ELMETRIC_MetricTypes)idx,
       NULL);
  }

  pclose(f);
  return result;

}


/* Returns 1 if #server# specifies an MDS server, else 0. */
static int
MdsValidServer(const char *server) {
  Socket sock;
  if((sock = ConnectToServer(server, MDS_DEFAULT_GIIS_PORT)) == NO_SOCKET)
    return 0;
  Disconnect(&sock);
  return 1;
}


/* === NWS === */


/*
 * NWS ServiceKeyInfo.  We keep memory contact info and the series name so that
 * we don't have to go to the name server each time we retrieve a metric.
 */
typedef struct _NwsKeyInfoStruct {
  char *memory;
  char *series;
} *NwsKeyInfo;

static const char *NWS_METRIC_NAMES[ELMETRIC_METRIC_COUNT + 1] = {
  "", "availableCpu", "tcpBandwidth", "tcpLatency", NULL
};


/*
 * Sets the serviceKeyInfo field of #k# from the fields in the NWS name server
 * object #object#.
 */
static void
NwsFillKeyFromObject(const char *object,
                     ELMETRIC_Key k) {
  NwsKeyInfo info;
  char *memory;
  char *series;
  k->serviceKeyInfo = NULL;
  if((memory = strstr(object, "memory:")) == NULL ||
     (series = strstr(object, "name:")) == NULL)
    return;
  memory += 7;
  series += 5;
  info = (NwsKeyInfo)malloc(sizeof(struct _NwsKeyInfoStruct));
  info->memory = StrNDup(memory, strchr(memory, '\t') - memory);
  info->series = StrNDup(series, strchr(series, '\t') - series);
  k->serviceKeyInfo = info;
}


/*
 * Returns 1 and sets the serviceKeyInfo field of #k# appropriately if the NWS
 * server #server# reports #entity#'s #metric# value; else returns 0.
 */
static int
NwsFillKey(const char *server,
           const char *entity,
           ELMETRIC_MetricTypes metric,
           ELMETRIC_Key k) {

  const char *attr = NWS_METRIC_NAMES[metric];
  char *c;
  char *filter;
  NWSC_HostConnection ns;
  char *objects;

  if(*attr == '\0' || !NWSC_HostConnect(server, NWSC_NAME_SERVER_PORT, &ns))
    return 0;

  filter = StrAppend(NULL, "(&(objectclass=nwsSeries)(resource=", attr,
                     ")(host=", entity, NULL);
  if((c = strchr(filter, '-')) != NULL)
    StrReplace(&filter, c - filter, 1, ":*)(target=");
  filter = StrAppend(filter, ":*))", NULL);

  objects = NULL;
  NWSC_NameServerSearch(ns, filter, &objects);
  NWSC_HostDisconnect(&ns);
  free(filter);
  if(objects == NULL)
    return 0;
  NwsFillKeyFromObject(objects, k);
  free(objects);
  return 1;

}


/*
 * Contacts the NWS server specified by #k#, then sets the value and expiration
 * fields of #m# for #k#'s metric.  Returns 1 if successful, else 0.
 */
static int
NwsFillMetric(ELMETRIC_Key k,
              ELMETRIC_Metric m) {

  NwsKeyInfo info = (NwsKeyInfo)k->serviceKeyInfo;
  char *end;
  unsigned long lastTime;
  NWSC_HostConnection mem;
  char *objects;
  unsigned size;
  unsigned long thisTime;

  if(!NWSC_HostConnect(info->memory, NWSC_MEMORY_PORT, &mem))
    return 0;

  objects = NULL;
  NWSC_MemoryFetch(mem, info->series, 2, &objects, &size);
  NWSC_HostDisconnect(&mem);
  if(objects == NULL)
    return 0;

  lastTime = strtol(objects + size, NULL, 10);
  thisTime = strtol(objects, &end, 10);
  objects[size] = '\0';
  while(*end == ' ')
    end++;
  if(k->metric == ELMETRIC_CPU_LOAD)
    sprintf(end, "%f", 1.0 / strtod(end, NULL) - 1.0);

  m->value = strdup(end);
  m->expiration = thisTime + (thisTime - lastTime);
  free(objects);
  return 1;

}


/* Frees NWS info stored in the serviceKeyInfo field of #k#. */
static void
NwsFreeKey(ELMETRIC_Key k) {
  NwsKeyInfo info = (NwsKeyInfo)k->serviceKeyInfo;
  if(info == NULL)
    return;
  if(info->memory != NULL)
    free(info->memory);
  if(info->series != NULL)
    free(info->series);
  free(info);
}


/*
 * Returns in an allocated, NULL-terminated array keys for all metrics that are
 * available from the Ganglia server #server#, or NULL on error.
 */
static ELMETRIC_Key *
NwsGetAllKeys(const char *server) {

  char *c;
  char *host;
  unsigned long idx;
  unsigned len = 0;
  NWSC_HostConnection ns;
  char *objects = NULL;
  char *resource;
  ELMETRIC_Key *result = NULL;

  if(!NWSC_HostConnect(server, NWSC_NAME_SERVER_PORT, &ns))
    return NULL;
  NWSC_NameServerSearch(ns, "&(objectclass=nwsSeries)", &objects);
  NWSC_HostDisconnect(&ns);
  if(objects == NULL)
    return NULL;
  for(c = objects; *c != '\0'; c = strstr(c, "\t\t") + 2) {
    if((host = strstr(c, "host:")) == NULL ||
       (resource = strstr(c, "resource:")) == NULL)
      continue;
    host += 5;
    resource += 9;
    if(!StrArrayNFind(NWS_METRIC_NAMES, ELMETRIC_METRIC_COUNT, resource,
                      strchr(resource, '\t') - resource, &idx))
      continue;
    result = (ELMETRIC_Key *)realloc(result, (len + 1) * sizeof(ELMETRIC_Key));
    result[len++] = NewKey
      (ELMETRIC_NWS, strdup(server), StrNDup(host, strchr(host, '\t') - host),
       (ELMETRIC_MetricTypes)idx, NULL);
    NwsFillKeyFromObject(c, result[len - 1]);
  }
  free(objects);
  result = (ELMETRIC_Key *)realloc(result, (len + 1) * sizeof(ELMETRIC_Key));
  result[len] = NULL;
  return result;

}


/* Returns 1 if #server# specifies an NWS server, else 0. */
static int
NwsValidServer(const char *server) {
  NWSC_HostConnection ns;
  if(!NWSC_HostConnect(server, NWSC_NAME_SERVER_PORT, &ns))
    return 0;
  NWSC_HostDisconnect(&ns);
  return 1;
}


/* === PUBLIC API === */


void
ELMETRIC_FreeKey(ELMETRIC_Key k) {
  if(k->server != NULL)
    free(k->server);
  if(k->entity != NULL)
    free(k->entity);
   if(k->service == ELMETRIC_GANGLIA)
     GangliaFreeKey(k);
   else if(k->service == ELMETRIC_LOCAL)
     LocalFreeKey(k);
   else if(k->service == ELMETRIC_MDS)
     MdsFreeKey(k);
   else if(k->service == ELMETRIC_NWS)
     NwsFreeKey(k);
  free(k);
}


void
ELMETRIC_FreeMetric(ELMETRIC_Metric m) {
  if(m->value != NULL)
    free(m->value);
  free(m);
}

/* for GHS -START*/
ELMETRIC_Key *
ELMETRIC_GetAllKeys(ELMETRIC_MetricServiceTypes service,
                    const char *server) {
  return service == ELMETRIC_GANGLIA ? GangliaGetAllKeys(server) :
         service == ELMETRIC_LOCAL   ? LocalGetAllKeys(server) :
         service == ELMETRIC_MDS     ? MdsGetAllKeys(server) :
         service == ELMETRIC_NWS     ? NwsGetAllKeys(server) :
         service == ELMETRIC_GHS     ? GhsGetAllKeys(server) :
         NULL;
}
/* for GHS -END*/

/* for GHS -START*/
int
GhsFillKey(const char *server,
               const char *entity,
               ELMETRIC_MetricTypes metric,
               ELMETRIC_Key k) {

  int sockServerfd, numBytes;
  char buf[MAX_DATA_SIZE];
  struct hostent *he;
  struct sockaddr_in serverAddr; // GHS server's address information
  int messageType;

  char str[4][256];
  //char ghs_host[256];
  int result = 0;

  //printf("the server is %s \n", server);
  if ((he = gethostbyname(server)) == NULL) // get the host info
  {
    perror("gethostbyname");
    exit(1);
  }

  if ((sockServerfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
  {
    perror("socket");
    exit(1);
  }

  serverAddr.sin_family = AF_INET;    // host byte order
  serverAddr.sin_port = htons(SERVER_PORT);  // short, network byte order
  serverAddr.sin_addr = *((struct in_addr *)he->h_addr);
  memset(&(serverAddr.sin_zero), '\0', 8);  // zero the rest of the struct

  if (connect(sockServerfd, (struct sockaddr *)&serverAddr, sizeof(struct sockaddr)) == -1)
  {
    perror("connect");
    exit(1);
  }

  sprintf(buf, "%d %s %d \n", SERVER_INQUIRY, entity, PRED_SCOPE);
  //printf("the buf is %s", buf);
  if (send(sockServerfd, buf, MAX_DATA_SIZE, 0) == -1) perror("send");

  if ((numBytes = recv(sockServerfd, buf, MAX_DATA_SIZE - 1, 0)) == -1)
  {
    perror("recv");
    exit(1);
  }
  //printf("the buf received is %s", buf);
 
  sscanf(buf, "%d %s %s %s %s", &messageType, &str[0], &str[1], &str[2], &str[3]);

  close(sockServerfd);

  if((metric == ELMETRIC_CPU_LOAD || metric == ELMETRIC_JOB_ARRIVAL_RATE || metric == ELMETRIC_SERVICE_TIME_DEVIATION || metric == ELMETRIC_COMPUTING_CAPACITY)) {
    k->serviceKeyInfo = metric == ELMETRIC_CPU_LOAD ? &str[0] : metric == ELMETRIC_JOB_ARRIVAL_RATE ? &str[1] : metric ==  ELMETRIC_SERVICE_TIME_DEVIATION ? &str[2] : metric == ELMETRIC_COMPUTING_CAPACITY ? &str[3]:0;

    result = 1;
  }

  return result;

}
/* for GHS -END*/

/* for GHS -START*/
ELMETRIC_Key
ELMETRIC_GetKey(ELMETRIC_MetricServiceTypes service,
                const char *server,
                const char *entity,
                ELMETRIC_MetricTypes metric) {
  ELMETRIC_Key k = NewKey
    (service, server == NULL ? NULL : strdup(server),
     entity == NULL ? NULL : strdup(entity), metric, NULL);
  if(service == ELMETRIC_GANGLIA ? GangliaFillKey(server, entity, metric, k) :
     service == ELMETRIC_LOCAL   ? LocalFillKey(server, entity, metric, k) :
     service == ELMETRIC_MDS     ? MdsFillKey(server, entity, metric, k) :
     service == ELMETRIC_GHS     ? GhsFillKey(server, entity, metric, k) :
     service == ELMETRIC_NWS     ? NwsFillKey(server, entity, metric, k) : 0)
  {
    return k;
  }

  ELMETRIC_FreeKey(k);
  return ELMETRIC_NO_KEY;
}
/* for GHS -END*/

/* for GHS -START*/
static int
GhsFillMetric(ELMETRIC_Key k, ELMETRIC_Metric m) {
  m->value = strdup((char*)k->serviceKeyInfo);
  return 1;
}
/* for GHS - END*/

char *
ELMETRIC_GetKeyEntity(ELMETRIC_Key k) {
  return k->entity;
}


ELMETRIC_MetricTypes
ELMETRIC_GetKeyMetric(ELMETRIC_Key k) {
  return k->metric;
}


char *
ELMETRIC_GetKeyServer(ELMETRIC_Key k) {
  return k->server;
}


ELMETRIC_MetricServiceTypes
ELMETRIC_GetKeyService(ELMETRIC_Key k) {
  return k->service;
}

/* for GHS -START*/
ELMETRIC_Metric
ELMETRIC_GetMetric(ELMETRIC_Key k) {
  ELMETRIC_Metric m =
    (ELMETRIC_Metric)malloc(sizeof(struct _ELMETRIC_MetricStruct));
  m->expiration = 0;
  m->key = k;
  m->value = NULL;
  //printf("%s %s \n", k->server, k->entity);
  if(k->service == ELMETRIC_GANGLIA ? GangliaFillMetric(k, m) :
     k->service == ELMETRIC_LOCAL   ? LocalFillMetric(k, m) :
     k->service == ELMETRIC_MDS     ? MdsFillMetric(k, m) :
     k->service == ELMETRIC_GHS     ? GhsFillMetric(k, m) :
     k->service == ELMETRIC_NWS     ? NwsFillMetric(k, m) : 0)
{
    return m;
}
  ELMETRIC_FreeMetric(m);
  return ELMETRIC_NO_METRIC;
}
/* for GHS -END*/


unsigned long
ELMETRIC_GetMetricExpiration(ELMETRIC_Metric m) {
  return m->expiration;
}


ELMETRIC_Key
ELMETRIC_GetMetricKey(ELMETRIC_Metric m) {
  return m->key;
}


char *
ELMETRIC_GetMetricValue(ELMETRIC_Metric m) {
  return m->value;
}


int
ELMETRIC_ValidServer(ELMETRIC_MetricServiceTypes type,
                     const char *server) {
  return
    type == ELMETRIC_GANGLIA ? GangliaValidServer(server) :
    type == ELMETRIC_LOCAL   ? 1 :
    type == ELMETRIC_MDS     ? MdsValidServer(server) :
    type == ELMETRIC_NWS     ? NwsValidServer(server) : 0;
}
