#! /usr/bin/perl -w
# $Id: rash.pl,v 1.20 2004/03/24 19:30:48 jhayes Exp $

my $USAGE = 'rash [-w] [name@host[:port] [command]]';

# This is a rash client.  The name@host:port argument specifies the host to
# connect to--the host:port portion is contact information for the rash daemon
# and the name portion the host name.  (Instead of a name, a number can be used
# to select a host positionally.)  The default is '0@localhost:7000', with the
# leading '0' selecting the rash daemon.  If -w is specified, the client will
# block until the specified host connects to the daemon; otherwise, it will
# exit immediately with a non-zero value if the specified host is not already
# connected.  If given a command, the client will execute that command and then
# exit; otherwise, it will accept and execute commands interactively.  The
# client exit code is the exit code of the last command completed.  When
# running interactively, the '?' command will show the command history,
# including the exit code for any commands that have completed.
#
# See rashd for description of the rash protocol.

use IO::Select;
use IO::Socket;

my $CRLF = "\015\012";
my $agent;
my $command;
my $foreground;
my $hosts = '';
my $pid;
my %pidCommand;
my %pidExit;
my @pidOrder;
my $sock;
my $waitForHost;

$waitForHost = shift if $#ARGV >= 0 && $ARGV[0] eq '-w';
unshift @ARGV, '0@localhost' if $#ARGV < 0;
die "usage: $USAGE" if $ARGV[0] !~ /^([^\@]+)\@([^:]+)(:(\d+))?$/;
$sock = IO::Socket->new
  (Domain => AF_INET, PeerAddr => $2, PeerPort => defined $4 ? $4 : 7000)
  or die "socket: $!";
$agent = $1;
shift and ($command = join(' ', @ARGV)) if $#ARGV > 0;

$SIG{PIPE} = 'IGNORE';
autoflush STDOUT 1;

# Asks that #cmd# be executed; identifies the resulting process by #pid#.
sub DoCommand {
  my ($cmd, $pid) = @_;
  my $host = $agent;
  if($cmd=~/^(AUTOWHO|KILL \S+|ME|QUIT|RENAME \S+ \S+|STOP \S+|WHO)$/i) {
    $cmd =~ s/(\S+)//;
    $cmd = uc($1) . $cmd;
    $host = '_rashd';
  }
  if($cmd =~ /^KILL (\S)/) {
    $pidExit{$1} = 'K';
    $foreground = undef if defined $foreground && $foreground eq $1;
  } elsif($cmd !~ /^QUIT/) {
    $pidCommand{$pid} = $cmd;
    push(@pidOrder, $pid);
    $cmd = "RUN $host $pid $cmd";
  }
  &Send($sock, $cmd);
}

# Disconnects from the daemon and exits with the appropriate code.
sub Done {
  my ($exitCode, $pid);
  &DoCommand('QUIT');
  $sock->close();
  foreach $pid (@pidOrder) {
    $exitCode = $pidExit{$pid}
      if defined $pidExit{$pid} and $pidExit{$pid} ne 'K';
  }
  exit(defined $exitCode ? $exitCode : 0);
}

# Processes the newly-arrived #msg#.
sub ProcessMessage {
  my ($msg) = @_;
  die "Lost connection\n" if !defined $msg;
  return if $msg !~ s/^(STDERR|STDOUT|FINISHED) (\S+) //;
  return if defined $pidExit{$2};
  return print STDERR $msg if $1 eq 'STDERR';
  return $2 eq 'AW' ? ($hosts = $msg) : print STDOUT $msg if $1 eq 'STDOUT';
  $foreground = undef if defined $foreground && $foreground eq $2;
  $pidExit{$2} = $msg;
}

# Returns a message received on #from#, minus the CRLF terminator and with
# special characters decoded.
sub Receive {
  my ($from) = @_;
  my $buf = ' ';
  my $msg = '';
  $msg .= $buf while $msg !~ s/$CRLF$// && sysread($from, $buf, 1);
  $msg =~ s/\\n/\n/g;
  $msg =~ s/\\\\/\\/g;
  return $msg =~ /./ ? $msg : undef;
}

# Sends #msg#, followed by CRLF and with special characters encoded, to #to#.
sub Send {
  my ($to, $msg) = @_;
  $msg =~ s/\\/\\\\/g;
  $msg =~ s/\n/\\n/g;
  $to->print("$msg$CRLF");
}

# Wait (-w) or test (no -w) for the presence of the named host.
&DoCommand('AUTOWHO', 'AW');
while(1) {
  my @names;
  &ProcessMessage(&Receive($sock));
  last if " $hosts " =~ / $agent /;
  # Interpret a numeric host as an index into the host list.
  ($agent = $names[$agent]) and last
     if $agent =~ /^\d+/ && (@names = split(' ', $hosts)) > $agent + 0;
  exit 1 if !defined $waitForHost;
}

if(!defined $command) {
  my $prompt = defined $ENV{'RASHPROMPT'} ? $ENV{'RASHPROMPT'} : 'rash> ';
  $SIG{INT} = sub {
    &DoCommand("KILL $foreground", undef) if defined $foreground;
  };
  while(1) {
    print $prompt if !defined $foreground;
    IO::Select->new(\*STDIN, $sock)->can_read();
    &ProcessMessage(&Receive($sock)) while IO::Select->new($sock)->can_read(0);
    if(IO::Select->new(\*STDIN)->can_read(0)) {
      my $text = <STDIN>;
      last if !defined $text;
      if(defined $foreground) {
        &Send($sock, "STDIN $foreground $text");
      } else {
        $text =~ s/(^\s+)|(\s+$)//g;
        next if $text eq '';
        if($text eq '?') {
          foreach $pid (@pidOrder) {
            print STDOUT "[$pid] $pidCommand{$pid}";
            print STDOUT " [$pidExit{$pid}]" if defined $pidExit{$pid};
            print STDOUT "\n";
          }
          next;
        }
        last if $text eq 'quit';
        $pid = $#pidOrder + 1;
        if($text =~ s/\s*&\s*$//) {
          print "[$pid]\n";
        } elsif($text !~ /^\s*kill /i) {
          $foreground = $pid;
        }
        &DoCommand($text, $pid);
      }
    }
  }
} else {
  $SIG{INT} = \&Done;
  $pid = '0';
  $command =~ s/(^\s+)|(\s+$)//g;
  &DoCommand($command, $pid);
  $pidExit{$pid} = $pid if $command eq 'quit';
  &ProcessMessage(&Receive($sock)) while !defined $pidExit{$pid};
}

&Done();
