#! /usr/bin/perl -w
# $Id: rashd.pl,v 1.17 2004/03/24 19:30:48 jhayes Exp $

my $USAGE = "rashd [-d] [-v] [-p port]";

# This is the rash daemon, normally run on the gateway to a set of batch or
# firewall-protected resources.  The default port is 7000.  The daemon accepts
# and transmits the following CRLF-terminated text messages.
#
# FINISHED pid code
#   From agent; forwarded to client.  Indicates that process #pid# has exited
#   with code #code#.
#
# KILL pid
#   From client; forwarded to agent.  Requests that process #pid# be terminated.
#
# QUIT
#   From agent or client.  Requests that the connection be closed.
#
# RUN agent pid command
#   From client; forwarded to agent.  Requests that the agent #agent# be asked
#   to run #command#.  #pid# identifies the resulting process.  The daemon
#   itself serves as the agent for these commands:
#   AUTOWHO
#     Returns a STDOUT message that contains the list of connection names and
#     sends a new STDOUT message every time the list changes.  Never completes;
#     the client must kill the process to terminate it.
#   ME
#     Returns a STDOUT message that contains the client's connection name
#     followed by a FINISHED 0 message.
#   RENAME old new
#     Requests that the connection #old# be reidentified with #new#.
#   STOP name
#     Requests that the connection to #name# be closed.
#   WHO
#     Returns a STDOUT message that contains the list of connection names
#     followed by a FINISHED 0 message.
#
# STDERR pid text
#   From agent; forwarded to client.  Indicates that process #pid# has written
#   #text# on standard error.
#
# STDIN
#   From client; forwarded to agent.  Indicates that #text# should be written
#   to process #pid# on standard input.
#   
# STDOUT
#   From agent; forwarded to client.  Indicates that process #pid# has written
#   #text# on standard output.

use IO::Select;
use IO::Socket;

my $CRLF = "\015\012";
my $connection;
my %connections;
my $count = 0;
my $debug;
my $ear;
my %order;
my %pidAgents;
my %pidClients;
my $port = 7000;
my $verbose;

$debug = shift if $#ARGV >= 0 && $ARGV[0] eq '-d';
$verbose = shift if $#ARGV >= 0 && $ARGV[0] eq '-v';
if($#ARGV >= 0 && $ARGV[0] eq '-p') {
  shift;
  die "usage: $USAGE" if !defined ($port = shift);
}
die "usage: $USAGE" if $#ARGV >= 0;
$ear = IO::Socket->new
  (Domain => AF_INET, ReuseAddr => 1, LocalPort => $port, Listen => SOMAXCONN)
  or die "listen: $!";
$connections{'_rashd'} = $ear;
$order{$ear} = $count++;
$SIG{HUP} = $SIG{PIPE} = 'IGNORE';

# Reports connection population to all AUTOWHO pids.
sub DoAutoWhos {
  my $pid;
  foreach $pid (keys %pidAgents) {
    &Reply($pid, 'STDOUT', &SortNames(), undef) if $pidAgents{$pid} == $ear;
  }
}

# Performs the daemon-specific command #cmd# newly arrived on #connection#.
# Labels all command output with #pid#.
sub DoCommand {
  my ($connection, $pid, $cmd) = @_;
  if($cmd eq 'AUTOWHO') {
    &Reply($pid, 'STDOUT', &SortNames(), undef);
  } elsif($cmd eq 'ME') {
    &Reply($pid, 'STDOUT', &GetConnectionName($connection), 0);
  } elsif($cmd =~ /^RENAME (\S+) (\S+)$/) {
    return &Reply($pid, 'STDERR', "Unknown host '$1'", 1)
      if !defined $connections{$1};
    return &Reply($pid, 'STDERR', "'$2' in use", 1) if defined $connections{$2};
    $connections{$2} = $connections{$1};
    delete $connections{$1};
    &Reply($pid, undef, undef, 0);
    &DoAutoWhos();
  } elsif($cmd =~ s/^STOP (\S+)$//) {
    return &Reply($pid, 'STDERR', "Unknown host '$1'", 1)
      if !defined $connections{$1};
    &Reply($pid, undef, undef, 0);
    return &Forget($connections{$1}) if $connections{$1} != $ear;
    close $ear;
    exit 0;
  } elsif($cmd eq 'WHO') {
    &Reply($pid, 'STDOUT', &SortNames(), 0);
  } else {
    &Reply($pid, 'STDERR', "Unknown command '$cmd'", 1);
  }
}

# Drops #connection# after killing any associated processes.
sub Forget {
  my ($connection) = @_;
  my $pid;
  foreach $pid (keys %pidAgents) {
    next if $pidAgents{$pid} != $connection && $pidClients{$pid} != $connection;
    &Send($pidAgents{$pid}, "KILL $pid")
      if $pidAgents{$pid} != $ear && $pidAgents{$pid} != $connection;
    $pid =~ /#(.*)$/;
    &Send($pidClients{$pid}, "FINISHED $1 255")
      if $pidClients{$pid} != $connection;
    delete $pidAgents{$pid};
    delete $pidClients{$pid};
  }
  $connection->close();
  delete $connections{&GetConnectionName($connection)};
  delete $order{$connection};
  &DoAutoWhos();
}

# Returns the identifier associated with #connection#.
sub GetConnectionName {
  my ($connection) = @_;
  my $name;
  return if !defined $connection;
  foreach $name (keys %connections) {
    return $name if $connections{$name} == $connection;
  }
}

# Responds to the message #msg# newly-arrived on #connection#.
sub ProcessMessage {
  my ($connection, $message) = @_;
  my $pid;

  return &Forget($connection) if !defined $message;
  print STDOUT
    "Received '$message' from " . &GetConnectionName($connection) . "\n";

  if($message =~ s/^FINISHED (\S+) //) {
    $pid = $1;
    $pid =~ /#(.*)$/;
    &Send($pidClients{$pid}, "FINISHED $1 $message")
      if defined $pidClients{$pid};
    delete $pidAgents{$pid};
    delete $pidClients{$pid};
  } elsif($message =~ /^KILL (\S+)$/) {
    $pid = &GetConnectionName($connection) . "#$1";
    &Send($pidAgents{$pid}, "KILL $pid")
      if defined $pidAgents{$pid} && $pidAgents{$pid} != $ear;
    delete $pidAgents{$pid};
    delete $pidClients{$pid};
  } elsif($message eq 'QUIT') {
    &Forget($connection);
  } elsif($message =~ s/^RUN (\S+) (\S+) //) {
    if(!defined $connections{$1}) {
      &Send($connection, "STDERR $2 Unknown host '$1'\\n");
      &Send($connection, "FINISHED $2 255");
      return;
    }
    $pid = &GetConnectionName($connection) . "#$2";
    &Send($pidAgents{$pid}, "KILL $pid")
      if defined $pidAgents{$pid} && $pidAgents{$pid} != $ear;
    $pidAgents{$pid} = $connections{$1};
    $pidClients{$pid} = $connection;
    if($pidAgents{$pid} == $ear) {
      &DoCommand($pidClients{$pid}, $pid, $message);
    } else {
      &Send($pidAgents{$pid}, "RUN $1 $pid $message");
    }
  } elsif($message =~ s/^(STDERR|STDOUT) (\S+) //) {
    my $stream = $1;
    $pid = $2;
    $pid =~ /#(.*)$/;
    &Send($pidClients{$pid}, "$stream $1 $message")
      if defined $pidClients{$pid};
  } elsif($message =~ s/^STDIN (\S+) //) {
    $pid = &GetConnectionName($connection) . "#$1";
    &Send($pidAgents{$pid}, "STDIN $pid $message")
      if defined $pidAgents{$pid} && $pidAgents{$pid} != $ear;
  }

}

# Returns a message received on #from#, minus the CRLF terminator.
sub Receive {
  my ($from) = @_;
  my $buf = ' ';
  my $msg = '';
  $msg .= $buf while $msg !~ s/$CRLF$// && sysread($from, $buf, 1);
  return $msg =~ /./ ? $msg : undef;
}

# A convenience function for generating a pair of messages about process #pid#.
# The first is #stream# output containing #output#; the second a FINISHED
# message that contains #exitCode#.
sub Reply {
  my ($pid, $stream, $output, $exitCode) = @_;
  &ProcessMessage($ear, "$stream $pid $output\\n") if defined $stream;
  &ProcessMessage($ear, "FINISHED $pid $exitCode") if defined $exitCode;
}

# Sends #msg# followed by CRLF to #to#. (Separated into a function for debug.)
sub Send {
  my ($to, $msg) = @_;
  print STDOUT "$msg\n";
  $to->print("$msg$CRLF");
}

# Returns the set of connection names, sorted by connection order.
sub SortNames {
  return join(' ', sort {$order{$connections{$a}} - $order{$connections{$b}}}
                   keys %connections);
}

if(!defined $debug) {
  # Put ourselves into the background.
  exit 0 if fork() != 0;
  setpgrp;
}
if(!defined $verbose) {
  # Suppress debugging output.
  close STDIN;
  open STDOUT, '/dev/null';
  open STDERR, '/dev/null';
}

while(1) {
  my $sel = IO::Select->new(values %connections);
  foreach $connection ($sel->can_read()) {
    print STDOUT "Read from $connection\n";
    if($connection == $ear) {
      my $newClient = $ear->accept();
      print STDERR "accept: $!" and next if !defined $newClient;
      my ($port, $addr) = sockaddr_in($newClient->peername());
      my $host = gethostbyaddr($addr, AF_INET);
      print STDOUT "Connection from $host:$port\n";
      $connections{"$host:$port"} = $newClient;
      $order{$newClient} = $count++;
      &DoAutoWhos();
    } else {
      &ProcessMessage($connection, &Receive($connection));
    }
  }
}
