/* $Id: diagnosticseed.h,v 1.15 2003/12/04 18:58:47 jhayes Exp $ */
/*
 * Copyright  2003 The Regents of the University of California. 
 * All Rights Reserved. 
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes, without
 * fee, and without a written agreement is hereby granted, provided that the
 * above copyright notice, this paragraph and the following three paragraphs
 * appear in all copies. 
 *
 * Permission to incorporate this software into commercial products may be
 * obtained by contacting
 * Technology Transfer Office 
 * 9500 Gilman Drive 
 * 411 University Center 
 * University of California 
 * La Jolla, CA 92093-0093
 * (858) 534-5815
 * invent@ucsd.edu
 *
 * This software program and documentation are copyrighted by The Regents of
 * the University of California. The software program and documentation are
 * supplied "as is", without any accompanying services from The Regents. The
 * Regents does not warrant that the operation of the program will be
 * uninterrupted or error-free. The end-user understands that the program was
 * developed for research purposes and is advised not to rely exclusively on
 * the program for any reason. 
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
 * EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED
 * HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO
 * OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS. 
 */


#ifndef DIAGNOSTICSEED_H
#define DIAGNOSTICSEED_H


/*
 * This package provides facilities for producing diagnostic messages.
 * Multiple types (e.g., severity levels) of messages may be directed to
 * independent destinations; the package predefines six types (DEBUG, ERROR,
 * FATAL, LOG, INFO, and WARNING).  By default, the package macros prepend
 * the file and line number where the message was generated to each message;
 * to turn off this behavior, #define ASDIAG_SUPPRESS_FILE_AND_LINE before
 * including this header.  Any message type may be suppressed by passing
 * ASDIAG_SUPPRESS as the second parameter to ASDIAG_DirectDiagnostics.
 */


#include <stdlib.h> /* exit */


#ifdef __cplusplus
extern "C" {
#endif


/* A value for suppressing messages of a particular type. */
#define ASDIAG_SUPPRESS (-1)


/* An internal representation of diagnostic directions. */
typedef struct ASDIAG__DiagnosticState *ASDIAG_DiagnosticState;


/*
 * Produces a message to the fd indicated for #messageType# messages in #ds#.
 * #message# is used as the fprintf format string; additional fprintf arguments
 * may be passed after it.
 */
void
ASDIAG_Diagnostic(ASDIAG_DiagnosticState ds,
                  int messageType,
                  const char *message,
                  ...);


/* Frees #ds# and any allocated memory it contains. */
void
ASDIAG_DiagnosticStateFree(ASDIAG_DiagnosticState ds);


/* Returns a new diagnostic state in which all messages are suppressed. */
ASDIAG_DiagnosticState
ASDIAG_DiagnosticStateNew(void);


/* Returns the fd to which #messageType# messages are directed in #ds#. */
int
ASDIAG_DiagnosticsDirection(ASDIAG_DiagnosticState ds,
                            int messageType);


/* Directs #messageType# diagnostic messages in #ds# to the fd #whereTo#. */
void
ASDIAG_DirectDiagnostics(ASDIAG_DiagnosticState ds,
                         int messageType,
                         int whereTo);


/*
 * Message types for debugging, error, fatal error, user information, log, and
 * warning messages, provided for convenience.
 */
#define ASDIAG_DEBUG_MESSAGE   0
#define ASDIAG_ERROR_MESSAGE   1
#define ASDIAG_FATAL_MESSAGE   2
#define ASDIAG_INFO_MESSAGE    3
#define ASDIAG_LOG_MESSAGE     4
#define ASDIAG_WARNING_MESSAGE 5


#ifdef ASDIAG_SUPPRESS_FILE_AND_LINE
#  define ASDIAG__MESSAGE(message) message
#else
#  define ASDIAG__MESSAGE(message) "%s %d: " message, __FILE__, __LINE__
#endif


/*
 * Issues the fatal message #message# with the file, line number, and any
 * specified parameters, then exits the program.
 */
#define ASDIAG_ABORT(ds,message) \
  do { \
    ASDIAG_Diagnostic(ds, ASDIAG_FATAL_MESSAGE,ASDIAG__MESSAGE(message)); \
    if(1) exit(1); \
  } while(0)
#define ASDIAG_ABORT1(ds,message,a) \
  do { \
    ASDIAG_Diagnostic(ds, ASDIAG_FATAL_MESSAGE, ASDIAG__MESSAGE(message), a); \
    if(1) exit(1); \
  } while(0)
#define ASDIAG_ABORT2(ds,message,a,b) \
  do { \
    ASDIAG_Diagnostic \
      (ds, ASDIAG_FATAL_MESSAGE,ASDIAG__MESSAGE(message), a, b); \
    if(1) exit(1); \
  } while(0)
#define ASDIAG_ABORT3(ds,message,a,b,c) \
  do { \
    ASDIAG_Diagnostic \
      (ds, ASDIAG_FATAL_MESSAGE,ASDIAG__MESSAGE(message), a, b, c); \
    if(1) exit(1); \
  } while(0)
#define ASDIAG_ABORT4(ds,message,a,b,c,d) \
  do { \
    ASDIAG_Diagnostic \
      (ds,ASDIAG_FATAL_MESSAGE,ASDIAG__MESSAGE(message),a,b,c,d); \
    if(1) exit(1); \
  } while(0)
#define ASDIAG_ABORT5(ds,message,a,b,c,d,e) \
  do { \
    ASDIAG_Diagnostic \
      (ds,ASDIAG_FATAL_MESSAGE,ASDIAG__MESSAGE(message),a,b,c,d,e);\
    if(1) exit(1); \
  } while(0)

#ifdef ASDIAG_DISPLAY_DEBUG_DIAGNOSTICS
#  define ASDIAG_DEBUG(ds,message) \
     ASDIAG_Diagnostic(ds,ASDIAG_DEBUG_MESSAGE,ASDIAG__MESSAGE(message)
#  define ASDIAG_DEBUG1(ds,message,a) \
     ASDIAG_Diagnostic(ds,ASDIAG_DEBUG_MESSAGE,ASDIAG__MESSAGE(message),a)
#  define ASDIAG_DEBUG2(message,a,b) \
     ASDIAG_Diagnostic(ds,ASDIAG_DEBUG_MESSAGE,ASDIAG__MESSAGE(message),a,b)
#  define ASDIAG_DEBUG3(ds,message,a,b,c) \
     ASDIAG_Diagnostic(ds,ASDIAG_DEBUG_MESSAGE,ASDIAG__MESSAGE(message),a,b,c)
#  define ASDIAG_DEBUG4(ds,message,a,b,c,d) \
     ASDIAG_Diagnostic(ds,ASDIAG_DEBUG_MESSAGE,ASDIAG__MESSAGE(message),a,b,c,d)
#  define ASDIAG_DEBUG5(ds,message,a,b,c,d,e) \
     ASDIAG_Diagnostic \
       (ds,ASDIAG_DEBUG_MESSAGE,ASDIAG__MESSAGE(message),a,b,c,d,e)
#else
#  define ASDIAG_DEBUG(ds,message)
#  define ASDIAG_DEBUG1(ds,message,a)
#  define ASDIAG_DEBUG2(ds,message,a,b)
#  define ASDIAG_DEBUG3(ds,message,a,b,c)
#  define ASDIAG_DEBUG4(ds,message,a,b,c,d)
#  define ASDIAG_DEBUG5(ds,message,a,b,c,d,e)
#endif

/* Issues the error message #message# with any specified parameters. */
#define ASDIAG_ERROR(ds,message) \
  ASDIAG_Diagnostic(ds,ASDIAG_ERROR_MESSAGE,ASDIAG__MESSAGE(message))
#define ASDIAG_ERROR1(ds,message,a) \
  ASDIAG_Diagnostic(ds,ASDIAG_ERROR_MESSAGE,ASDIAG__MESSAGE(message),a)
#define ASDIAG_ERROR2(ds,message,a,b) \
  ASDIAG_Diagnostic(ds,ASDIAG_ERROR_MESSAGE,ASDIAG__MESSAGE(message),a,b)
#define ASDIAG_ERROR3(ds,message,a,b,c) \
  ASDIAG_Diagnostic(ds,ASDIAG_ERROR_MESSAGE,ASDIAG__MESSAGE(message),a,b,c)
#define ASDIAG_ERROR4(ds,message,a,b,c,d) \
  ASDIAG_Diagnostic(ds,ASDIAG_ERROR_MESSAGE,ASDIAG__MESSAGE(message),a,b,c,d)
#define ASDIAG_ERROR5(ds,message,a,b,c,d,e) \
  ASDIAG_Diagnostic(ds,ASDIAG_ERROR_MESSAGE,ASDIAG__MESSAGE(message),a,b,c,d,e)

/*
 * Issues the error message #message# any specified parameters, then returns 0
 * from the current function.
 */
#define ASDIAG_FAIL(ds,message) ASDIAG_FAILRETURN(ds,message,0)
#define ASDIAG_FAIL1(ds,message,a) ASDIAG_FAILRETURN1(ds,message,a,0)
#define ASDIAG_FAIL2(ds,message,a,b) ASDIAG_FAILRETURN2(ds,message,a,b,0)
#define ASDIAG_FAIL3(ds,message,a,b,c) ASDIAG_FAILRETURN3(ds,message,a,b,c,0)
#define ASDIAG_FAIL4(ds,message,a,b,c,d) \
  ASDIAG_FAILRETURN4(ds,message,a,b,c,d,0)
#define ASDIAG_FAIL5(ds,message,a,b,c,d,e) \
  ASDIAG_FAILRETURN5(ds,message,a,b,c,d,e,0)

/*
 * Issues the error message #message# any specified parameters, then returns
 * #ret# from the current function.
 */
#define ASDIAG_FAILRETURN(ds,message,ret) \
  do { \
    ASDIAG_Diagnostic(ds,ASDIAG_ERROR_MESSAGE,ASDIAG__MESSAGE(message)); \
    if(1) return ret; \
  } while(0)
#define ASDIAG_FAILRETURN1(ds,message,a,ret) \
  do { \
    ASDIAG_Diagnostic(ds,ASDIAG_ERROR_MESSAGE,ASDIAG__MESSAGE(message),a); \
    if(1) return ret; \
  } while(0)
#define ASDIAG_FAILRETURN2(ds,message,a,b,ret) \
  do { \
    ASDIAG_Diagnostic(ds,ASDIAG_ERROR_MESSAGE,ASDIAG__MESSAGE(message),a,b); \
    if(1) return ret; \
  } while(0)
#define ASDIAG_FAILRETURN3(ds,message,a,b,c,ret) \
  do { \
    ASDIAG_Diagnostic(ds,ASDIAG_ERROR_MESSAGE,ASDIAG__MESSAGE(message),a,b,c); \
    if(1) return ret; \
  } while(0)
#define ASDIAG_FAILRETURN4(ds,message,a,b,c,d,ret) \
  do { \
    ASDIAG_Diagnostic \
      (ds,ASDIAG_ERROR_MESSAGE,ASDIAG__MESSAGE(message),a,b,c,d); \
    if(1) return ret; \
  } while(0)
#define ASDIAG_FAILRETURN5(ds,message,a,b,c,d,e,ret) \
  do { \
    ASDIAG_Diagnostic \
      (ds,ASDIAG_ERROR_MESSAGE,ASDIAG__MESSAGE(message),a,b,c,d,e);\
    if(1) return ret; \
  } while(0)

/* Issues the information message #message# with any specified parameters. */
#define ASDIAG_INFO(ds,message) \
  ASDIAG_Diagnostic(ds,ASDIAG_INFO_MESSAGE,ASDIAG__MESSAGE(message))
#define ASDIAG_INFO1(ds,message,a) \
  ASDIAG_Diagnostic(ds,ASDIAG_INFO_MESSAGE,ASDIAG__MESSAGE(message),a)
#define ASDIAG_INFO2(ds,message,a,b) \
  ASDIAG_Diagnostic(ds,ASDIAG_INFO_MESSAGE,ASDIAG__MESSAGE(message),a,b)
#define ASDIAG_INFO3(ds,message,a,b,c) \
  ASDIAG_Diagnostic(ds,ASDIAG_INFO_MESSAGE,ASDIAG__MESSAGE(message),a,b,c)
#define ASDIAG_INFO4(ds,message,a,b,c,d) \
  ASDIAG_Diagnostic(ds,ASDIAG_INFO_MESSAGE,ASDIAG__MESSAGE(message),a,b,c,d)
#define ASDIAG_INFO5(ds,message,a,b,c,d,e) \
  ASDIAG_Diagnostic(ds,ASDIAG_INFO_MESSAGE,ASDIAG__MESSAGE(message),a,b,c,d,e)

/* Issues the log message #message# with any specified parameters. */
#define ASDIAG_LOG(ds,message) \
  ASDIAG_Diagnostic(ds,ASDIAG_LOG_MESSAGE,ASDIAG__MESSAGE(message))
#define ASDIAG_LOG1(ds,message,a) \
  ASDIAG_Diagnostic(ds,ASDIAG_LOG_MESSAGE,ASDIAG__MESSAGE(message),a)
#define ASDIAG_LOG2(ds,message,a,b) \
  ASDIAG_Diagnostic(ds,ASDIAG_LOG_MESSAGE,ASDIAG__MESSAGE(message),a,b)
#define ASDIAG_LOG3(ds,message,a,b,c) \
  ASDIAG_Diagnostic(ds,ASDIAG_LOG_MESSAGE,ASDIAG__MESSAGE(message),a,b,c)
#define ASDIAG_LOG4(ds,message,a,b,c,d) \
  ASDIAG_Diagnostic(ds,ASDIAG_LOG_MESSAGE,ASDIAG__MESSAGE(message),a,b,c,d)
#define ASDIAG_LOG5(ds,message,a,b,c,d,e) \
  ASDIAG_Diagnostic(ds,ASDIAG_LOG_MESSAGE,ASDIAG__MESSAGE(message),a,b,c,d,e)

/* Issues the warning message #message# with any specified parameters. */
#define ASDIAG_WARN(ds,message) \
  ASDIAG_Diagnostic(ds,ASDIAG_WARNING_MESSAGE,ASDIAG__MESSAGE(message))
#define ASDIAG_WARN1(ds,message,a) \
  ASDIAG_Diagnostic(ds,ASDIAG_WARNING_MESSAGE,ASDIAG__MESSAGE(message),a)
#define ASDIAG_WARN2(ds,message,a,b) \
  ASDIAG_Diagnostic(ds,ASDIAG_WARNING_MESSAGE,ASDIAG__MESSAGE(message),a,b)
#define ASDIAG_WARN3(ds,message,a,b,c) \
  ASDIAG_Diagnostic(ds,ASDIAG_WARNING_MESSAGE,ASDIAG__MESSAGE(message),a,b,c)
#define ASDIAG_WARN4(ds,message,a,b,c,d) \
  ASDIAG_Diagnostic(ds,ASDIAG_WARNING_MESSAGE,ASDIAG__MESSAGE(message),a,b,c,d)
#define ASDIAG_WARN5(ds,message,a,b,c,d,e) \
  ASDIAG_Diagnostic \
    (ds,ASDIAG_WARNING_MESSAGE,ASDIAG__MESSAGE(message),a,b,c,d,e)


#ifdef ASDIAG_SHORT_NAMES

#define SUPPRESS ASDIAG_SUPPRESS
#define DiagnosticState ASDIAG_DiagnosticState
#define Diagnostic ASDIAG_Diagnostic
#define DiagnosticStateFree ASDIAG_DiagnosticStateFree
#define DiagnosticStateNew ASDIAG_DiagnosticStateNew
#define DiagnosticsDirection ASDIAG_DiagnosticsDirection
#define DirectDiagnostics ASDIAG_DirectDiagnostics

#define DEBUG_MESSAGE   ASDIAG_DEBUG_MESSAGE
#define ERROR_MESSAGE   ASDIAG_ERROR_MESSAGE
#define FATAL_MESSAGE   ASDIAG_FATAL_MESSAGE
#define INFO_MESSAGE    ASDIAG_INFO_MESSAGE
#define LOG_MESSAGE     ASDIAG_LOG_MESSAGE
#define WARNING_MESSAGE ASDIAG_WARNING_MESSAGE

#define ABORT  ASDIAG_ABORT
#define ABORT1 ASDIAG_ABORT1
#define ABORT2 ASDIAG_ABORT2
#define ABORT3 ASDIAG_ABORT3
#define ABORT4 ASDIAG_ABORT4
#define ABORT5 ASDIAG_ABORT5
#undef DEBUG
#define DEBUG  ASDIAG_DEBUG
#define DEBUG1 ASDIAG_DEBUG1
#define DEBUG2 ASDIAG_DEBUG2
#define DEBUG3 ASDIAG_DEBUG3
#define DEBUG4 ASDIAG_DEBUG4
#define DEBUG5 ASDIAG_DEBUG5
#undef ERROR
#define ERROR  ASDIAG_ERROR
#define ERROR1 ASDIAG_ERROR1
#define ERROR2 ASDIAG_ERROR2
#define ERROR3 ASDIAG_ERROR3
#define ERROR4 ASDIAG_ERROR4
#define ERROR5 ASDIAG_ERROR5
#define FAIL   ASDIAG_FAIL
#define FAIL1  ASDIAG_FAIL1
#define FAIL2  ASDIAG_FAIL2
#define FAIL3  ASDIAG_FAIL3
#define FAIL4  ASDIAG_FAIL4
#define FAIL5  ASDIAG_FAIL5
#define FAILRETURN  ASDIAG_FAILRETURN
#define FAILRETURN1 ASDIAG_FAILRETURN1
#define FAILRETURN2 ASDIAG_FAILRETURN2
#define FAILRETURN3 ASDIAG_FAILRETURN3
#define FAILRETURN4 ASDIAG_FAILRETURN4
#define FAILRETURN5 ASDIAG_FAILRETURN5
#define INFO   ASDIAG_INFO
#define INFO1  ASDIAG_INFO1
#define INFO2  ASDIAG_INFO2
#define INFO3  ASDIAG_INFO3
#define INFO4  ASDIAG_INFO4
#define INFO5  ASDIAG_INFO5
#define LOG    ASDIAG_LOG
#define LOG1   ASDIAG_LOG1
#define LOG2   ASDIAG_LOG2
#define LOG3   ASDIAG_LOG3
#define LOG4   ASDIAG_LOG4
#define LOG5   ASDIAG_LOG5
#define WARN   ASDIAG_WARN
#define WARN1  ASDIAG_WARN1
#define WARN2  ASDIAG_WARN2
#define WARN3  ASDIAG_WARN3
#define WARN4  ASDIAG_WARN4
#define WARN5  ASDIAG_WARN5

#endif


#ifdef __cplusplus
}
#endif


#endif
