/* $Id: compute.c,v 1.134 2003/11/25 23:03:02 jhayes Exp $ */
/*
 * Copyright  2002 The Regents of the University of California. 
 * All Rights Reserved. 
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes, without
 * fee, and without a written agreement is hereby granted, provided that the
 * above copyright notice, this paragraph and the following three paragraphs
 * appear in all copies. 
 *
 * Permission to incorporate this software into commercial products may be
 * obtained by contacting
 * Technology Transfer Office 
 * 9500 Gilman Drive 
 * 411 University Center 
 * University of California 
 * La Jolla, CA 92093-0093
 * (858) 534-5815
 * invent@ucsd.edu
 *
 * This software program and documentation are copyrighted by The Regents of
 * the University of California. The software program and documentation are
 * supplied "as is", without any accompanying services from The Regents. The
 * Regents does not warrant that the operation of the program will be
 * uninterrupted or error-free. The end-user understands that the program was
 * developed for research purposes and is advised not to rely exclusively on
 * the program for any reason. 
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
 * EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED
 * HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO
 * OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS. 
 */


#include "config.h"
#include <sys/stat.h> /* struct stat */
#include <ctype.h>    /* isalnum isspace */
#include <stdio.h>    /* sprintf */
#include <stdlib.h>   /* strtod strtol */
#include <string.h>   /* string functions */
#include <time.h>     /* time */
#include <unistd.h>   /* getpid stat */
#include <fileseed.h>
#define ASSTR_SHORT_NAMES
#include <strseed.h>
#define ASV_SHORT_NAMES
#include <vectorseed.h>
#include <elbase.h>
#include <rashgrp.h>
#include "global.h"
#include "misc.h"
#include "scheduler.h"
#include "storage.h"
#include "compute.h"


/* Info retained for each rash host group. */
typedef enum {
  GROUP_NEW, GROUP_PROXIED, GROUP_SUBMITTED, GROUP_RUNNING, GROUP_FAILED
} RashGroupStatus;
typedef struct _RashGroupInfo {
  Host batchHost;           /* Front-end host. */
  HostList rashHosts;       /* List of batch node hosts. */
  HostScheduler scheduler;  /* Submission information. */
  RASHGRP_GroupInfo group;  /* RASH host group. */
  RashGroupStatus status;   /* What we've done so far. */
  int stderrKey;            /* stderr redirection key for batch submission. */
} *RashGroupInfo;

/* Info retained for each running task. */
typedef struct _TaskInfo {
  Task task;
  ELBASE_Process pid; /* Local process id; -1 if simulated. */
  int terminated;     /* Done? ... */
  int exitValue;      /* ... exit value if so. */
  int canceled;       /* Did we cancel it? */
  int stderrKey;      /* stderr redirection key. */
} *TaskInfo;

/* Info about task variants used during expansion. */
typedef enum {
  EXPAND_TASK, EXPAND_VALUE, EXPAND_WORD
} VariantExpandTypes;
typedef struct _TaskVariant {
  VariantExpandTypes type;
  char *ref;
  char **values;
} *TaskVariant;


static RashGroupInfo *rashGroups;
static TaskInfo *pendingTasks;
static const ELBASE_Process SIMULATED_PROCESS_PID = NULL;
static const char *METHOD_IMAGES[ELBASE_COMPUTE_SERVICE_COUNT + 1] = {
  "local", "globus", "netsolve", "rash", "ssh", NULL
};
static const char *VARIANT_EXPAND_IMAGES[] = {
  "task", "value", "word"
};
static const char *SCHEDULER_IMAGES[ELBASE_SCHEDULER_SERVICE_COUNT + 1] = {
  "shell", "condor", "dqs", "loadleveler", "lsf", "pbs", "sge", NULL
};
static const char *TASK_STATUS_IMAGES[] = {
  "canceled", "done", "failed", "running", "new", NULL
};


#define isquote(c) ((c) == '"' || (c) == '\'')


/* Frees #tv# and all its fields. */
static void
Variant_Free(TaskVariant tv) {
  if(tv->ref != NULL)
    FREE(tv->ref);
  if(tv->values != NULL)
    StrArrayFree(tv->values);
  FREE(tv);
}


/* Returns a newly-allocated TaskVariant containing the specified values. */
static TaskVariant
Variant_New(VariantExpandTypes type,
            char *ref,
            char **values) {
  TaskVariant result = (TaskVariant)MALLOC(sizeof(struct _TaskVariant));
  result->type = type;
  result->ref = ref;
  result->values = values;
  return result;
}


/*
 * Returns an allocated string of XML containing a copy of the XML string
 * #template# with the values of each element of #variants#, starting with the
 * #first# one, appropriately expanded with its value(s).
 */
static char *
ExpandTaskXml(const char *template,
              TaskVariant *variants,
              unsigned first) {

  char *c;
  char *context;
  const char *contextBegin;
  const char *contextEnd;
  char *expansion;
  char **expansions;
  unsigned i;
  char *rangeCurr;
  char *rangeLast;
  unsigned rangeStep;
  char *ref;
  char *refBegin;
  char *result = STRDUP(template);
  const char *sep;
  VariantExpandTypes type;

  if(first >= Size(variants))
    return result;

  ref = variants[first]->ref;
  type = variants[first]->type;
  sep = type == EXPAND_TASK ? "\n" : " ";

  /* For each context containing a reference to this variant ... */
  while((refBegin = strstr(result, ref)) != NULL) {

    if(type == EXPAND_TASK) {
      contextBegin = result;
      contextEnd = result + strlen(result) - 1;
    }
    else {
      contextBegin = contextEnd = refBegin;
      while(!isquote((int)*contextBegin) &&
            (type == EXPAND_VALUE || !isspace((int)*contextBegin)))
        contextBegin--;
      contextBegin++;
      while(!isquote((int)*contextEnd) &&
            (type == EXPAND_VALUE || !isspace((int)*contextEnd)))
        contextEnd++;
      contextEnd--;
    }

    context = StrNDup(contextBegin, contextEnd - contextBegin + 1);
    expansions = VectorNew(char *);

    /* For each value range of the variant ... */
    for(i = 0; variants[first]->values[i] != NULL; i++) {
      rangeCurr = STRDUP(variants[first]->values[i]);
      c = strrchr(rangeCurr, ':');
      rangeLast = STRDUP(c == NULL ? rangeCurr : c + 1);
      if((c = strchr(rangeCurr, ':')) != NULL)
        *c = '\0';
      rangeStep = c == NULL || strchr(c + 1, ':') == NULL ?
                  1 : strtol(c + 1, NULL, 10);
      StrIncr(&rangeLast, rangeStep);
      /* For each value in the range ... */
      for( ; strcmp(rangeCurr,rangeLast) != 0; StrIncr(&rangeCurr,rangeStep)) {
        expansion = STRDUP(context);
        StrReplaceAll(&expansion, ref, rangeCurr);
        Append(expansions, expansion);
      }
      FREE(rangeCurr);
      FREE(rangeLast);
    }

    expansion = StrArrayNJoin((const char **)expansions, Size(expansions), sep);
    StrReplace(&result, contextBegin - result, strlen(context), expansion);
    FREE(context);
    FREE(expansion);
    for(i = 0; i < Size(expansions); i++)
      FREE(expansions[i]);
    VectorFree(expansions);

  }
  c = result;
  result = ExpandTaskXml(c, variants, first + 1);
  FREE(c);
  return result;

}


/*
 * Returns NULL if #f# is NULL; otherwise, the path to #f# on #d# in an
 * allocated string.  If the path to #f# is relative, prepends #d#'s datadir
 * if #makeAbsolute# is true; otherwise, returns the relative path.
 */
static char *
FilePath(Disk d,
         File f,
         int makeAbsolute) {
  FileCopy fc;
  const char *path;
  if(f == NULL)
    return NULL;
  path = (fc = FileCopyList_Get(f->copies, d->id)) == NULL ? f->name : fc->path;
  return makeAbsolute ? AbsolutePath(path, d->datadir) : STRDUP(path);
}


/* Returns the byte size of #f# on #d#; -1 if #f# is not on #d#. */
static long
FileSizeOnDisk(Disk d,
               File f) {

  const char *args[3] = {"-l", NULL, NULL};
  char *c;
  unsigned col;
  char *dirPath;
  FILE *dir = NULL;
  char *fileName;
  char *filePath = FilePath(d, f, 1);
  Host h;
  ASH_HashIterator it;
  char lsLine[256 + 1];
  char *lsPath;
  ELBASE_Process pid;
  char pidImage[20 + 1];
  long result;
  struct stat st;

  ASFILE_MinimizeFilePath(filePath);
  if(d->method == ELBASE_CP) {
    result = stat(filePath, &st) == 0 ? st.st_size : -1;
    FREE(filePath);
    return result;
  }

  /* Remote #f#.  Construct a file name to hold an ls of its directory. */
  fileName = strrchr(filePath, '/') + 1;
  *(fileName - 1) = '\0';
  dirPath = filePath;
  sprintf(pidImage, "%d", (int)getpid());
  lsPath = StrAppend
    (NULL, "/tmp/", d->server, ".", dirPath, ".", pidImage, ".ls", NULL);
  for(c = lsPath + 5; *c != '\0'; c++)
    if(!isalnum((int)*c))
      *c = '.';

  if((dir = fopen(lsPath, "r")) == NULL) {
    /* Generate the ls remotely, then copy it to the local disk. */
    args[1] = dirPath;
    ASH_ForEachMapping(global.hosts, it) {
      h = (Host)ASH_IteratorValue(it);
      if(h->disk == d)
        break;
    }
    if(it.valid &&
       ELBASE_Spawn(h->method, h->server, "/bin/ls", d->datadir, args, NULL,
                    lsPath, NULL, NULL, &pid) &&
       ELBASE_Poll(pid, 1, NULL) &&
       ELBASE_CopyFile(d->method,d->server,lsPath,ELBASE_CP,"",lsPath,&pid) &&
       ELBASE_Poll(pid, 1, NULL))
      dir = fopen(lsPath, "r");
  }

  /* Find the line for #f# in the ls output; get size from the 5th column. */
  result = -1;
  if(dir != NULL) {
    while(fgets(lsLine, sizeof(lsLine), dir) != NULL) {
      lsLine[strlen(lsLine) - 1] = '\0';
      if((c = strrchr(lsLine, ' ')) == NULL || strcmp(c + 1, fileName) != 0)
        continue;
      for(c = lsLine; isspace((int)*c); c++)
        ; /* empty */
      for(col = 1;  col < 5 && *c != '\0'; col++) {
        while(!isspace((int)*c) && *c != '\0')
          c++;
        while(isspace((int)*c))
          c++;
      }
      result = strtol(c, NULL, 10);
      break;
    }
    fclose(dir);
  }

  FREE(filePath);
  FREE(lsPath);
  return result;

}


/*
 * Finds on the global files list and returns the file with id #id#, creating
 * it if necessary.  #outputOf# is a task that produces the file as output; it
 * is NULL for input files.
 */
static File
GetFile(const char *id,
        Task outputOf) {
  File file = FileList_Get(global.files, (char *)id);
  if(file == NULL) {
    file = File_New(STRDUP(id));
    file->download = outputOf != NULL;
    file->transferable = 1;
    FileList_Add(global.files, file);
  }
  if(outputOf != NULL)
    file->outputOf = outputOf;
  else if(file->outputOf == NULL) {
    /* (Possibly staged) input file; test for copies as appropriate. */
    Disk d;
    ASH_HashIterator it;
    long size;
    FindLocalDisk(); /* Force inclusion of local disk if undefined by user. */
    ASH_ForEachMapping(global.disks, it) {
      d = (Disk)ASH_IteratorValue(it);
      if(FileCopyList_Get(file->copies, d->id) == NULL &&
         (global.findInput || d->method == ELBASE_CP) &&
         (size = FileSizeOnDisk(d, file)) >= 0) {
          NewFileCopy(d, file);
          file->size = size;
      }
    }
  }
  return file;
}


/*
 * Stores in the vector #toWhere# the files included in the space-delimited
 * list of file ids #spec#.  #outputOf# is a task that produces the files as
 * output; it is NULL for input files.
 */
static void
GetFileList(const char *spec,
            Task outputOf,
            File **toWhere) {
  unsigned i;
  char **words = StrSplit(spec, NULL);
  for(i = 0; words[i] != NULL; i++)
    Append(*toWhere, GetFile(words[i], outputOf));
  StrArrayFree(words);
}


/* Returns a host containing the information under #node#, NULL on error. */
static Host
ProcessHostXml(ASXML_XmlNode node) {

  char *allAttrs;
  ASXML_XmlNode child;
  Disk disk;
  const char *diskId;
  const char *field;
  ASXML_XmlNode grandChild;
  Host host;
  unsigned long i;
  const char *id;
  const char *optionName;
  const char *optionValue;
  HostScheduler scheduler;
  char *wd;

  if(*(id = ASXML_ElementGetAttribute(node, "id")) == '\0')
    FAIL("'host' tag requires an 'id' attribute\n");
  if(HostList_Get(global.hosts, (char *)id) != NULL)
    FAIL1("Duplicate definition for host '%s'\n", id);
  if(*(diskId = ASXML_ElementGetAttribute(node, "disk")) == '\0')
    disk = FindLocalDisk();
  else if((disk = DiskList_Get(global.disks, (char *)diskId)) == NULL)
    FAIL1("Unknown disk '%s'\n", diskId);
  wd = *(field = ASXML_ElementGetAttribute(node, "wd")) == '\0' ?
       STRDUP(disk->datadir) : AbsolutePath(field, disk->datadir);

  host = Host_New(STRDUP(id), disk, wd, ELBASE_FORK, STRDUP(""));

  host->cpus = *(field = ASXML_ElementGetAttribute(node, "cpus")) == '\0' ?
               1 : strtol(field, NULL, 10);
  if(*(field = ASXML_ElementGetAttribute(node, "dnsname")) != '\0')
    host->dnsName = STRDUP(field);
  if(*(field = ASXML_ElementGetAttribute(node, "memory")) != '\0') {
    host->memory = strtod(field, (char **)&field);
    if(*field == 'K' || *field == 'k')
      host->memory *= 1000;
    else if(*field == 'M' || *field == 'm')
      host->memory *= 1000000;
    else if(*field == 'G' || *field == 'g')
      host->memory *= 1000000000;
  }

  for(child = ASXML_NodeFirstChildByType(node, ASXML_SHOW_ELEMENT);
      child != NULL;
      child = ASXML_NodeNextSiblingByType(child, ASXML_SHOW_ELEMENT)) {

    if(StrArrayFind(METHOD_IMAGES, ASXML_NodeName(child), &i)) {
      host->method = (ELBASE_ComputeServiceTypes)i;
      FREE(host->server);
      host->server = STRDUP
        (host->method == ELBASE_FORK ? "" :
         *(field = ASXML_ElementGetAttribute(child, "server")) == '\0' ?
         id : field);
      continue;
    }
    if(!StrArrayFind(SCHEDULER_IMAGES, ASXML_NodeName(child), &i)) {
      Host_Free(host);
      FAIL1("Unknown host subtag '%s'\n", ASXML_NodeName(child));
    }

    scheduler = (HostScheduler)MALLOC(sizeof(struct _HostScheduler));
    scheduler->method = (ELBASE_SchedulerServiceTypes)i;
    allAttrs = STRDUP("");
    if(*(field = ASXML_ElementGetAttribute(child, "account")) != '\0')
      allAttrs = StrAppend(allAttrs, ELBASE_ACCOUNT, "=", field, "\t", NULL);
    if(*(field = ASXML_ElementGetAttribute(child, "memory")) != '\0')
      allAttrs = StrAppend(allAttrs, ELBASE_MEMORY, "=", field, "\t", NULL);
    if(*(field = ASXML_ElementGetAttribute(child, "nodes")) == '\0')
      field = "1";
    allAttrs = StrAppend(allAttrs, ELBASE_NODES, "=", field, "\t", NULL);
    scheduler->agentCount = strtol(field, NULL, 10);
    if(*(field = ASXML_ElementGetAttribute(child, "nodetype")) != '\0')
      allAttrs = StrAppend(allAttrs, ELBASE_NODETYPE, "=", field, "\t", NULL);
    if(*(field = ASXML_ElementGetAttribute(child, "queue")) != '\0')
      allAttrs = StrAppend(allAttrs, ELBASE_QUEUE, "=", field, "\t", NULL);
    if(*(field = ASXML_ElementGetAttribute(child, "stderr")) != '\0')
      allAttrs = StrAppend(allAttrs, ELBASE_STDERR, "=", field, "\t", NULL);
    if(*(field = ASXML_ElementGetAttribute(child, "stdin")) != '\0')
      allAttrs = StrAppend(allAttrs, ELBASE_STDIN, "=", field, "\t", NULL);
    if(*(field = ASXML_ElementGetAttribute(child, "stdout")) != '\0')
      allAttrs = StrAppend(allAttrs, ELBASE_STDOUT, "=", field, "\t", NULL);
    if(*(field = ASXML_ElementGetAttribute(child, "time")) == '\0')
      field = "60";
    allAttrs = StrAppend(allAttrs, ELBASE_TIME, "=", field, "\t", NULL);
    if(*(field = ASXML_ElementGetAttribute(child, "rashport")) == '\0')
      field = "0";
    scheduler->port = strtol(field, NULL, 10);
    for(grandChild = ASXML_NodeFirstChildByType(child, ASXML_SHOW_ELEMENT);
        grandChild != NULL;
        grandChild=ASXML_NodeNextSiblingByType(grandChild,ASXML_SHOW_ELEMENT)) {
      if(strcmp(ASXML_NodeName(grandChild), "option") != 0) {
        WARN1("Unknown scheduler subtag %s\n", ASXML_NodeName(grandChild));
        continue;
      }
      if(*(optionName = ASXML_ElementGetAttribute(grandChild, "name"))=='\0' ||
         *(optionValue = ASXML_ElementGetAttribute(grandChild,"value"))=='\0') {
        WARN("Missing attribute in option tag\n");
        continue;
      }
      allAttrs = StrAppend(allAttrs, optionName, "=", optionValue, "\t", NULL);
    }
    scheduler->attrs = StrSplit(allAttrs, "\t");
    FREE(allAttrs);
    scheduler->next = host->schedulers;
    host->schedulers = scheduler;

  }

  return host;

}


/* */
static int
ProcessTaskXml(ASXML_XmlNode node,
               const char *defaultGroups) {
  File f;
  const char *field;
  int i;
  Task t;
  if(*(field = ASXML_ElementGetAttribute(node, "executable")) == '\0')
    FAIL("'task' tag requires an 'executable' attribute\n");
  t = Task_New();
  t->exec = GetFile(field, NULL);
  if(*(field = ASXML_ElementGetAttribute(node, "arguments")) != '\0')
    t->arguments = StrSplit(field, NULL);
  if(*(field = ASXML_ElementGetAttribute(node, "cost")) != '\0')
    t->meta.cost = strtol(field, NULL, 10);
  if(*(field = ASXML_ElementGetAttribute(node, "groups")) == '\0')
    field = defaultGroups;
  if(*field != '\0') {
    char **groups = StrSplit(field, NULL);
    for(i = 0; groups[i] != NULL; i++)
      TaskGroupAdd(groups[i], t);
    StrArrayFree(groups);
  }
  if(*(field = ASXML_ElementGetAttribute(node, "host")) != '\0' &&
     (t->host = HostList_Get(global.hosts, (char *)field)) == NULL)
    WARN1("Unknown host %s\n", field);
  if(*(field = ASXML_ElementGetAttribute(node, "id")) != '\0') {
    if(ASH_GetKey(global.userIds, (ASH_KeyType)field, NULL) != NULL)
      WARN1("Duplication definition for task '%s'\n", field);
    else
      ASH_PutMapping(global.userIds, t, STRDUP(field));
  }
  if(*(field = ASXML_ElementGetAttribute(node, "input")) != '\0')
    GetFileList(field, NULL, &t->input);
  if(*(field = ASXML_ElementGetAttribute(node, "memory")) != '\0') {
    t->meta.memory = strtod(field, (char **)&field);
    if(*field == 'K' || *field == 'k')
      t->meta.memory *= 1000;
    else if(*field == 'M' || *field == 'm')
      t->meta.memory *= 1000000;
    else if(*field == 'G' || *field == 'g')
      t->meta.memory *= 1000000000;
  }
  if(*(field = ASXML_ElementGetAttribute(node, "output")) != '\0')
    GetFileList(field, t, &t->output);
  if(*(field = ASXML_ElementGetAttribute(node, "priority")) != '\0')
    t->priority = strtol(field, NULL, 10);
  if(*(field = ASXML_ElementGetAttribute(node, "stderr")) != '\0') {
    f = GetFile(field, t);
    if(global.downloadstd && (*f->name!='/' || strcmp(f->name,"/dev") != 0)) {
      for(i = 0; i < Size(t->output) && t->output[i] != f; i++)
        ; /* empty */
      if(i == Size(t->output))
        Append(t->output, f);
    }
    ASH_PutMapping(global.stderrs, t, f);
  }
  if(*(field = ASXML_ElementGetAttribute(node, "stdin")) != '\0') {
    f = GetFile(field, NULL);
    for(i = 0; i < Size(t->output) && t->output[i] != f; i++)
      ; /* empty */
    if(i == Size(t->output))
      Append(t->input, f);
    ASH_PutMapping(global.stdins, t, f);
  }
  if(*(field = ASXML_ElementGetAttribute(node, "stdout")) != '\0') {
    f = GetFile(field, t);
    if(global.downloadstd && (*f->name!='/' || strcmp(f->name,"/dev") != 0)) {
      for(i = 0; i < Size(t->output) && t->output[i] != f; i++)
        ; /* empty */
      if(i == Size(t->output))
        Append(t->output, f);
    }
    ASH_PutMapping(global.stdouts, t, f);
  }
  if(*(field = ASXML_ElementGetAttribute(node, "wd")) != '\0')
    ASH_PutMapping(global.wds, t, STRDUP(field));
  TaskList_Add(global.tasks, t);
  TaskGroupAdd("new", t);
  sched_new_task(t);
  return 1;
}


/* Returns a task variant containing the info under #node#, NULL on error. */
static TaskVariant
ProcessVariantXml(ASXML_XmlNode node) {
  const char *expand;
  unsigned long i;
  const char *name;
  VariantExpandTypes type;
  const char *values;
  if((name = ASXML_ElementGetAttribute(node, "name")) == NULL)
    FAIL("'variant' tag requires a 'name' attribute\n");
  else if((values = ASXML_ElementGetAttribute(node, "values")) == NULL)
    FAIL("'variant' tag requires a 'values' attribute\n");
  if(*(expand = ASXML_ElementGetAttribute(node, "expand")) == '\0')
    type = EXPAND_TASK;
  else if(StrArrayFind(VARIANT_EXPAND_IMAGES, expand, &i))
    type = (VariantExpandTypes)i;
  else
    FAIL1("Unknown value %s for 'expand' attribute\n", expand);
  return Variant_New
    (type, StrAppend(NULL, "${", name, "}", NULL), StrSplit(values, NULL));
}


/*
 * Process the variant task info headed by #node#, added tasks from the
 * expanded XML to the global tables.  Returns 1 if successful, else 0.
 */
static int
ProcessVariantTaskXml(ASXML_XmlNode node,
                      const char *groups) {

  ASXML_XmlNode child;
  char *expandedXml;
  unsigned i;
  ASXML_XmlNode invariant;
  int result = 1;
  char *tasksXml;
  char *template;
  TaskVariant variant;
  TaskVariant *variants = VectorNew(TaskVariant);

  for(child = ASXML_NodeFirstChildByType(node, ASXML_SHOW_ELEMENT);
      child != NULL && result;
      child = ASXML_NodeNextSiblingByType(child, ASXML_SHOW_ELEMENT)) {
    if(strcmp(ASXML_NodeName(child), "variant") != 0) {
      ERROR1("Unknown task subtag '%s'\n", ASXML_NodeName(child));
      result = 0;
    }
    else if((variant = ProcessVariantXml(child)) == NULL)
      result = 0;
    else
      Append(variants, variant);
  }

  if(result) {
    /* Get task XML w/out variants. */
    invariant = ASXML_NodeCloneNode(node, 0);
    template = ASXML_XmlToString(invariant, NULL);
    ASXML_NodeFree(invariant);
    /* Expand and add enclosing <task></task> to make it legal XML. */
    expandedXml = ExpandTaskXml(template, variants, 0);
    FREE(template);
    tasksXml = StrAppend
      (NULL, "<tasks groups='", groups, "'>", expandedXml, "</tasks>", NULL);
    FREE(expandedXml);
    /* Parse and process. */
    if((node = ASXML_XmlFromString(tasksXml,NULL,1,1,NULL,NULL,NULL)) == NULL)
      result = 0;
    else {
      result = compute_process_xml(ASXML_DocumentElement(node));
      ASXML_NodeFree(node);
    }
    FREE(tasksXml);
  }
  for(i = 0; i < Size(variants); i++)
    Variant_Free(variants[i]);
  VectorFree(variants);
  return result;

}


/* Releases any batch resources associated with #group#. */
static void
ReleaseBatchResources(RashGroupInfo group) {
  ASH_HashIterator it;
  if(group->status != GROUP_RUNNING)
    return;
  ASH_ForEachMapping(group->rashHosts, it)
    HostDisabled((Host)ASH_IteratorValue(it));
  RASHGRP_StopAgents(group->group);
  group->status = GROUP_PROXIED;
}


int
compute_cancel_task(Task t) {
  unsigned i;
  for(i = 0; i < Size(pendingTasks) && pendingTasks[i]->task != t; i++)
    ; /* empty */
  if(i == Size(pendingTasks))
    FAIL1("compute_cancel_task: Cannot find task %ld\n", t->seq);
  if(pendingTasks[i]->pid != SIMULATED_PROCESS_PID &&
     !ELBASE_Kill(pendingTasks[i]->pid))
    WARN2("compute_cancel_task: Cannot kill pid %d of task %ld\n",
          pendingTasks[i]->pid, t->seq);
  pendingTasks[i]->canceled = 1;
  return 1;
}


int
compute_do_task(Task t) {

  char *execPath;
  TaskInfo info;
  ELBASE_SchedulerServiceTypes scheduler;
  char **schedulerAttrs;
  char *stderrPath;
  char *stdinPath;
  char *stdoutPath;
  char *wd;

  info = (TaskInfo)MALLOC(sizeof(struct _TaskInfo));
  info->task = t;
  info->terminated = 0;
  info->canceled = 0;
  wd = ASH_GetValue(global.wds, t, NULL);
  wd = wd == NULL ? t->host->wd : AbsolutePath(wd, t->host->wd);
  execPath = FilePath(t->host->disk, t->exec, 0);
  stdinPath = FilePath(t->host->disk, ASH_GetValue(global.stdins, t, NULL), 1);
  stdoutPath = FilePath(t->host->disk, ASH_GetValue(global.stdouts,t,NULL), 1);
  stderrPath = FilePath(t->host->disk, ASH_GetValue(global.stderrs,t,NULL), 1);
  scheduler =
    t->host->schedulers == NULL ? ELBASE_SHELL : t->host->schedulers->method;
  schedulerAttrs =
    t->host->schedulers == NULL ? NULL : t->host->schedulers->attrs;
  info->stderrKey = StderrCaptureBegin();
  if(global.simHost >= 0)
    info->pid = SIMULATED_PROCESS_PID;
  else if(!ELBASE_Submit(t->host->method, t->host->server, scheduler,
                         (const char **)schedulerAttrs,
                         execPath, wd, (const char **)t->arguments,
                         stdinPath, stdoutPath, stderrPath, NULL, &info->pid)) {
    info->terminated = 1;
    info->exitValue = 1;
  }
  StderrCaptureEnd(info->stderrKey);
  TaskStarted(t, t->host);
  if(wd != t->host->wd)
    FREE(wd);
  FREE(execPath);
  if(stdinPath != NULL)
    FREE(stdinPath);
  if(stdoutPath != NULL)
    FREE(stdoutPath);
  if(stderrPath != NULL)
    FREE(stderrPath);

  Append(pendingTasks, info);
  return 1;

}

 
int
compute_init(void) {
  pendingTasks = VectorNew(TaskInfo);
  rashGroups = VectorNew(RashGroupInfo);
  return 1;
}


int
compute_poll(void) {

  char *c;
  int exitCode;
  RashGroupInfo group;
  Host h;
  int i;
  int j;
  int hostDead;
  TaskInfo info;
  ELBASE_Process pid;
  int stderrKey;

  /* Test any pending rash requests. */
  for(i = 0; i < Size(rashGroups); i++) {
    group = rashGroups[i];
    if(group->status == GROUP_NEW) {
      group->status = GROUP_PROXIED;
      stderrKey = StderrCaptureBegin();
      if(global.simHost < 0 && !RASHGRP_StartProxy(group->group))
        group->status = GROUP_FAILED;
      StderrCaptureEnd(stderrKey);
      c = StderrCaptureRetrieve(stderrKey);
      if(group->status == GROUP_FAILED) {
        WARN1("Unable to start proxy on %s\n", group->batchHost->id);
        if(c != NULL)
          WARN1("Proxy stderr follows\n%s\n", c);
      }
      if(c != NULL)
        FREE(c);
    }
    else if(TaskGroupMembership("new") == 0 &&
            TaskGroupMembership("running") == 0) {
      /* Avoid useless allocation burning. */
      if(group->status == GROUP_SUBMITTED || group->status == GROUP_RUNNING)
        ReleaseBatchResources(group);
    }
    else if(group->status == GROUP_PROXIED) {
      group->status = GROUP_SUBMITTED;
      group->stderrKey = StderrCaptureBegin();
      if(global.simHost < 0)
        RASHGRP_SubmitAgents
          (group->group, group->scheduler->method,
           (const char **)group->scheduler->attrs);
      StderrCaptureEnd(group->stderrKey);
    }
    else if(group->status == GROUP_SUBMITTED) {
      if(global.simHost >= 0)
        exitCode = 0;
      else if(RASHGRP_AgentsPending(group->group, &exitCode))
        continue;
      group->status = GROUP_RUNNING;
      c = StderrCaptureRetrieve(group->stderrKey);
      if(exitCode != 0) {
        group->status = GROUP_FAILED;
        WARN2("Unable to submit job to %s (exit code %d\n) \n",
              group->batchHost->id, exitCode);
        if(c != NULL)
          WARN1("Submission stderr follows\n%s\n", c);
      }
      if(c != NULL)
        FREE(c);
      if(group->status == GROUP_RUNNING) {
        char **agents;
        unsigned agentSubid;
        Host h;
        char *id;
        char image[20];
        if(global.simHost < 0)
          agents = RASHGRP_GetAgentContacts(group->group);
        else {
          /* Simulate agent contact ids. */
          j = group->scheduler->agentCount;
          agents = MALLOC((j + 1) * sizeof(char *));
          agents[j] = NULL;
          while(--j >= 0) {
            sprintf(image, "%d", j + 1);
            agents[j] = StrAppend(NULL, image, "@", group->batchHost->id, NULL);
          }
        }
        for(j = 0; agents[j] != NULL; j++) {
          for(agentSubid = 0; ; agentSubid++) {
            sprintf(image, "%d", agentSubid);
            id = StrAppend(NULL, group->batchHost->id, "-", image, NULL);
            if(HostList_Get(global.hosts, id) == NULL)
              break;
            FREE(id);
          }
          h = Host_New(id, group->batchHost->disk,
                       STRDUP(group->batchHost->wd), ELBASE_RASH, NULL);
          h->cpus = group->batchHost->cpus;
          h->memory = group->batchHost->memory;
          HostList_Add(global.hosts, h);
          HostList_Add(group->rashHosts, h);
          h->server = agents[j];
          HostEnabled(h);
        }
        FREE(agents);
      }
    }
  }

  for(i = Size(pendingTasks) - 1; i >= 0; i--) {
    info = pendingTasks[i];
    if(info->pid == SIMULATED_PROCESS_PID) {
      long runTime = !global.simRandom ? global.simHost :
                     (long)(global.simHost * ((double)random() / RAND_MAX));
      if(time(NULL) >= info->task->meta.timeStarted + runTime) {
        info->terminated = 1;
        info->exitValue = 0;
      }
    }
    else if(!info->terminated)
      info->terminated = ELBASE_Poll(info->pid, 0, &info->exitValue);
    if(!info->terminated)
      continue;
    c = StderrCaptureRetrieve(info->stderrKey);
    if(info->canceled)
      ; /* empty */
    else if(info->exitValue == 0)
      TaskDone(info->task, 0);
    else {
      TaskDone(info->task, info->exitValue);
      if(c != NULL)
        WARN1("Task stderr follows\n%s\n", c);
      /*
       * The task failed for one of three reasons: the host died, the task spec
       * is bad, or there's a task/host mismatch.  Differentiating the last two
       * is hard, but we can at least check for host failure.
       */
       h = info->task->host;
       hostDead = 1;
       if(!ELBASE_Spawn(h->method, h->server, "/bin/ls", NULL, NULL, NULL,
                        "/dev/null", NULL, NULL, &pid))
         WARN1("Unable to spawn /bin/ls to %s\n", h->server);
       else if(!ELBASE_Poll(pid, 1, &exitCode))
         WARN1("Unable to poll /bin/ls on %s\n", h->server);
       else if(exitCode != 0)
         WARN2("/bin/ls failed w/exit code %d on %s\n", exitCode, h->server);
       else
         hostDead = 0;
      if(hostDead) {
        HostDisabled(h);
        TaskReset(info->task);
        /*
         * If it's a rash host, it probably died because the allocated time
         * expired.  Resubmit.
         */
        if(h->method == ELBASE_RASH) {
          for(j = 0;
              j < Size(rashGroups) &&
              HostList_Get(rashGroups[j]->rashHosts, h->id) == NULL;
              j++)
             ; /* empty */
          if(j < Size(rashGroups))
            ReleaseBatchResources(rashGroups[j]);
        }
      }
    }
    if(c != NULL)
      FREE(c);
    pendingTasks[i] = pendingTasks[Size(pendingTasks) - 1];
    ShrinkSize(pendingTasks, Size(pendingTasks) - 1);
    FREE(info);
  }

  return 1;

}


int
compute_process_xml(ASXML_XmlNode node) {

  ASXML_XmlNode child;
  ASXML_XmlNode grand;
  const char *groups = ASXML_ElementGetAttribute(node, "groups");
  Host host;
  int result = 1;
  HostScheduler s;

  for(child = ASXML_NodeFirstChildByType(node, ASXML_SHOW_ELEMENT);
      child != NULL;
      child = ASXML_NodeNextSiblingByType(child, ASXML_SHOW_ELEMENT)) {
    if(strcmp(ASXML_NodeName(child), "host") == 0) {
      if((host = ProcessHostXml(child)) == NULL)
        result = 0;
      else {
        HostList_Add(global.hosts, host);
        HostEnabled(host);
        for(s = host->schedulers; s != NULL; s = s->next) {
          RASHGRP_GroupInfo group = NULL;
          RashGroupInfo info;
          if(s->method == ELBASE_SHELL || s->method == ELBASE_CONDOR)
            continue;
          HostDisabled(host);
          if(global.simHost < 0) {
            char *c;
            int stderrKey = StderrCaptureBegin();
            if(s->port == 0)
              s->port = getpid() % 4000 + 4000 + Size(rashGroups);
            group = RASHGRP_GroupNew
              (host->disk->method, host->disk->server, host->method,
               host->server, s->port, host->wd);
            StderrCaptureEnd(stderrKey);
            c = StderrCaptureRetrieve(stderrKey);
            if(group == NULL) {
              WARN1("Unable to acquire nodes on %s\n", host->id);
              if(c != NULL)
                WARN1("Acquire stderr follows\n%s\n", c);
              result = 0;
            }
            if(c != NULL)
              FREE(c);
          }
          if(group != NULL || global.simHost >= 0) {
            info = MALLOC(sizeof(struct _RashGroupInfo));
            info->batchHost = host;
            info->rashHosts = HostList_New(SMALL_HASH);
            info->scheduler = s;
            info->group = group;
            info->status = GROUP_NEW;
            Append(rashGroups, info);
          }
        }
      }
    }
    else if(strcmp(ASXML_NodeName(child), "task") != 0) {
      ERROR1("Unknown compute subtag '%s'\n", ASXML_NodeName(child));
      result = 0;
    }
    else if((grand = ASXML_NodeFirstChildByType(child, ASXML_SHOW_ELEMENT)) !=
            NULL && strcmp(ASXML_NodeName(grand), "variant") == 0)
      result = result && ProcessVariantTaskXml(child, groups);
    else
      result = result && ProcessTaskXml(child, groups);
  }

  return result;

}


char *
compute_host_to_xml(Host h) {

  char *attr;
  char buffer[24 + 1];
  unsigned i;
  char *result;
  HostScheduler s;

  result = StrAppend
    (NULL, "<host id='", h->id, "' disk='", h->disk->id, "' wd='", h->wd, "'",
     NULL);
  if(h->dnsName != NULL)
    result = StrAppend(result, " dnsname='", h->dnsName, "'", NULL);
  if(h->cpus != 1) {
    sprintf(buffer, " cpus='%d'", h->cpus);
    result = StrAppend(result, buffer, NULL);
  }
  if(h->memory != 0) {
    sprintf(buffer, " memory='%.0f'", h->memory);
    result = StrAppend(result, buffer, NULL);
  }
  if(h->meta.timeDisabled != 0)
    result = StrAppend(result, " status='disabled'", NULL);
  result = StrAppend(result, "><", METHOD_IMAGES[h->method], NULL);
  if(*h->server != '\0' && strcmp(h->server, h->id) != 0)
    result = StrAppend(result, " server='", h->server, "'", NULL);
  result = StrAppend(result, "/>", NULL);
  for(s = h->schedulers; s != NULL; s = s->next) {
    if(s->attrs == NULL)
      result = StrAppend(result, "<", SCHEDULER_IMAGES[s->method], "/>", NULL);
    else {
      result = StrAppend(result, "<", SCHEDULER_IMAGES[s->method], ">", NULL);
      for(i = 0; s->attrs[i] != NULL; i++) {
        attr = STRDUP(s->attrs[i]);
        *strchr(attr, '=') = '\0';
        result = StrAppend(result,
          "<option name='", attr,
          "' value='", attr + strlen(attr) + 1, "'/>", NULL);
        FREE(attr);
      }
      result = StrAppend(result, "</", SCHEDULER_IMAGES[s->method], ">", NULL);
    }
  }
  result = StrAppend(result, "</host>", NULL);
  return result;

}


char *
compute_task_to_xml(Task t) {

  char *arguments;
  File f;
  char **files;
  char *host;
  unsigned i;
  char *id;
  char *input;
  char *joined;
  char meta[200] = "";
  char *output;
  char priority[30] = "";
  char *result;
  char seqImage[10];
  char *stderrf;
  char *stdinf;
  char *stdoutf;
  char *wd;

  id = ASH_GetValue(global.userIds, t, NULL);
  id = id == NULL ? "" : StrAppend(NULL, " id='", id, "'", NULL);
  sprintf(seqImage, "%ld", t->seq);
  if(t->priority != 0)
    sprintf(priority, " priority='%d'", t->priority);
  wd = ASH_GetValue(global.wds, t, NULL);
  wd = wd == NULL ? "" : StrAppend(NULL, " wd='", wd, "'", NULL);
  if(t->arguments == NULL)
    arguments = "";
  else {
    joined = StrArrayJoin((const char **)t->arguments, " ");
    arguments = StrAppend(NULL, " arguments='", joined, "'", NULL);
    FREE(joined);
  }
  if(t->meta.memory > 0)
    sprintf(meta + strlen(meta), " memory='%.0f'", t->meta.memory);
  if(t->meta.cost > 0)
    sprintf(meta + strlen(meta), " cost='%.0f'", t->meta.cost);
  if(t->meta.timeStarted > 0)
    sprintf(meta + strlen(meta), " started='%ld'", t->meta.timeStarted);
  if(t->meta.timeCompleted > 0)
    sprintf(meta + strlen(meta), " finished='%ld'", t->meta.timeCompleted);
  if(Size(t->input) == 0)
    input = "";
  else {
    files = (char **)MALLOC(Size(t->input) * sizeof(char *));
    for(i = 0; i < Size(t->input); i++)
      files[i] = t->input[i]->name;
    joined = StrArrayNJoin((const char **)files, i, " ");
    input = StrAppend(NULL, " input='", joined, "'", NULL);
    FREE(joined);
    FREE(files);
  }
  if(Size(t->output) == 0)
    output = "";
  else {
    files = (char **)MALLOC(Size(t->output) * sizeof(char *));
    for(i = 0; i < Size(t->output); i++)
      files[i] = t->output[i]->name;
    joined = StrArrayNJoin((const char **)files, i, " ");
    output = StrAppend(NULL, " output='", joined, "'", NULL);
    FREE(joined);
    FREE(files);
  }
  f = ASH_GetValue(global.stdins, t, NULL);
  stdinf = f == NULL ? "" : StrAppend(NULL, " stdin='", f->name, "'", NULL);
  f = ASH_GetValue(global.stdouts, t, NULL);
  stdoutf = f == NULL ? "" : StrAppend(NULL, " stdout='", f->name, "'", NULL);
  f = ASH_GetValue(global.stderrs, t, NULL);
  stderrf = f == NULL ? "" : StrAppend(NULL, " stderr='", f->name, "'", NULL);
  host = t->host == NULL ? "" :
         StrAppend(NULL, " host='", t->host->id, "'", NULL);
  result = StrAppend
    (NULL, "<task", id, " position='", seqImage, "'", priority,
     " executable='", t->exec->name, "'", arguments, input, output, stdinf,
     stdoutf, stderrf, " status='", TASK_STATUS_IMAGES[t->stat], "'", meta,
     host, "/>", NULL);
  if(*arguments != '\0')
    FREE(arguments);
  if(*host != '\0')
    FREE(host);
  if(*id != '\0')
    FREE(id);
  if(*input != '\0')
    FREE(input);
  if(*output != '\0')
    FREE(output);
  if(*stderrf != '\0')
    FREE(stderrf);
  if(*stdinf != '\0')
    FREE(stdinf);
  if(*stdoutf != '\0')
    FREE(stdoutf);
  if(*wd != '\0')
    FREE(wd);
  return result;

}


int
compute_shutdown(void) {
  char *c;
  unsigned i;
  for(i = 0; i < Size(pendingTasks); i++) {
    if((c = StderrCaptureRetrieve(pendingTasks[i]->stderrKey)) != NULL)
      FREE(c);
    FREE(pendingTasks[i]);
  }
  VectorFree(pendingTasks);
  for(i = 0; i < Size(rashGroups); i++) {
    RashGroupInfo info = rashGroups[i];
    HostList_Free(info->rashHosts, 0);
    if(info->group != NULL)
      RASHGRP_GroupFree(info->group);
    if(info->status == GROUP_SUBMITTED &&
       (c = StderrCaptureRetrieve(info->stderrKey)) != NULL)
      FREE(c);
    FREE(info);
  }
  VectorFree(rashGroups);
  return 1;
}
