/* $Id: daemon.c,v 1.93 2003/11/25 23:03:02 jhayes Exp $ */
/*
 * Copyright  2002 The Regents of the University of California. 
 * All Rights Reserved. 
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes, without
 * fee, and without a written agreement is hereby granted, provided that the
 * above copyright notice, this paragraph and the following three paragraphs
 * appear in all copies. 
 *
 * Permission to incorporate this software into commercial products may be
 * obtained by contacting
 * Technology Transfer Office 
 * 9500 Gilman Drive 
 * 411 University Center 
 * University of California 
 * La Jolla, CA 92093-0093
 * (858) 534-5815
 * invent@ucsd.edu
 *
 * This software program and documentation are copyrighted by The Regents of
 * the University of California. The software program and documentation are
 * supplied "as is", without any accompanying services from The Regents. The
 * Regents does not warrant that the operation of the program will be
 * uninterrupted or error-free. The end-user understands that the program was
 * developed for research purposes and is advised not to rely exclusively on
 * the program for any reason. 
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
 * EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED
 * HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO
 * OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS. 
 */


#include "config.h"
#include <stdio.h>    /* file functions */
#include <stdlib.h>   /* strtol */
#include <string.h>   /* strcmp strlen */
#include <strings.h>  /* strcasecmp */
#include <time.h>     /* time */
#include <fileseed.h>
#define ASSTR_SHORT_NAMES
#include <strseed.h>
#define ASV_SHORT_NAMES
#include <vectorseed.h>
#define ASXML_SHORT_NAMES
#include <xmlseed.h>
#include "compute.h"
#include "global.h"
#include "misc.h"
#include "predictor.h"
#include "protocol.h"
#include "scheduler.h"
#include "storage.h"
#include "daemon.h"


static char *errorMessage = NULL;
struct Global global;


/* Sets the value returned by DaemonErrorMessage to #message#. */
static void
SetError(char *message) {
  if(errorMessage != NULL)
    FREE(errorMessage);
  errorMessage = message;
}


/* Returns the task sequence number associated with #s#; -1 on error. */
static SequenceNumber
GetTaskSequence(const char *s) {
  char *c;
  int result = strtol(s, &c, 10);
  Task t;
  if(*c == '\0')
    return result;
  t = ASH_GetKey(global.userIds, (ASH_ValueType)s, &ASH_StringCompare);
  return t == NULL ? -1 : t->seq;
}


/*
 * Translates the task specifications included in #tasks# into a set of task
 * sequence numbers, returned in the #sequencesCount#-long array #sequences#.
 * The specifications can contain ranges (n-m), task names, task group names
 * and/or the word "all". #sequences# is allocated and must be freed after use.
 *  Returns 1 if successful, else 0.
 */
static int
GetTaskSequences(const char *tasks,
                 SequenceNumber **sequences,
                 unsigned *sequencesCount) {

  char *c;
  char *group;
  int high;
  unsigned i;
  ASH_HashIterator it;
  TaskList included = TaskList_New(SMALL_HASH);
  unsigned j;
  int low;
  int negate;
  int result = 1;
  Task t;
  TaskList tl;
  char **words = StrSplit(tasks, NULL);

  for(i = 0; words[i] != NULL; i++) {
    group = words[i];
    negate = *group == '!';
    if(negate) {
      group++;
      if(i == 0) {
        /* First group negated; prepend implicit "all". */
        ASH_ForEachMapping(global.tasks, it)
          TaskList_Add(included, (Task)ASH_IteratorValue(it));
      }
    }
    tl = strcmp(group, "all") == 0 ? global.tasks :
         (TaskList)ASH_GetValue(global.taskGroups, group, NULL);
    if(tl != NULL) {
      ASH_ForEachMapping(tl, it) {
        t = (Task)ASH_IteratorValue(it);
        if(negate)
          TaskList_Remove(included, t);
        else
          TaskList_Add(included, t);
      }
    }
    else {
      if((c = strchr(group, '-')) != NULL)
        *c = '\0';
      if((low = GetTaskSequence(group)) < 0) {
        low = 0;
        SetError(StrAppend(NULL, "Unknown group '", group, "'", NULL));
        result = 0;
      }
      if(c == NULL)
        high = low;
      else if((high = GetTaskSequence(c + 1)) < 0) {
        high = 0;
        SetError(StrAppend(NULL, "Unknown group '", c + 1, "'", NULL));
        result = 0;
      }
      for(j = low; j <= high; j++) {
        if((t = TaskList_Get(global.tasks, j)) == NULL)
          ; /* empty */
        else if(negate)
          TaskList_Remove(included, t);
        else
          TaskList_Add(included, t);
      }
    }
  }
  StrArrayFree(words);

  if(result) {
    Task *tasks = (Task *)MALLOC(ASH_GetMappingCount(included) * sizeof(Task));
    unsigned tasksCount = 0;
    ASH_ForEachMapping(included, it)
      tasks[tasksCount++] = (Task)ASH_IteratorValue(it);
    SortTasks(tasks, tasksCount, 0);
    *sequences = (SequenceNumber *)MALLOC(tasksCount * sizeof(SequenceNumber));
    for(i = 0; i < tasksCount; i++)
      (*sequences)[i] = tasks[i]->seq;
    *sequencesCount = tasksCount;
    FREE(tasks);
  }
  TaskList_Free(included, 0);
  return result;

}


const char *
DaemonErrorMessage(void) {
  return errorMessage;
}


int
DoAdd(const char *xml) {

  static const char *ERROR_MESSAGES[] = {
    "",
    "invalid closing tag",
    "invalid attribute name",
    "missing attribute value",
    "missing closing tag",
    "missing '-->'",
    "missing '='",
    "missing '>'",
    "missing '<'",
    "missing '?>'",
    "missing tag name",
    "empty document",
    "multiple topmost tags"
  };

  XmlNode apst;
  char *c;
  XmlNode child;
  unsigned long col;
  XmlNode document;
  XmlParseErrorCodes err;
  unsigned long line;
  int result = 1;
  XmlNodeIterator textIterator;

  if(*xml == '@') {
    FILE *f;
    if((f = fopen(xml + 1, "r")) == NULL) {
      SetError(StrAppend(NULL, "Unable to open ", xml + 1, NULL));
      FAIL1("DoAdd: %s\n", errorMessage);
    }
    document = XmlFromStream((XmlGet)&fgetc, f, NULL, 0, 0, NULL, &err, &col);
    if(document == NULL) {
      char buffer[256];
      for(rewind(f), line = 1;
          fgets(buffer, sizeof(buffer), f) != NULL && strlen(buffer) < col;
          col -= strlen(buffer))
        if(strchr(buffer, '\n') != NULL)
          line++;
    }
    fclose(f);
  }
  else {
    document = XmlFromString((DOMString)xml, NULL, 0, 0, NULL, &err, &col);
    if(document == NULL) {
      const char *c;
      for(line = 1;
          (c = strchr(xml, '\n')) != NULL && c - xml < col;
          col -= xml - c) {
        xml = c + 1;
        line++;
      }
    }
  }
  if(document == NULL) {
    char position[20];
    sprintf(position, "%ld:%ld", line, col);
    SetError(StrAppend
      (NULL, ERROR_MESSAGES[(int)err], " at ", position, " in XML", NULL));
    FAIL1("DoAdd: %s\n", errorMessage);
  }
  apst = DocumentElement(document);

  /* Chew through disk definitions first, since the others depend on them. */
  for(child = NodeFirstChildByType(apst, SHOW_ELEMENT);
      child != NULL;
      child = NodeNextSiblingByType(child, SHOW_ELEMENT)) {
    if(strcmp(NodeName(child), "storage") != 0)
      continue;
    if(!storage_process_xml(child))
      result = 0;
  }

  for(child = NodeFirstChildByType(apst, SHOW_ELEMENT);
      child != NULL;
      child = NodeNextSiblingByType(child, SHOW_ELEMENT)) {
    const char *name = NodeName(child);
    if(strcmp(name, "compute") == 0 || strcmp(name, "tasks") == 0)
      result = result && compute_process_xml(child);
    else if(strcmp(name, "files") == 0)
      result = result && storage_process_xml(child);
    else if(strcmp(name, "gridinfo") == 0) {
	//printf("I am coming before predictor_process_xml and the result is %d \n", result);
      result = result && predictor_process_xml(child);
	//printf("I am coming after predictor_process_xml and the result is %d \n", result);
	}
    else if(strcmp(name, "log") == 0)
      result = result && ProcessLogXml(child);
    else if(strcmp(name, "storage") == 0)
      ; /* Handled above */
    else {
      ERROR1("Unknown tag '%s'\n", name);
      result = 0;
    }
  }

  /* Log a warning about any non-blank text in the XML. */
  textIterator = DocumentCreateNodeIterator
    (document, DocumentElement(document), SHOW_TEXT, NULL, 0);
  while((child = NodeIteratorNextNode(textIterator)) != NULL) {
    for(c = NodeValue(child); *c !='\0'&& strchr(" \f\n\r\t\v", *c)!=NULL; c++)
      ; /* empty */
    if(*c != '\0')
      WARN1("Ignoring text '%s' in XML\n", c);
  }
  NodeIteratorDetach(textIterator);

  NodeFree(document);
  if(!result)
    SetError(STRDUP("Semantic error in XML"));
  return result;

}


int
DoCancelTasks(const char *tasks) {
  unsigned i;
  SequenceNumber *sequences;
  unsigned sequencesCount;
  Task t;
  if(!GetTaskSequences(tasks, &sequences, &sequencesCount))
    return 0;
  for(i = 0; i < sequencesCount; i++) {
    if((t = TaskList_Get(global.tasks, sequences[i])) != NULL)
      TaskCanceled(t);
  }
  FREE(sequences);
  return 1;
}


int
DoDisable(const char *resources) {
  Host host;
  Host *hosts;
  unsigned long hostsLen;
  if(strcasecmp(resources, "ALL") == 0) {
    ASH_GetAllValues(global.hosts, (ASH_ValueType **)&hosts, &hostsLen);
    while(hostsLen > 0) {
      host = hosts[--hostsLen];
      HostDisabled(host);
      host->meta.timeDisabled = -1; /* Mark as client disabled. */
    }
    FREE(hosts);
  }
  else if((host = HostList_Get(global.hosts, (char *)resources)) == NULL) {
    SetError(StrAppend(NULL, "Unknown resource '", resources, "'", NULL));
    FAIL1("DoDisable: %s\n", errorMessage);
  }
  else {
    HostDisabled(host);
    host->meta.timeDisabled = -1; /* Mark as client disabled. */
  }
  return 1;
}


int
DoDownloadFile(const char *fileName) {

  FileCopy copy;
  File file;
  FileTransfer ft;
  FileTransfer oldFt;

  /* Verify that file is known and that we have a copy available. */
  file = FileList_Get(global.files, (char *)fileName);
  if(file == NULL) {
    SetError(StrAppend(NULL, "Unknown file '", fileName, "'", NULL));
    FAIL1("DoDownloadFile: %s\n", errorMessage);
  } 
  if(ASH_GetMappingCount(file->copies) == 0) {
    SetError(StrAppend(NULL, "File '", fileName, "' is unavailable", NULL));
    FAIL1("DoDownloadFile: %s\n", errorMessage);
  }

  /* Transfer file */
  copy = (FileCopy)ASH_IteratorValue(ASH_IteratorFirst(file->copies));
  ft = FileTransfer_New(file, copy->disk, FindLocalDisk()); 
  if((oldFt = FileTransferList_Get(global.transfers, ft->id)) == NULL)
    FileTransferList_Add(global.transfers, ft);
  else {
    FileTransfer_Free(ft);
    FileTransferReset(oldFt);
    ft = oldFt;
  }
  storage_do_file_transfer(ft); 
  return 1;

}


int
DoEnable(const char *resources,
         long after) {
  Host host;
  Host *hosts;
  unsigned long hostsLen;
  if(strcasecmp(resources, "ALL") == 0) {
    ASH_GetAllValues(global.hosts, (ASH_ValueType **)&hosts, &hostsLen);
    while(hostsLen > 0) {
      host = hosts[--hostsLen];
      if(host->meta.timeDisabled >= after)
        HostEnabled(host);
    }
    FREE(hosts);
  }
  else if((host = HostList_Get(global.hosts, (char *)resources)) == NULL) {
    SetError(StrAppend(NULL, "Unknown resource '", resources, "'", NULL));
    FAIL1("DoEnable: %s\n", errorMessage);
  }
  else if(host->meta.timeDisabled >= after)
    HostEnabled(host);
  return 1;
}


int
DoGetFile(const char *path,
          char **contents) {
  if((*contents = ASFILE_FileContents(path)) == NULL) {
    SetError(STRDUP("Unable to read file"));
    FAIL1("DoGetFile: %s\n", errorMessage);
  }
  return 1;
}


int
DoGetState(const char *tasks,
           char **state) {
  SequenceNumber *sequences;
  unsigned sequencesCount;
  if(!GetTaskSequences(tasks, &sequences, &sequencesCount))
    return 0;
  *state = GetStateXML(sequences, sequencesCount);
  FREE(sequences);
  return 1;
}


int
DoGetStatus(char **status) {
  char buffer[256];
  sprintf(buffer, "%ld %d %ld %ld %ld %ld %ld",
          (long)(time(NULL) - global.timeStarted),
          global.commandCount,
          ASH_GetMappingCount(global.tasks),
          TaskGroupMembership("new"),
          TaskGroupMembership("running"),
          TaskGroupMembership("done"),
          TaskGroupMembership("failed"));
  *status = STRDUP(buffer);
  return 1;
}


int
DoNotify(const char *address) {
  AddEmailNotification(address);
  return 1;
}


int
DoPrioritize(unsigned priority,
             const char *tasks) {
  unsigned i;
  SequenceNumber *sequences;
  unsigned sequencesCount;
  Task t;
  if(!GetTaskSequences(tasks, &sequences, &sequencesCount))
    return 0;
  for(i = 0; i < sequencesCount; i++) {
    if((t = TaskList_Get(global.tasks, sequences[i])) != NULL)
      TaskPrioritized(t, priority);
  }
  FREE(sequences);
  return 1;
}


int
DoReset(const char *tasks) {
  unsigned i;
  SequenceNumber *sequences;
  unsigned sequencesCount;
  Task t;
  if(!GetTaskSequences(tasks, &sequences, &sequencesCount))
    return 0;
  for(i = 0; i < sequencesCount; i++) {
    if((t = TaskList_Get(global.tasks, sequences[i])) != NULL)
      TaskReset(t);
  }
  FREE(sequences);
  return 1;
}
