/* $Id: daemonmain.c,v 1.94 2004/04/07 21:39:55 jhayes Exp $ */
/*
 * Copyright  2002 The Regents of the University of California. 
 * All Rights Reserved. 
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes, without
 * fee, and without a written agreement is hereby granted, provided that the
 * above copyright notice, this paragraph and the following three paragraphs
 * appear in all copies. 
 *
 * Permission to incorporate this software into commercial products may be
 * obtained by contacting
 * Technology Transfer Office 
 * 9500 Gilman Drive 
 * 411 University Center 
 * University of California 
 * La Jolla, CA 92093-0093
 * (858) 534-5815
 * invent@ucsd.edu
 *
 * This software program and documentation are copyrighted by The Regents of
 * the University of California. The software program and documentation are
 * supplied "as is", without any accompanying services from The Regents. The
 * Regents does not warrant that the operation of the program will be
 * uninterrupted or error-free. The end-user understands that the program was
 * developed for research purposes and is advised not to rely exclusively on
 * the program for any reason. 
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
 * EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED
 * HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO
 * OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS. 
 */


/* 
 * This is the main code of the APST daemon. The daemon executes
 * an infinite loop and reacts to:
 *   - messages from the APST client (adding/canceling tasks,...)
 *   - events from the underlying Grid software (task completed/failed,...)
 */

#include "config.h"
#include <sys/types.h>
#include <sys/stat.h>     /* permission values */
#include <ctype.h>        /* isspace */
#include <fcntl.h>        /* open */
#include <signal.h>       /* signal */
#include <stdio.h>        /* file funcs */
#include <stdlib.h>       /* getenv srandom strtol */
#include <string.h>       /* strchr strcmp strlen */
#include <time.h>         /* time */
#include <unistd.h>       /* close fork getcwd getpid setpgid sleep */
#include <elbase.h>
#include <rashgrp.h>
#define ASENV_SHORT_NAMES
#include <envseed.h>      /* argv parsing */
#include <fileseed.h>     /* FileContents */
#define ASIP_SHORT_NAMES
#include <ipseed.h>       /* Sockets */
#define ASPROP_SHORT_NAMES
#include <propertyseed.h> /* rc file parsing */
#define ASSTR_SHORT_NAMES
#include <strseed.h>      /* StrAppend */
#define ASV_SHORT_NAMES
#include <vectorseed.h>
#include "compute.h"
#include "daemon.h"
#include "global.h"
#include "predictor.h"
#include "protocol.h"
#include "scheduler.h"
#include "storage.h"


typedef enum {
  SECURITY_NONE, SECURITY_MINIMAL, SECURITY_MEDIUM, SECURITY_STRICT
} SecurityLevelTypes;


static const char *APSTRC_NAME = ".apstrc";
/* Compute/storage polls done every TIME_STEP seconds. */
static const int TIME_STEP = 3;
static const int FILE_FLAGS = O_WRONLY | O_CREAT | O_TRUNC;
static const int FILE_PERMISSIONS = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP;
static const int STDERR = 2;


ASDIAG_DiagnosticState ds;


/* Module globals. */
static Port apstdPort;
static ASIP_Socket apstdSocket;
static int autoenableInterval;
static char *checkpointPath;
static Socket *clients;
static char *logPath;
static SecurityLevelTypes security;
static int stepsUntilBreak = -1;


/*
 * If #switchName# was included on the command line, returns its value;
 * otherwise, if #apstrcName# is defined in #apstrcProperties#, returns its
 * value; otherwise, returns #defaultValue#.
 */
static const char *
Setting(ParsedArgv pa,
        const char *switchName,
        PropertyList apstrcProperties,
        const char *apstrcName,
        const char *defaultValue) {
  const char *result;
  if((result = SwitchValue(pa, switchName, NULL)) != NULL)
    return result;
  if(apstrcProperties != NO_PROPERTY_LIST &&
     (result = FindPropertyValueByName(apstrcProperties, apstrcName)) != NULL)
    return result;
  return defaultValue;
}


/* Sets global and module settings from .apstrc, switches, and defaults. */
static void
DaemonInit(ParsedArgv pa) {

  char defaultPath[256];
  char *homeDir;
  PropertyList rc = NO_PROPERTY_LIST;
  const char *value;

  ASDIAG_DirectDiagnostics(ds, ASDIAG_DEBUG_MESSAGE, STDERR);
  ASDIAG_DirectDiagnostics(ds, ASDIAG_INFO_MESSAGE, STDERR);
  ASDIAG_DirectDiagnostics(ds, ASDIAG_LOG_MESSAGE, STDERR);
  ASDIAG_DirectDiagnostics(ds, ASDIAG_WARNING_MESSAGE, STDERR);
  ASDIAG_DirectDiagnostics(ds, ASDIAG_ERROR_MESSAGE, STDERR);
  ASDIAG_DirectDiagnostics(ds, ASDIAG_FATAL_MESSAGE, STDERR);

  /* Read $HOME/.apstrc, if possible. */
  if((homeDir = getenv("HOME")) != NULL) {
    char *apstrcPath;
    FILE *f;
    apstrcPath = StrAppend(NULL, homeDir, "/", APSTRC_NAME, NULL);
    if((f = fopen(apstrcPath, "r")) != NULL) {
      rc = ReadPropertyList(f);
      fclose(f);
      if(rc == NO_PROPERTY_LIST)
        WARN1("Error reading %s\n", apstrcPath);
    }
    FREE(apstrcPath);
  }

  /* Get module settings from switches and .apstrc. */
  apstdPort = (Port)
    strtol(Setting(pa,"--port",rc,"APSTD_PORT",DEFAULT_APSTD_PORT), NULL, 10);
  autoenableInterval =
    strtol(Setting(pa, "--autoenable", rc, "AUTOENABLE", "0"), NULL, 10);
  sprintf(defaultPath, "/tmp/apstd%dcheckpoint.xml", apstdPort);
  value = Setting(pa, "--checkpointfile", rc, "CHECKPOINT_FILE", defaultPath);
  checkpointPath =
    strcmp(value, "none") == 0 || strcmp(value, "/dev/null") == 0 ?
    NULL : STRDUP(value);
  sprintf(defaultPath, "/tmp/apstd%d.log", apstdPort);
  value = Setting(pa, "--logfile", rc, "LOG_FILE", defaultPath);
  logPath = strcmp(value, "none") == 0 || strcmp(value, "/dev/null") == 0 ?
                NULL : STRDUP(value);
  value = Setting(pa, "--security", rc, "SECURITY", "minimal");
  security = strcmp(value, "none") == 0 ? SECURITY_NONE :
             strcmp(value, "minimal") == 0 ? SECURITY_MINIMAL :
             strcmp(value, "medium") == 0 ? SECURITY_MEDIUM :
             SECURITY_STRICT;
  stepsUntilBreak = *Setting(pa, "--hold", rc, "HOLD", "no") == 'y' ? 0 : -1;

  /* Get globals settings from switches and .apstrc. */
  global.bulkCopy = *Setting(pa, "--bulkcopy", rc, "BULK_COPY", "yes") == 'y';
  global.commandCount = 0;
  global.downloadstd =
    *Setting(pa, "--downloadstd", rc, "DOWNLOAD_STD", "yes") == 'y';
  global.findInput = *Setting(pa, "--findinput", rc, "FIND_INPUT", "no") == 'y';
  value = Setting(pa, "--heuristic", rc, "HEURISTIC", "wq");
  global.heuristic = strcmp(value, "maxmin") == 0 ? HEURISTIC_MAXMIN :
                     strcmp(value, "minmin") == 0 ? HEURISTIC_MINMIN :
                     strcmp(value, "roundrobin") == 0 ? HEURISTIC_ROUNDROBIN :
                     strcmp(value, "sufferage") == 0 ? HEURISTIC_SUFFERAGE :
                     strcmp(value, "wq") == 0 ? HEURISTIC_WORK_QUEUE :
                     strcmp(value, "ghs") == 0 ? HEURISTIC_GHS:
                     HEURISTIC_XSUFFERAGE;
  global.simDisk =
    strtol(Setting(pa, "--simdisk", rc, "SIMDISK", "-1"), NULL, 10);
  global.simHost =
    strtol(Setting(pa, "--simhost", rc, "SIMHOST", "-1"), NULL, 10);
  global.simRandom = *Setting(pa, "--simrand", rc, "SIMRAND", "no") == 'y';
  global.timeStarted = time(NULL);
  srandom(time(NULL));

  if(rc != NO_PROPERTY_LIST)
    PropertyListFree(&rc);

}


/* Enables any hosts that have been disabled for the autoenable interval. */
static void
AutoEnableHosts(void) {
  static time_t lastCall = 0;
  if(autoenableInterval <= 0 || time(NULL) - lastCall < autoenableInterval)
    return;
  DoEnable("ALL", 0);
  lastCall = time(NULL);
}

/* Returns 1 or 0 based on whether or not #command# is allowed under #level#. */
static int
CommandAllowed(SecurityLevelTypes level,
               const char *command) {
  static const char *HARMLESS_COMMANDS =
    APST_CLIENT_GET_ID APST_CLIENT_NOOP;
  static const char *INFO_COMMANDS =
    APST_CLIENT_GET_FILE APST_CLIENT_GET_STATE APST_CLIENT_GET_STATUS
    APST_CLIENT_NOTIFY;
  static const char *STATE_COMMANDS =
    APST_CLIENT_CANCEL_TASKS APST_CLIENT_CLEAR APST_CLIENT_DISABLE_HOSTS
    APST_CLIENT_DOWNLOAD_FILE APST_CLIENT_ENABLE_HOSTS APST_CLIENT_PRIORITIZE
    APST_CLIENT_RESET_TASKS APST_CLIENT_STEP APST_CLIENT_STOP;
  return level==SECURITY_NONE || strstr(HARMLESS_COMMANDS,command) != NULL ? 1 :
         level == SECURITY_STRICT ? 0 :
         strstr(INFO_COMMANDS, command) != NULL ? 1 :
         (level == SECURITY_MINIMAL && strstr(STATE_COMMANDS, command) != NULL);
}


/* Executes #command#, passing #args#; returns command output in #output#. */
static void
DoCommand(const char *command,
          const char *args,
          char **output) {

  char *content = NULL;
  int result;
  const char *tag;

  if(strcmp(command, APST_CLIENT_ADD) == 0)
    result = DoAdd(args);
  else if(strcmp(command, APST_CLIENT_CANCEL_TASKS) == 0)
    result = DoCancelTasks(args);
  else if(strcmp(command, APST_CLIENT_CLEAR) == 0) {
    if((result = DoCancelTasks("new running"))) {
      sched_shutdown();
      predictor_shutdown();
      compute_shutdown();
      storage_shutdown();
      GlobalDeinit();
      GlobalInit();
      storage_init();
      compute_init();
      predictor_init();
      sched_init();
    }
  }
  else if(strcmp(command, APST_CLIENT_DISABLE_HOSTS) == 0)
    result = DoDisable(args);
  else if(strcmp(command, APST_CLIENT_DOWNLOAD_FILE) == 0)
    result = DoDownloadFile(args);
  else if(strcmp(command, APST_CLIENT_ENABLE_HOSTS) == 0)
    result = DoEnable(args, -1);
  else if(strcmp(command, APST_CLIENT_GET_FILE) == 0)
    result = DoGetFile(logPath, &content);
  else if(strcmp(command, APST_CLIENT_GET_STATE) == 0)
    result = DoGetState(args, &content);
  else if(strcmp(command, APST_CLIENT_GET_STATUS) == 0)
    result = DoGetStatus(&content);
  else if(strcmp(command, APST_CLIENT_NOOP) == 0)
    result = 1;
  else if(strcmp(command, APST_CLIENT_NOTIFY) == 0)
    result = DoNotify(args);
  else if(strcmp(command, APST_CLIENT_PRIORITIZE) == 0) {
    unsigned priority = strtol(args, (char **)&args, 10);
    while(*args == ' ')
      args++;
    result = DoPrioritize(priority, args);
  }
  else if(strcmp(command, APST_CLIENT_RESET_TASKS) == 0)
    result = DoReset(args);
  else if(strcmp(command, APST_CLIENT_STEP) == 0) {
    stepsUntilBreak = strtol(args, NULL, 10);
    result = 1;
  }
  else if(strcmp(command, APST_CLIENT_STOP) == 0)
    result = DoReset("running");
  else {
    content = StrAppend(NULL, "unknown command '", command, "'", NULL);
    WARN1("%s\n", content);
    result = 0;
  }
  tag = result ? APST_CLIENT_OK : APST_CLIENT_FAIL;
  if(content == NULL)
    content = STRDUP(result ? "" : DaemonErrorMessage());
  *output = StrAppend(NULL, tag, "\n", content, NULL);
  FREE(content);
  global.commandCount++;

}


/*
 * Processes any messages that have arrived from a client.  Sets #stopReceived#
 * to 1 iff a message indicating that the daemon should terminate is received;
 * otherwise leaves it unchanged.
 */
static void
CheckClients(int *stopReceived) {

  char *args;
  char *command;
  char *contents;
  unsigned i;
  char *reply;
  Socket sock;

  while(CheckIfReadable(apstdSocket, 0)) {
    sock = Accept(apstdSocket);
    if(sock == NO_SOCKET) {
      WARN("Accept failed\n");
      continue;
    }
    Append(clients, sock);
  }

  for(i = 0; i < Size(clients); i++) {
    sock = clients[i];
    if(!CheckIfReadable(sock, 0))
      continue;
    else if(!ReceiveString(sock, &command)) {
      Disconnect(&sock);
      Remove(clients, i);
      continue;
    }
    for(args = command; *args != '\0' && !isspace((int)*args); args++)
      ; /* empty */
    for( ; isspace((int)*args); args++)
      *args = '\0';
    reply = NULL;
    if(strcmp(command, APST_CLIENT_READ) == 0) {
      contents = *args != '\0' && strchr(args, '/') == NULL ?
                 ASFILE_FileContents(args) : NULL;
      if(contents == NULL)
        reply = StrAppend(NULL, APST_CLIENT_FAIL, "\n", "No such file", NULL);
      else {
        FREE(command);
        command = contents;
        for(args = command; *args != '\0' && !isspace((int)*args); args++)
          ; /* empty */
        for( ; isspace((int)*args); args++)
          *args = '\0';
      }
    }
    else if(!CommandAllowed(security, command))
      reply = StrAppend(NULL,APST_CLIENT_FAIL,"\n","Permission denied",NULL);
    if(reply == NULL) {
      if(strcmp(command, APST_CLIENT_GET_ID) == 0) {
        char cwd[512];
        char id[20];
        getcwd(cwd, sizeof(cwd));
        sprintf(id, "%d", (int)sock);
        reply = StrAppend(NULL, APST_CLIENT_OK, "\n", id, " ", cwd, NULL);
      }
      else {
        DoCommand(command, args, &reply);
        if(strcmp(command, APST_CLIENT_STOP) == 0)
          *stopReceived = 1;
      }
    }
    SendString(sock, reply);
    FREE(reply);
    FREE(command);
  }

}


/* Checks to see if any compute or storage work needs to be done. */
static void
CheckPolls(void) {
  static time_t lastCall = 0;
  if(time(NULL) - lastCall >= TIME_STEP) {
    lastCall = time(NULL);
    compute_poll();
    storage_poll();
  }
}


/*
 * Passes the contents of #path# to the DoAdd function.  Returns 1 if
 * successful, else 0.
 */
static int
ProcessXmlFile(const char *path) {
  char *atPath;
  FILE *f;
  int result;
  if((f = fopen(path, "r")) != NULL)
    atPath = StrAppend(NULL, "@", path, NULL);
  else {
    char *withExt = StrAppend(NULL, path, ".xml", NULL);
    f = fopen(withExt, "r");
    FREE(withExt);
    if(f == NULL)
      FAIL1("No such file %s\n", path);;
    atPath = StrAppend(NULL, "@", path, ".xml", NULL);
  }
  fclose(f);
  result = DoAdd(atPath);
  FREE(atPath);
  if(!result)
    FAIL1("%s\n", DaemonErrorMessage());
  return 1;
}


int
main(int argc,
     char **argv) {
 
  /* Split to avoid ISO 509 character limit on string literals. */
  const char *VALID1 =
    "-d                void     keep daemon in foreground\n"
    "-h                void     display help and exit\n"
    "-v                void     display version and exit\n"
    "--autoenable      int      seconds between autoenable tests\n"
    "--bulkcopy        yes|no   transfer files in bulk whenever possible\n"
    "--checkpointfile  string   path to checkpoint file or /dev/null\n";
  const char *VALID2 =
    "--downloadstd     yes|no   automatically download stdout/err\n"
    "--findinput       yes|no   look for pre-staged input files\n"
    "--heuristic       maxmin|minmin|roundrobin|sufferage|wq|xsufferage|ghs\n"
    "--hold            yes|no   wait for user command to begin work\n"
    "--logfile         string   path to log file or /dev/null\n";
  const char *VALID3 =
    "--port            int      daemon port\n"
    "--security        none|minimal|medium|strict\n"
    "--simdisk         int      simulation time for file transfers\n"
    "--simhost         int      simulation time for task execution\n"
    "--simrand         yes|no   randomize simulation times\n"
    "                  string   XML file";

  const char *fileName;
  int halt = 0;
  int i;
  int logFile = STDERR;
  ParsedArgv pa;
  char *s;
  char *VALID;

  VALID = StrAppend(NULL, VALID1, VALID2, VALID3, NULL);
  if((pa = ParseArgv((const char *const *)argv, VALID, NULL)) == NULL ||
      SwitchPresent(pa, "-h")){
    fprintf(stderr, "USAGE: apstd [<option> ...] [<XML file> ...]\n%s\n",VALID);
    if(pa != NULL)
      ParsedArgvFree(pa);
    return 1;
  }
  FREE(VALID);
  if(SwitchPresent(pa, "-v")) {
    fprintf(stderr,
            "AppLeS Parameter Sweep Template (APST) Daemon %s\n", APST_VERSION);
    ParsedArgvFree(pa);
    return 0;
  }

  GlobalInit();
  ds = ASDIAG_DiagnosticStateNew();
  DaemonInit(pa);
  if(checkpointPath != NULL) {
    unlink(checkpointPath);
    if((global.checkpointFile = fopen(checkpointPath, "w+")) == NULL)
      fprintf(stderr, "Cannot create checkpoint file: '%s'\n", checkpointPath);
  }
  clients = VectorNew(Socket);

  if((apstdSocket = OpenTcpPort(ANY_ADDRESS, apstdPort)) == NO_SOCKET) {
    fprintf(stderr, "Cannot listen on port %d\n", apstdPort);
    return 1;
  }

  RASHGRP_Init();
  storage_init();
  compute_init();
  predictor_init();
  sched_init();

  if(logPath != NULL) {
    if((logFile = open(logPath, FILE_FLAGS, FILE_PERMISSIONS)) == -1)
      fprintf(stderr, "Cannot create log file: '%s'\n", logPath);
    else {
      ASDIAG_DirectDiagnostics(ds, ASDIAG_DEBUG_MESSAGE, logFile);
      ASDIAG_DirectDiagnostics(ds, ASDIAG_INFO_MESSAGE, logFile);
      ASDIAG_DirectDiagnostics(ds, ASDIAG_LOG_MESSAGE, logFile);
      ASDIAG_DirectDiagnostics(ds, ASDIAG_WARNING_MESSAGE, logFile);
      ASDIAG_DirectDiagnostics(ds, ASDIAG_ERROR_MESSAGE, logFile);
      ASDIAG_DirectDiagnostics(ds, ASDIAG_FATAL_MESSAGE, logFile);
    }
  }

  for(i = 1; (fileName = SwitchValueN(pa, "", i, NULL)) != NULL; i++) {
    if(!ProcessXmlFile(fileName)) {
      fprintf(stderr, "Failed to parse the xml file %s\n", fileName);
      return 1;
    }
  }

  s = MyMachineName();
  LOG3("apstd version %s running on %s:%d\n", APST_VERSION, s, apstdPort);
  FREE(s);
  if(!SwitchPresent(pa, "-d")) {
    fprintf(stdout, "going into background\n");
    if(fork() != 0)
      return 0;
    setpgid(getpid(), getpid());
#if 0
    /* Closing stdin has caused ssh spawning problems in the past. */
    fclose(stdin);
#endif
  }
  ParsedArgvFree(pa);
  signal(SIGPIPE, SIG_IGN);

  while(1) { 
    CheckClients(&halt);
    if(halt)
      break;
    if(!sched_run(stepsUntilBreak != 0)) {
      ERROR("Error while running the scheduler\n");
      break;
    }
    if(stepsUntilBreak > 0)
      stepsUntilBreak--;
    CheckPolls();
    AutoEnableHosts();
    sleep(1);
  }

  /* Shutting down the sub-systems */
  sched_shutdown();
  predictor_shutdown();
  compute_shutdown();
  storage_shutdown();
  RASHGRP_Shutdown();

  for(i = 0; i < Size(clients); i++)
    Disconnect(&clients[i]);
  Disconnect(&apstdSocket);
  VectorFree(clients);
  GlobalDeinit();
  LOG("apstd terminated\n");
  if(logFile != STDERR)
    close(logFile);
  if(global.checkpointFile != NULL)
    fclose(global.checkpointFile);
  ASDIAG_DiagnosticStateFree(ds);
  return 0;

}
