/* $Id: gantt.c,v 1.8 2003/09/15 19:56:23 jhayes Exp $ */
/*
 * Copyright  2002 The Regents of the University of California. 
 * All Rights Reserved. 
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes, without
 * fee, and without a written agreement is hereby granted, provided that the
 * above copyright notice, this paragraph and the following three paragraphs
 * appear in all copies. 
 *
 * Permission to incorporate this software into commercial products may be
 * obtained by contacting
 * Technology Transfer Office 
 * 9500 Gilman Drive 
 * 411 University Center 
 * University of California 
 * La Jolla, CA 92093-0093
 * (858) 534-5815
 * invent@ucsd.edu
 *
 * This software program and documentation are copyrighted by The Regents of
 * the University of California. The software program and documentation are
 * supplied "as is", without any accompanying services from The Regents. The
 * Regents does not warrant that the operation of the program will be
 * uninterrupted or error-free. The end-user understands that the program was
 * developed for research purposes and is advised not to rely exclusively on
 * the program for any reason. 
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
 * EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED
 * HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO
 * OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS. 
 */


#include "config.h"
#include <stdlib.h> /* calloc free */
#define ASV_SHORT_NAMES
#include <vectorseed.h>
#include "gantt.h"
#define HOLE_TYPE -1


/*
 * We implement a GanttChart as a vector of GanttColumns, each of which
 * contains a doubly-linked list of GanttEvent.  Gaps in a column's events are
 * occupied by events with a type field of HOLE_TYPE.  Holes in a column are
 * linked together (singly) via their #details# field.
 */


/*
 * Returns the hole in #c# in which a new event would be inserted if the event
 * had a begin value of no less than #earliestBegin# and a running time of
 * #runTime#.  Returns NULL if the new event would be appended to #c#.
 */
static GanttEvent
FindEventHole(GanttColumn c,
              unsigned long earliestStart,
              unsigned long runTime) {
  GanttEvent e;
  unsigned long earliestEnd = earliestStart + runTime;
  for(e = c->firstHole; e != NULL; e = (GanttEvent)e->details)
    if(e->end >= earliestEnd && e->end - e->begin >= runTime)
      break; /* sufficient room for insertion */
  return e;
}


/* Adds a hole from #begin# to #end# in #c# just before #nextEvent#. */
static void
AddHole(GanttColumn c,
        unsigned long begin,
        unsigned long end,
        GanttEvent nextEvent) {
  GanttEvent e;
  GanttEvent h = GanttEvent_New(HOLE_TYPE, begin, end, NULL);
  if((h->previous = nextEvent->previous) == NULL)
    c->firstEvent = h;
  else
    h->previous->next = h;
  h->next = nextEvent;
  nextEvent->previous = h;
  if(c->firstHole == NULL || c->firstHole->begin >= end) {
    h->details = c->firstHole;
    c->firstHole = h;
  }
  else {
    for(e = c->firstHole;
        e->details != NULL && ((GanttEvent)e->details)->begin < end;
        e = (GanttEvent)e->details)
      ; /* empty */
    h->details = e->details;
    e->details = h;
  }
}


/* Removes #hole# from #c#'s hole list.  Returns #hole# for convenience. */
static GanttEvent
RemoveHole(GanttColumn c,
           GanttEvent hole) {
  GanttEvent e;
  if(c->firstHole == hole)
    c->firstHole = (GanttEvent)hole->details;
  else {
    for(e = c->firstHole;
        e->details != NULL && (GanttEvent)e->details != hole;
        e = (GanttEvent)e->details)
      ; /* empty */
    if((GanttEvent)e->details == hole)
      e->details = hole->details;
  }
  return hole;
}


void
GanttEvent_Free(GanttEvent e) {
  free(e);
}


GanttEvent
GanttEvent_New(int type,
               unsigned long begin,
               unsigned long end,
               void *details) {
  GanttEvent e = malloc(sizeof(struct _GanttEvent));
  e->type = type;
  e->begin = begin;
  e->end = end;
  e->details = details;
  return e;
}


void
GanttColumn_AddEvent(GanttColumn c,
                     GanttEvent e) {
  GanttEvent hole = FindEventHole(c, e->begin, e->end - e->begin);
  unsigned long priorEnd;
  if(hole == NULL) {
    e->next = NULL;
    e->previous = c->lastEvent;
  }
  else {
    e->next = (GanttEvent)hole->next;
    e->previous = hole->previous;
    GanttEvent_Free(RemoveHole(c, hole));
  }
  if(e->next == NULL)
    c->lastEvent = e;
  else
    e->next->previous = e;
  if(e->previous == NULL)
    c->firstEvent = e;
  else
    e->previous->next = e;
  priorEnd = e->previous == NULL ? 0 : e->previous->end;
  if(e->begin != priorEnd)
    AddHole(c, priorEnd, e->begin, e);
  if(e->next != NULL && e->end != e->next->begin)
    AddHole(c, e->end, e->next->begin, e->next);
}


unsigned long
GanttColumn_EarliestStartTime(GanttColumn c,
                              unsigned long earliestStart,
                              unsigned long runTime) {
  GanttEvent e = FindEventHole(c, earliestStart, runTime);
  unsigned long priorEnd =
    e != NULL ? e->begin : c->lastEvent == NULL ? 0 : c->lastEvent->end;
  return priorEnd > earliestStart ? priorEnd : earliestStart;
}


unsigned long
GanttColumn_GetEventCount(GanttColumn c) {
  GanttEvent e;
  unsigned long result;
  for(e = GanttColumn_GetFirstEvent(c), result = 0;
      e != NULL;
      e = GanttColumn_GetNextEvent(c, e), result++)
    ; /* empty */
  return result;
}


GanttEvent
GanttColumn_GetEventForDetails(GanttColumn c,
                               void *details) {
  GanttEvent e;
  for(e = c->firstEvent; e != NULL && e->details != details; e = e->next)
    ; /* empty */
  return e;
}


GanttEvent
GanttColumn_GetFirstEvent(GanttColumn c) {
  GanttEvent e;
  for(e = c->firstEvent; e != NULL && e->type == HOLE_TYPE; e = e->next)
    ; /* empty */
  return e;
}


void *
GanttColumn_GetKey(GanttColumn c) {
  return c->key;
}


GanttEvent
GanttColumn_GetNextEvent(GanttColumn c,
                         GanttEvent e) {
  for(e = e->next; e != NULL && e->type == HOLE_TYPE; e = e->next)
    ; /* empty */
  return e;
}


GanttEvent
GanttColumn_RemoveEvent(GanttColumn c,
                        GanttEvent e) {
  if(e->previous == NULL)
    c->firstEvent = e->next;
  else
    e->previous->next = e->next;
  if(e->next == NULL)
    c->lastEvent = e->previous;
  else
    e->next->previous = e->previous;
  if(e->previous != NULL && e->previous->type == HOLE_TYPE) {
    if(e->next == NULL)
      GanttEvent_Free(GanttColumn_RemoveEvent(c, RemoveHole(c, e->previous)));
    else if(e->next->type != HOLE_TYPE)
      e->previous->end = e->end;
    else {
      e->previous->end = e->next->end;
      GanttEvent_Free(GanttColumn_RemoveEvent(c, RemoveHole(c, e->next)));
    }
  }
  else if(e->next == NULL)
    ; /* empty */
  else if(e->next->type == HOLE_TYPE)
    e->next->begin = e->begin;
  else
    AddHole(c, e->begin, e->end, e->next);
  return e;
}


GanttColumn
GanttChart_AddColumn(GanttChart g,
                     void *key) {
  GanttColumn c = calloc(1, sizeof(struct _GanttColumn));
  c->key = key;
  Append(g->columns, c);
  return c;
}


void
GanttChart_Free(GanttChart g) {
  GanttColumn c;
  unsigned i;
  for(i = 0; i < Size(g->columns); i++) {
    c = g->columns[i];
    while(!GanttColumn_IsEmpty(c))
      GanttEvent_Free(GanttColumn_RemoveEvent(c, GanttColumn_GetFirstEvent(c)));
    free(c);
  }
  VectorFree(g->columns);
  free(g);
}


GanttColumn
GanttChart_GetColumn(GanttChart g,
                     unsigned n) {
  return n < Size(g->columns) ? g->columns[n] : NULL;
}


int
GanttChart_GetColumnCount(GanttChart g) {
  return Size(g->columns);
}


GanttColumn
GanttChart_GetColumnForDetails(GanttChart g,
                               void *details) {
  unsigned i;
  for(i = 0;
      i < Size(g->columns) &&
      GanttColumn_GetEventForDetails(g->columns[i], details) == NULL;
      i++)
    ; /* empty */
  return i < Size(g->columns) ? g->columns[i] : NULL;
}


GanttColumn
GanttChart_GetColumnForKey(GanttChart g,
                           void *key) {
  int i;
  for(i = 0; i < Size(g->columns) && g->columns[i]->key != key; i++)
    ; /* empty */
  return i < Size(g->columns) ? g->columns[i] : NULL;
}


int
GanttChart_IsEmpty(GanttChart g) {
  unsigned i;
  for(i = 0; i < Size(g->columns) && GanttColumn_IsEmpty(g->columns[i]); i++)
    ; /* empty */
  return i == Size(g->columns);
}


GanttChart
GanttChart_New(void) {
  GanttChart result = (GanttChart)calloc(1, sizeof(struct _GanttChart));
  result->columns = VectorNew(GanttColumn);
  return result;
}
