/* $Id: global.c,v 1.90 2004/04/01 16:56:09 jhayes Exp $ */
/*
 * Copyright  2002 The Regents of the University of California. 
 * All Rights Reserved. 
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes, without
 * fee, and without a written agreement is hereby granted, provided that the
 * above copyright notice, this paragraph and the following three paragraphs
 * appear in all copies. 
 *
 * Permission to incorporate this software into commercial products may be
 * obtained by contacting
 * Technology Transfer Office 
 * 9500 Gilman Drive 
 * 411 University Center 
 * University of California 
 * La Jolla, CA 92093-0093
 * (858) 534-5815
 * invent@ucsd.edu
 *
 * This software program and documentation are copyrighted by The Regents of
 * the University of California. The software program and documentation are
 * supplied "as is", without any accompanying services from The Regents. The
 * Regents does not warrant that the operation of the program will be
 * uninterrupted or error-free. The end-user understands that the program was
 * developed for research purposes and is advised not to rely exclusively on
 * the program for any reason. 
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
 * EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED
 * HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO
 * OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS. 
 */


#include "config.h"
#include <sys/types.h>
#include <stdio.h>    /* file functions */
#include <stdlib.h>   /* strtol */
#include <string.h>   /* string functions */
#include <time.h>     /* ctime time */
#define ASH_SHORT_NAMES
#include <hashseed.h>
#define ASSTR_SHORT_NAMES
#include <strseed.h>
#define ASV_SHORT_NAMES
#include <vectorseed.h>
#include <xmlseed.h>
#include "corestructures.h"
#include "compute.h"
#include "misc.h"
#include "scheduler.h"
#include "storage.h"
#include "global.h"


typedef enum {
  HOST_DISABLED, HOST_ENABLED, TASK_CANCELED, TASK_DONE,
  TASK_PRIORITIZED, TASK_RESET, TASK_STARTED, TRANSFER_CANCELED,
  TRANSFER_DONE, TRANSFER_RESET, TRANSFER_STARTED
} EventTypes;
static const char *EVENT_TYPE_IMAGES[] = {
  "host_disabled", "host_enabled", "task_canceled", "task_done",
  "task_prioritized", "task_reset", "task_started", "transfer_canceled",
  "transfer_done", "transfer_reset", "transfer_started", NULL
};


/* Garbage collect callback for hash table entries. */
static void
FreeValue(KeyType key,
          ValueType value) {
  FREE(value);
}


/*
 * Returns the batch submission host that corresponds to the batch node host
 * #id#, NULL if none.  NOTE: this function uses implicit knowledge of the
 * naming convention that the compute module uses for batch nodes.
 */
static Host
GetSubmitHost(const char *id) {
  Host result;
  char *submitId;
  if(strrchr(id, '-') == NULL)
    return NULL;
  submitId = STRDUP(id);
  *strrchr(submitId, '-') = '\0';
  result = HostList_Get(global.hosts, submitId);
  FREE(submitId);
  return result;
}


/*
 * Appends an #e# event to the checkpoint file.  #id# is the id of the event
 * target, #param# the optional event parameter.
 */
static void
RecordEvent(EventTypes e,
            const char *id,
            char *param) {

  static const char *END_TAG = "</log></apst>\n";
  static const int END_TAG_LEN = 14;
  char *tag;
  char timeImage[30];

  if(global.checkpointFile == NULL)
    return; /* TODO Write something to log? */
  if(ftell(global.checkpointFile) == 0)
    fprintf(global.checkpointFile, "%s", "<apst><log>\n");
  else
    fseek(global.checkpointFile, -END_TAG_LEN, SEEK_END);
  sprintf(timeImage, "%ld", (long)time(NULL));
  tag = StrAppend
    (NULL, "<event when='", timeImage, "' type='", EVENT_TYPE_IMAGES[e],
     "' which='", id, "'", NULL);
  if(param != NULL)
    tag = StrAppend(tag, " param='", param, "'", NULL);
  tag = StrAppend(tag, "/>", NULL);
  fprintf(global.checkpointFile, "%s\n%s", tag, END_TAG);
  FREE(tag);
  fflush(global.checkpointFile);

}


/*
 * Common steps taken when a task stops for any reason.  Removes information
 * about #t# from global tables and peers; does not change #t#.
 */
static void
TaskStopped(Task t) {
  unsigned i;
  TaskGroupRemove("canceled", t);
  TaskGroupRemove("done", t);
  TaskGroupRemove("failed", t);
  TaskGroupRemove("new", t);
  TaskGroupRemove("running", t);
  if(t->stat == ACTION_IN_PROGRESS) {
    for(i = 0; i < Size(t->host->tasksInProgress); i++) {
      if(t->host->tasksInProgress[i] == t)
        Remove(t->host->tasksInProgress, i);
    }
  }
}


/*
 * Common steps taken when a file transfer stops for any reason.  Removes
 * information about #ft# from global tables and peers; does not change #ft#.
 */
static void
TransferStopped(FileTransfer ft) {
  unsigned i;
  FileTransferList_Remove(ft->stat==ACTION_CANCELED ? global.transfersCanceled :
                          ft->stat==ACTION_DONE ? global.transfersDone :
                          ft->stat==ACTION_FAILED ? global.transfersFailed :
                          global.transfersInProgress, ft);
  if(ft->stat == ACTION_IN_PROGRESS) {
    for(i = 0; i < Size(ft->source->transfersInProgress); i++) {
      if(ft->source->transfersInProgress[i] == ft)
        Remove(ft->source->transfersInProgress, i);
    }
    for(i = 0; i < Size(ft->dest->transfersInProgress); i++) {
      if(ft->dest->transfersInProgress[i] == ft)
        Remove(ft->dest->transfersInProgress, i);
    }
  }
}


void
FileTransferCanceled(FileTransfer ft) {
  if(ft->stat == ACTION_CANCELED)
    return;
  else if(ft->stat == ACTION_DONE) {
    ERROR1("Transfer '%s' already finished\n", ft->id);
    return;
  }
  else if(ft->stat == ACTION_FAILED) {
    ERROR1("Transfer '%s' already failed\n", ft->id);
    return;
  }
  else if(ft->stat == ACTION_IN_PROGRESS)
    storage_cancel_transfer(ft);
  TransferStopped(ft);
  ft->stat = ACTION_CANCELED;
  FileTransferList_Add(global.transfersCanceled, ft);
  RecordEvent(TRANSFER_CANCELED, ft->id, NULL);
}


void
FileTransferDone(FileTransfer ft,
                 int exitCode) {
  char codeImage[20];
  ft->meta.timeCompleted = time(NULL);
  TransferStopped(ft);
  if(exitCode == 0) {
    ft->stat = ACTION_DONE;
    FileTransferList_Add(global.transfersDone, ft);
    Append(ft->source->transfersCompleted, ft);
    Append(ft->dest->transfersCompleted, ft);
    NewFileCopy(ft->dest, ft->file);
    sched_file_transfer_done(ft, 1);
  }
  else {
    ft->stat = ACTION_FAILED;
    FileTransferList_Add(global.transfersFailed, ft);
    sched_file_transfer_done(ft, 0);
  }
  sprintf(codeImage, "%d", exitCode);
  RecordEvent(TRANSFER_DONE, ft->id, codeImage);
}


void
FileTransferReset(FileTransfer ft) {
  if(ft->stat == ACTION_NEW)
    return;
  if(ft->stat == ACTION_IN_PROGRESS)
    storage_cancel_transfer(ft);
  TransferStopped(ft);
  ft->stat = ACTION_NEW;
  RecordEvent(TRANSFER_RESET, ft->id, NULL);
}


void
FileTransferStarted(FileTransfer ft) {
  FileTransferList_Add(global.transfersInProgress, ft);
  ft->meta.timeCompleted = -1;
  ft->meta.timeStarted = time(NULL);
  ft->stat = ACTION_IN_PROGRESS;
  Append(ft->source->transfersInProgress, ft);
  Append(ft->dest->transfersInProgress, ft);
  RecordEvent(TRANSFER_STARTED, ft->id, NULL);
}


Disk
FindLocalDisk(void) {
  Disk d;
  HashIterator iterator;
  ForEachMapping(global.disks, iterator) {
    Disk d = (Disk)IteratorValue(iterator);
    if(d->method == ELBASE_CP)
      return d;
  }
  d = Disk_New(STRDUP("local"), STRDUP("."), ELBASE_CP, STRDUP(""));
  DiskList_Add(global.disks, d);
  return d;
}


char *
HashToXML(HashTable hash,
          const char *section,
          XmlFn fn) {

  HashIterator it;
  char *contents;
  unsigned i;
  unsigned len;
  char *result;
  char *sectionClose;
  char *sectionOpen;
  char **xml;

  sectionOpen =
    section == NULL ? "" : StrAppend(NULL, "<", section, ">", NULL);
  if(GetMappingCount(hash) == 0)
    contents = "";
  else {
    xml = (char **)MALLOC((GetMappingCount(hash) + 1) * sizeof(char *));
    i = 0;
    len = 0;
    ForEachMapping(hash, it) {
      xml[i] = fn(IteratorValue(it));
      len += strlen(xml[i]);
      i++;
    }
    xml[i] = NULL;
    contents = StrArrayJoin((const char **)xml, NULL);
    StrArrayFree(xml);
  }
  sectionClose =
    section == NULL ? "" : StrAppend(NULL, "</", section, ">", NULL);
  result = StrAppend(NULL, sectionOpen, contents, sectionClose, NULL);
  if(*sectionOpen != '\0')
    FREE(sectionOpen);
  if(*contents != '\0')
    FREE(contents);
  if(*sectionClose != '\0')
    FREE(sectionClose);
  return result;

}


char *
GetStateXML(SequenceNumber *tasks,
            unsigned tasksCount) {

  unsigned i;
  unsigned len;
  char *result;
  char **taskXml;
  Task t;
  char *xml;

  result = STRDUP("<apst>");
  if(GetMappingCount(global.disks) > 0) {
    xml = HashToXML(global.disks, "storage", (XmlFn)&storage_disk_to_xml);
    result = StrAppend(result, xml, NULL);
    FREE(xml);
  }
  if(GetMappingCount(global.files) > 0) {
    xml = HashToXML(global.files, "files", (XmlFn)&storage_file_to_xml);
    result = StrAppend(result, xml, NULL);
    FREE(xml);
  }
  if(GetMappingCount(global.hosts) > 0) {
    xml = HashToXML(global.hosts, "compute", (XmlFn)&compute_host_to_xml);
    result = StrAppend(result, xml, NULL);
    FREE(xml);
  }
  if(tasksCount > 0) {
    taskXml = (char **)MALLOC((tasksCount + 1) * sizeof(char *));
    for(i = 0, len = 0; i < tasksCount; i++)
      if((t = TaskList_Get(global.tasks, tasks[i])) != NULL)
        taskXml[len++] = compute_task_to_xml(t);
    taskXml[i] = NULL;
    xml = StrArrayJoin((const char **)taskXml, NULL);
    result = StrAppend(result, "<tasks>", xml, "</tasks>", NULL);
    FREE(xml);
    StrArrayFree(taskXml);
  }
  result = StrAppend(result, "</apst>", NULL);
  return result;

}


void
GlobalDeinit(void) {
  HashIterator it;
  DiskList_Free(global.disks, 1);
  FileList_Free(global.files, 1);
  GridInfoSourceList_Free(global.gridInfoSources, 1);
  HostList_Free(global.hosts, 1);
  ForEachMapping(global.taskGroups, it) {
    TaskList_Free((TaskList)IteratorValue(it), 0);
    FREE(IteratorKey(it));
  }
  HashTableFree(global.taskGroups);
  TaskList_Free(global.tasks, 1);
  FileTransferList_Free(global.transfersCanceled, 0);
  FileTransferList_Free(global.transfersDone, 0);
  FileTransferList_Free(global.transfersFailed, 0);
  FileTransferList_Free(global.transfersInProgress, 0);
  FileTransferList_Free(global.transfers, 1);
  HashTableFree(global.stdins);
  HashTableFree(global.stdouts);
  HashTableFree(global.stderrs);
  HashTableFree(global.userIds);
  HashTableFree(global.wds);
}


void
GlobalInit(void) {
  global.disks = DiskList_New(SMALL_HASH);
  global.files = FileList_New(LARGE_HASH);
  global.gridInfoSources = GridInfoSourceList_New(MEDIUM_HASH);
  global.hosts = HostList_New(MEDIUM_HASH);
  global.tasks = TaskList_New(LARGE_HASH);
  global.taskGroups =
    HashTableNew(SMALL_HASH, &ASH_StringCompare, &ASH_StringHash, NULL);
  PutMapping
    (global.taskGroups, STRDUP("canceled"), TaskList_New(SMALL_HASH));
  PutMapping(global.taskGroups, STRDUP("done"), TaskList_New(LARGE_HASH));
  PutMapping(global.taskGroups, STRDUP("failed"), TaskList_New(SMALL_HASH));
  PutMapping(global.taskGroups, STRDUP("new"), TaskList_New(LARGE_HASH));
  PutMapping
    (global.taskGroups, STRDUP("running"), TaskList_New(MEDIUM_HASH));
  global.transfers = FileTransferList_New(MEDIUM_HASH);
  global.transfersCanceled = FileTransferList_New(LINKED_LIST);
  global.transfersDone = FileTransferList_New(MEDIUM_HASH);
  global.transfersFailed = FileTransferList_New(LINKED_LIST);
  global.transfersInProgress = FileTransferList_New(SMALL_HASH);
  global.stdins = HashTableNew(SMALL_HASH, NULL, NULL, NULL);
  global.stdouts = HashTableNew(SMALL_HASH, NULL, NULL, NULL);
  global.stderrs = HashTableNew(SMALL_HASH, NULL, NULL, NULL);
  global.userIds = HashTableNew(SMALL_HASH, NULL, NULL, &FreeValue);
  global.wds = HashTableNew(SMALL_HASH, NULL, NULL, &FreeValue);
}


void
HostDisabled(Host h) {
  if(h->meta.timeDisabled != 0)
    return;
  while(Size(h->tasksInProgress) > 0)
    TaskReset(*h->tasksInProgress);
  h->meta.timeDisabled = (long)time(NULL);
  sched_host_failed(h);
  RecordEvent(HOST_DISABLED, h->id, NULL);
}


void
HostEnabled(Host h) {
#if 0
  /*
   * NOTE: removed this so that the scheduler is made aware of newly-arrived
   * batch hosts (bug #347).
   */
  if(h->meta.timeDisabled == 0)
    return;
#endif
  h->meta.timeDisabled = 0;
  sched_host_available(h);
  RecordEvent(HOST_ENABLED, h->id, NULL);
}


void
NewFileCopy(Disk d,
            File f) {
  if(FileCopyList_Get(f->copies, d->id) == NULL)
    FileCopyList_Add
      (f->copies, FileCopy_New(d, AbsolutePath(f->name, d->datadir)));
}


int
ProcessLogXml(ASXML_XmlNode node) {

  char *attrId;
  char *attrParam;
  char *attrType;
  char *attrWhen;
  ASXML_XmlNode child;
  EventTypes e;
  Host h;
  unsigned long i;
  HashIterator it;
  int okay;
  int result = 1;
  TaskList startedTasks = TaskList_New(SMALL_HASH);
  FileTransferList startedTransfers = FileTransferList_New(SMALL_HASH);
  unsigned long when;

  for(child = ASXML_NodeFirstChildByType(node, ASXML_SHOW_ELEMENT);
      child != NULL;
      child = ASXML_NodeNextSiblingByType(child, ASXML_SHOW_ELEMENT)) {
    if(strcmp(ASXML_NodeName(child), "event") == 0) {
      okay = 0;
      if(*(attrType = ASXML_ElementGetAttribute(child, "type")) == '\0')
        WARN("'event' tag requires a 'type' atttribute\n");
      else if(!StrArrayFind(EVENT_TYPE_IMAGES, attrType, &i))
        WARN1("Unknown event type '%s'\n", attrType);
      else if(*(attrWhen = ASXML_ElementGetAttribute(child, "when")) == '\0')
        WARN("'event' tag requires a 'when' atttribute\n");
      else if(*(attrId = ASXML_ElementGetAttribute(child, "which")) == '\0')
        WARN("'event' tag requires a 'which' atttribute\n");
      else
        okay = 1;
      if(!okay) {
        result = 0;
        continue;
      }
      attrParam = ASXML_ElementGetAttribute(child, "param");
      e = (EventTypes)i;
      when = (unsigned long)strtol(attrWhen, NULL, 10);
      if(e == HOST_DISABLED || e == HOST_ENABLED) {
        h = HostList_Get(global.hosts, attrId);
        if(h == NULL) {
          if(GetSubmitHost(attrId) == NULL) {
            WARN1("Unknown host '%s'\n", attrId);
            result = 0;
          }
          continue;
        }
        if(e == HOST_DISABLED) {
          HostDisabled(h);
          h->meta.timeDisabled = when;
        }
        else
          HostEnabled(h);
      }
      else if(e == TASK_CANCELED || e == TASK_DONE || e == TASK_PRIORITIZED ||
              e == TASK_RESET || e == TASK_STARTED) {
        Task t = TaskList_Get(global.tasks, strtol(attrId, NULL, 10));
        if(t == NULL) {
          WARN1("Unknown task '%s'\n", attrId);
          result = 0;
          continue;
        }
        if(e == TASK_CANCELED)
          TaskCanceled(t);
        else if(e == TASK_DONE) {
          TaskDone(t, strtol(attrParam, NULL, 10));
          t->meta.timeCompleted = when;
        }
        else if(e == TASK_PRIORITIZED)
          TaskPrioritized(t, strtol(attrParam, NULL, 10));
        else if(e == TASK_RESET)
          TaskReset(t);
        else if(e == TASK_STARTED) {
          if((h = HostList_Get(global.hosts, attrParam)) == NULL &&
             (h = GetSubmitHost(attrParam)) == NULL) {
            WARN1("Unknown host '%s'\n", attrParam);
            result = 0;
            continue;
          }
          TaskStarted(t, h);
          TaskList_Add(startedTasks, t);
          t->meta.timeStarted = when;
        }
      }
      else {
        FileTransfer ft = FileTransferList_Get(global.transfers, attrId);
        if(ft == NULL) {
          /* NOTE: implicit knowledge of ft id format from corestructures.c. */
          Disk dest;
          File f;
          char **pieces = ASSTR_StrSplit(attrId, NULL);
          Disk source;
          if(pieces[0] == NULL || pieces[1] == NULL || pieces[2] == NULL)
            WARN1("Bad transfer id '%s'\n", attrId);
          else if((f = FileList_Get(global.files, pieces[0])) == NULL)
            WARN1("Unknown file '%s'\n", pieces[0]);
          else if((source = DiskList_Get(global.disks, pieces[1])) == NULL)
            WARN1("Unknown disk '%s'\n", pieces[1]);
          else if((dest = DiskList_Get(global.disks, pieces[2])) == NULL)
            WARN1("Unknown disk '%s'\n", pieces[2]);
          else {
            ft = FileTransfer_New(f, source, dest);
            FileTransferList_Add(global.transfers, ft);
          }
          StrArrayFree(pieces);
        }
        if(ft == NULL) {
          result = 0;
          continue;
        }
        if(e == TRANSFER_CANCELED)
          FileTransferCanceled(ft);
        else if(e == TRANSFER_DONE) {
          FileTransferDone(ft, strtol(attrParam, NULL, 10));
          ft->meta.timeCompleted = when;
        }
        else if(e == TRANSFER_RESET)
          FileTransferReset(ft);
        else if(e == TRANSFER_STARTED) {
          FileTransferStarted(ft);
          FileTransferList_Add(startedTransfers, ft);
          ft->meta.timeStarted = when;
        }
      }
    }
    else {
      WARN1("Unknown log subtag '%s'\n", ASXML_NodeName(child));
      result = 0;
    }
  }
  ForEachMapping(startedTasks, it) {
    Task t = (Task)IteratorValue(it);
    if(t->stat == ACTION_IN_PROGRESS)
      TaskCanceled(t);
  }
  TaskList_Free(startedTasks, 0);
  ForEachMapping(startedTransfers, it) {
    FileTransfer ft = (FileTransfer)IteratorValue(it);
    if(ft->stat == ACTION_IN_PROGRESS)
      FileTransferCanceled(ft);
  }
  FileTransferList_Free(startedTransfers, 0);
  sched_reschedule();
  return result;
}


void
TaskCanceled(Task t) {
  char seqImage[20];
  if(t->stat == ACTION_CANCELED)
    return;
  else if(t->stat == ACTION_DONE) {
    ERROR1("Task %ld already finished\n", t->seq);
    return;
  }
  else if(t->stat == ACTION_FAILED) {
    ERROR1("Task %ld already failed\n", t->seq);
    return;
  }
  else if(t->stat == ACTION_IN_PROGRESS)
    compute_cancel_task(t);
  TaskStopped(t);
  t->stat = ACTION_CANCELED;
  TaskGroupAdd("canceled", t);
  sched_task_canceled(t);
  sprintf(seqImage, "%ld", (unsigned long)t->seq);
  RecordEvent(TASK_CANCELED, seqImage, NULL);
}


void
TaskDone(Task t,
         int exitCode) {
  char codeImage[20];
  Disk disk = t->host->disk;
  unsigned i;
  char seqImage[20];
  t->meta.timeCompleted = time(NULL);
  TaskStopped(t);
  if(exitCode == 0) {
    t->stat = ACTION_DONE;
    TaskGroupAdd("done", t);
    Append(t->host->tasksCompleted, t);
    for(i = 0; i < Size(t->output); i++)
      NewFileCopy(disk, t->output[i]);
    sched_task_done(t, 1);
  }
  else {
    t->stat = ACTION_FAILED;
    TaskGroupAdd("failed", t);
    sched_task_done(t, 0);
  }
  sprintf(codeImage, "%d", exitCode);
  sprintf(seqImage, "%ld", (unsigned long)t->seq);
  RecordEvent(TASK_DONE, seqImage, codeImage);
  if(TaskGroupMembership("new") == 0 && TaskGroupMembership("running") == 0) {
    char body[512];
    time_t now = time(NULL);
    sprintf(body, "%s\tAll tasks completed\n\n\t\tAPST daemon\n", ctime(&now));
    SendEmailNotifications("automatic APST e-mail notification", body);
  }
}


void
TaskPrioritized(Task t,
                unsigned priority) {
  char priorityImage[20];
  char seqImage[20];
  if(t->priority == priority)
    return;
  t->priority = priority;
  if(t->stat == ACTION_NEW)
    sched_reschedule();
  sprintf(priorityImage, "%d", priority);
  sprintf(seqImage, "%ld", (unsigned long)t->seq);
  RecordEvent(TASK_PRIORITIZED, seqImage, priorityImage);
}


void
TaskReset(Task t) {
  char seqImage[20];
  if(t->stat == ACTION_NEW)
    return;
  if(t->stat == ACTION_IN_PROGRESS)
    compute_cancel_task(t);
  TaskStopped(t);
  t->host = NULL;
  t->stat = ACTION_NEW;
  TaskGroupAdd("new", t);
  sched_new_task(t);
  sprintf(seqImage, "%ld", (unsigned long)t->seq);
  RecordEvent(TASK_RESET, seqImage, NULL);
}


void
TaskStarted(Task t,
            Host h) {
  char seqImage[20];
  TaskGroupRemove("new", t);
  TaskGroupAdd("running", t);
  t->meta.timeCompleted = -1;
  t->meta.timeStarted = time(NULL);
  t->stat = ACTION_IN_PROGRESS;
  t->host = h;
  Append(h->tasksInProgress, t);
  sprintf(seqImage, "%ld", (unsigned long)t->seq);
  RecordEvent(TASK_STARTED, seqImage, h->id);
}


void
TaskGroupAdd(const char *name,
             Task t) {
  TaskList tl = (TaskList)GetValue(global.taskGroups, (char *)name, NULL);
  if(tl == NULL) {
    tl = TaskList_New(SMALL_HASH);
    PutMapping(global.taskGroups, STRDUP(name), tl);
  }
  TaskList_Add(tl, t);
}


unsigned long
TaskGroupMembership(const char *name) {
  TaskList tl = (TaskList)GetValue(global.taskGroups, (char *)name, NULL);
  return tl == NULL ? 0 : GetMappingCount(tl);
}


void
TaskGroupRemove(const char *name,
                Task t) {
  TaskList tl = (TaskList)GetValue(global.taskGroups, (char *)name, NULL);
  if(tl != NULL)
    TaskList_Remove(tl, t);
}
