/* $Id: storage.c,v 1.78 2003/12/01 17:45:02 jhayes Exp $ */
/*
 * Copyright  2002 The Regents of the University of California. 
 * All Rights Reserved. 
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes, without
 * fee, and without a written agreement is hereby granted, provided that the
 * above copyright notice, this paragraph and the following three paragraphs
 * appear in all copies. 
 *
 * Permission to incorporate this software into commercial products may be
 * obtained by contacting
 * Technology Transfer Office 
 * 9500 Gilman Drive 
 * 411 University Center 
 * University of California 
 * La Jolla, CA 92093-0093
 * (858) 534-5815
 * invent@ucsd.edu
 *
 * This software program and documentation are copyrighted by The Regents of
 * the University of California. The software program and documentation are
 * supplied "as is", without any accompanying services from The Regents. The
 * Regents does not warrant that the operation of the program will be
 * uninterrupted or error-free. The end-user understands that the program was
 * developed for research purposes and is advised not to rely exclusively on
 * the program for any reason. 
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
 * LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
 * EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED
 * HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO
 * OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS. 
 */


#include "config.h"
#include "errno.h"
#include <sys/types.h>
#include <sys/stat.h> /* mkdir */
#include <stdio.h>    /* file functions */
#include <stdlib.h>   /* strtod */
#include <string.h>   /* string functions */
#include <time.h>     /* time */
#include <unistd.h>   /* getcwd */
#include <fileseed.h>
#define ASH_SHORT_NAMES
#include <hashseed.h>
#define ASSTR_SHORT_NAMES
#include <strseed.h>
#define ASV_SHORT_NAMES
#include <vectorseed.h>
#include <bulkcopy.h>
#include "corestructures.h"
#include "global.h"
#include "misc.h"
#include "storage.h"


/* Info retained for each active set of file transfers. */
typedef struct _TransferInfo {
  FileTransfer ft;
  FileList files;     /* List of files for bulk transfers. */
  ELBASE_Process pid; /* Local process id; NULL if simulated. */
  int terminated;     /* Done? ... */
  int exitValue;      /* ... exit value if so. */
  int canceled;       /* Did we cancel it? */
  int stderrKey;      /* stderr redirection key. */
} *TransferInfo;


static TransferInfo *pendingTransfers;
static ELBASE_Process SIMULATED_PROCESS_PID = NULL;
static const char *METHOD_IMAGES[ELBASE_STORAGE_SERVICE_COUNT + 1] = {
  "local", "ftp", "gass", "scp", "sftp", "srb", NULL
};


/* Returns XML that represents the info in #fc#. */
static char *
CopyToXML(FileCopy fc) {
  return StrAppend
    (NULL, "<copy disk='", fc->disk->id, "' path='", fc->path, "'", "/>", NULL);
}


/* Returns a disk containing the information under #node#, NULL on error. */
static Disk
ProcessDiskXml(ASXML_XmlNode node) {

  char *datadir;
  char defaultDir[256];
  unsigned long i;
  const char *id;
  ELBASE_StorageServiceTypes method;
  ASXML_XmlNode methodNode;
  const char *server;

  if(*(id = ASXML_ElementGetAttribute(node, "id")) == '\0')
    FAIL("'disk' tag requires an 'id' attribute\n");
  if(DiskList_Get(global.disks, (KeyType)id) != NULL)
    FAIL1("Duplicate definition for disk '%s'\n", id);
  methodNode = ASXML_NodeFirstChildByType(node, ASXML_SHOW_ELEMENT);
  if(methodNode == NULL)
    method = ELBASE_CP;
  else if(!StrArrayFind(METHOD_IMAGES, ASXML_NodeName(methodNode), &i))
    FAIL1("Unknown disk subtag '%s'\n", ASXML_NodeName(methodNode));
  else
    method = (ELBASE_StorageServiceTypes)i;
  if(method == ELBASE_CP)
    getcwd(defaultDir, sizeof(defaultDir));
  else
    strcpy(defaultDir, ".");
  datadir = ASXML_ElementGetAttribute(node, "datadir");
  datadir = *datadir == '\0' ?
            STRDUP(defaultDir) : AbsolutePath(datadir, defaultDir);
  if(method == ELBASE_CP)
    server = "";
  else if(*(server = ASXML_ElementGetAttribute(methodNode, "server")) == '\0')
    server = id;

  return Disk_New(STRDUP(id), datadir, method, STRDUP(server));

}


/* Returns a file containing the information under #node#, NULL on error. */
static File
ProcessFileXml(ASXML_XmlNode node) {

  const char *attribute;
  ASXML_XmlNode child;
  FileCopy copy;
  char *end;
  Disk disk;
  const char *id;
  char *path;
  File result;

  if(*(id = ASXML_ElementGetAttribute(node, "id")) == '\0')
    FAIL("'file' tag requires an 'id' attribute\n");
  if(FileList_Get(global.files, (char *)id) != NULL)
    FAIL1("Duplicate definition for file '%s'\n", id);
  result = File_New(STRDUP(id));
  result->download = *ASXML_ElementGetAttribute(node, "download") != 'n';
  if(*(attribute = ASXML_ElementGetAttribute(node, "size")) != '\0') {
    result->size = strtod(attribute, &end);
    if(*end == 'k' || *end == 'K')
      result->size *= 1000;
    else if(*end == 'm' || *end == 'M')
      result->size *= 1000000;
    else if(*end == 'g' || *end == 'G')
      result->size *= 1000000000;
  }
  result->transferable = *ASXML_ElementGetAttribute(node, "transfer") != 'n';
  for(child = ASXML_NodeFirstChildByType(node, ASXML_SHOW_ELEMENT);
      child != NULL;
      child = ASXML_NodeNextSiblingByType(child, ASXML_SHOW_ELEMENT)) {
    if(strcmp(ASXML_NodeName(child), "copy") == 0) {
      if(*(attribute = ASXML_ElementGetAttribute(child, "disk")) == '\0')
        disk = FindLocalDisk();
      else if((disk = DiskList_Get(global.disks, (char *)attribute)) == NULL) {
        File_Free(result);
        FAIL1("Unknown disk %s\n", attribute);
      }
      if(FileCopyList_Get(result->copies, disk->id) != NULL)
        continue;
      if(*(path = ASXML_ElementGetAttribute(child, "path")) == '\0')
        path = result->name;
      path = AbsolutePath(path, disk->datadir);
      copy = FileCopy_New(disk, path);
      FileCopyList_Add(result->copies, copy);
    }
    else {
      File_Free(result);
      FAIL1("Unknown file subtag '%s'\n", ASXML_NodeName(child));
    }
  }
  return result;

}


int
storage_cancel_transfer(FileTransfer ft) {
  unsigned i;
  for(i = 0; i < Size(pendingTransfers) && pendingTransfers[i]->ft != ft; i++)
    ; /* empty */
  if(i == Size(pendingTransfers))
    FAIL1("storage_cancel_transfer: Cannot find transfer '%s'\n", ft->id);
  if(pendingTransfers[i]->pid != SIMULATED_PROCESS_PID &&
     !ELBASE_Kill(pendingTransfers[i]->pid))
    WARN2("storage_cancel_transfer: Cannot kill pid %d of transfer '%s'\n",
          pendingTransfers[i]->pid, ft->id);
  pendingTransfers[i]->canceled = 1;
  return 1;
}


int
storage_do_bulk_transfer(FileTransferList ftl) {

  Disk destDisk;
  char *destPath;
  FileTransfer ft;
  TransferInfo info;
  HashIterator it;
  ELBC_BulkLink link;
  Disk sourceDisk;
  char *sourcePath;

  ForEachMapping(ftl, it) {
    ft = (FileTransfer)IteratorValue(it);
    if(FileCopyList_Get(ft->file->copies, ft->source->id) == NULL)
      FAIL2("No copy of %s available on %s\n", ft->file->name, ft->source->id);
  }

  if(GetMappingCount(ftl) < 2)
    link = NULL; /* Bulk transfer unnecessary. */
  else {
    ft = (FileTransfer)IteratorValue(IteratorFirst(ftl));
    destDisk = ft->dest;
    sourceDisk = ft->source;
    ForEachMapping(ftl, it) {
      ft = (FileTransfer)IteratorValue(it);
      if(ft->source != sourceDisk || ft->dest != destDisk)
        break;
    }
    if(it.valid)
     link = NULL; /* ftl contains mixed links. */
    else {
      ELBASE_ComputeServiceTypes destService;
      char *destServer = NULL;
      ELBASE_ComputeServiceTypes sourceService;
      char *sourceServer = NULL;
      if(sourceDisk->method == ELBASE_CP) {
        sourceService = ELBASE_FORK;
        sourceServer = "localhost";
      }
      else if(sourceDisk->method == ELBASE_SCP) {
        sourceService = ELBASE_SSH;
        sourceServer = sourceDisk->server;
      }
      else {
        ForEachMapping(global.hosts, it) {
          Host h = (Host)IteratorValue(it);
          if(h->disk == sourceDisk && h->method != ELBASE_RASH) {
            sourceService = h->method;
            sourceServer = h->server;
            break;
          }
        }
      }
      if(destDisk->method == ELBASE_CP) {
        destService = ELBASE_FORK;
        destServer = "localhost";
      }
      else if(destDisk->method == ELBASE_SCP) {
        destService = ELBASE_SSH;
        destServer = destDisk->server;
      }
      else {
        ForEachMapping(global.hosts, it) {
          Host h = (Host)IteratorValue(it);
          if(h->disk == destDisk && h->method != ELBASE_RASH) {
            destService = h->method;
            destServer = h->server;
            break;
          }
        }
      }
      if(sourceServer == NULL || destServer == NULL)
        link = NULL; /* No compute host for disk. */
      else
        link = ELBC_BulkLinkNew
          (sourceDisk->method, sourceDisk->server, sourceService, sourceServer,
           destDisk->method, destDisk->server, destService, destServer);
    }
  }

  if(link == NULL) {

    /* Bulk transfer not supported; do each transfer individually. */
    ForEachMapping(ftl, it) {
      ft = (FileTransfer)IteratorValue(it);
      info = (TransferInfo)MALLOC(sizeof(struct _TransferInfo));
      info->ft = ft;
      info->files = NULL;
      info->terminated = 0;
      info->canceled = 0;
      info->stderrKey = StderrCaptureBegin();
      if(global.simDisk >= 0)
        info->pid = SIMULATED_PROCESS_PID;
      else {
        sourcePath = FileCopyList_Get(ft->file->copies, ft->source->id)->path;
        destPath = AbsolutePath(ft->file->name, ft->dest->datadir);
        if(!ELBASE_CopyFile(ft->source->method, ft->source->server, sourcePath,
                            ft->dest->method, ft->dest->server, destPath,
                            &info->pid)) {
          info->terminated = 1;
          info->exitValue = 1;
        }
        FREE(destPath);
      }
      Append(pendingTransfers, info);
      StderrCaptureEnd(info->stderrKey);
      FileTransferStarted(ft);
    }

  }
  else {

    File bulkFile;
    FileTransfer bulkFt;
    const char **destPaths;
    unsigned i;
    const char **sourcePaths;

    /* Create a pseudo-File to hold the bulk data and a FileTransfer for it. */
    bulkFile = File_New(ASFILE_UniqueFilePath("apst_bulk"));
    ForEachMapping(ftl, it)
      bulkFile->size += ((FileTransfer)IteratorValue(it))->file->size;
    FileList_Add(global.files, bulkFile);
    LOG4("Bulk transfer %d files from %s to %s via %s\n",
         GetMappingCount(ftl), sourceDisk->id, destDisk->id, bulkFile->name);
    bulkFt = FileTransfer_New(bulkFile, sourceDisk, destDisk);

    destPaths = VectorNew(const char *);
    sourcePaths = VectorNew(const char *);
    info = (TransferInfo)MALLOC(sizeof(struct _TransferInfo));
    info->ft = bulkFt;
    info->files = FileList_New(LINKED_LIST);
    ForEachMapping(ftl, it) {
      File f = ((FileTransfer)IteratorValue(it))->file;
      Append(sourcePaths, FileCopyList_Get(f->copies, sourceDisk->id)->path);
      Append(destPaths, AbsolutePath(f->name, destDisk->datadir));
      FileList_Add(info->files, f);
    }
    info->terminated = 0;
    info->canceled = 0;
    info->stderrKey = StderrCaptureBegin();

    if(global.simDisk >= 0)
      info->pid = SIMULATED_PROCESS_PID;
    else {
      if(!ELBC_BulkCopy(link, sourcePaths, destPaths, Size(sourcePaths),
         &info->pid)) {
        info->terminated = 1;
        info->exitValue = 1;
      }
    }

    Append(pendingTransfers, info);
    StderrCaptureEnd(info->stderrKey);
    FileTransferList_Add(global.transfers, bulkFt);
    FileTransferStarted(bulkFt);
    ELBC_BulkLinkFree(link);
    for(i = 0; i < Size(destPaths); i++)
      FREE((char *)destPaths[i]);
    VectorFree(sourcePaths);
    VectorFree(destPaths);

  }
  
  return 1;

}


int
storage_do_file_transfer(FileTransfer ft) {
  FileTransferList ftl;
  int result;
  ftl = FileTransferList_New(LINKED_LIST);
  FileTransferList_Add(ftl, ft);
  result = storage_do_bulk_transfer(ftl);
  FileTransferList_Free(ftl, 0);
  return result;
}


int
storage_init(void) {
  pendingTransfers = VectorNew(TransferInfo);
  return 1;
}


int
storage_poll(void) {
  char *c;
  int i;
  TransferInfo info;
  HashIterator it;
  for(i = Size(pendingTransfers) - 1; i >= 0; i--) {
    info = pendingTransfers[i];
    if(info->pid == SIMULATED_PROCESS_PID) {
      long runTime = !global.simRandom ? global.simDisk :
                     (long)(global.simDisk * ((double)random() / RAND_MAX));
      if(time(NULL) >= info->ft->meta.timeStarted + runTime) {
        info->terminated = 1;
        info->exitValue = 0;
      }
    }
    else if(!info->terminated)
      info->terminated = ELBASE_Poll(info->pid, 0, &info->exitValue);
    if(!info->terminated)
      continue;
    c = StderrCaptureRetrieve(info->stderrKey);
    if(info->canceled)
      ; /* empty */
    else {
      FileTransferDone(info->ft, info->exitValue);
      if(info->files != NULL) { /* Bulk transfer */
        if(info->exitValue == 0)
          ForEachMapping(info->files, it)
            NewFileCopy(info->ft->dest, (File)IteratorValue(it));
        FileList_Free(info->files, 0);
      }
      if(info->exitValue != 0 && c != NULL)
        WARN1("Transfer stderr follows\n%s\n", c);
    }
    if(c != NULL)
      FREE(c);
    pendingTransfers[i] = pendingTransfers[Size(pendingTransfers) - 1];
    ShrinkSize(pendingTransfers, Size(pendingTransfers) - 1);
    FREE(info);
  }
  return 1;
}


int
storage_process_xml(ASXML_XmlNode node) {
  ASXML_XmlNode child;
  Disk d;
  File f;
  int result = 1;
  for(child = ASXML_NodeFirstChildByType(node, ASXML_SHOW_ELEMENT);
      child != NULL;
      child = ASXML_NodeNextSiblingByType(child, ASXML_SHOW_ELEMENT)) {
    if(strcmp(ASXML_NodeName(child), "disk") == 0) {
      if((d = ProcessDiskXml(child)) == NULL)
        result = 0;
      else
        DiskList_Add(global.disks, d);
    }
    else if(strcmp(ASXML_NodeName(child), "file") == 0) {
      if((f = ProcessFileXml(child)) == NULL)
        result = 0;
      else
        FileList_Add(global.files, f);
    }
    else {
      ERROR1("Unknown storage subtag '%s'\n", ASXML_NodeName(child));
      result = 0;
    }
  }
  return result;
}


char *
storage_disk_to_xml(Disk d) {
  char *result;
  char *server;
  server = *d->server == '\0' || strcmp(d->server, d->id) == 0 ? "" :
           StrAppend(NULL, " server='", d->server, "'", NULL);
  result = StrAppend
    (NULL, "<disk id='", d->id, "' datadir='", d->datadir, "'><",
     METHOD_IMAGES[d->method], server, "/></disk>", NULL);
  if(*server != '\0')
    FREE(server);
  return result;
}


char *
storage_file_to_xml(File f) {

  char *copies;
  const char *download;
  char *result;
  char *size;
  char sizeImage[10];
  const char *transfer;

  if(f->download && f->size == 0 && GetMappingCount(f->copies) == 0)
    return STRDUP(""); /* No useful information */
  download = f->download ? "" : " download='no'";
  if(f->size == 0)
    size = "";
  else {
    sprintf(sizeImage, "%.0f", f->size);
    size = StrAppend(NULL, " size='", sizeImage, "'", NULL);
  }
  transfer = f->transferable ? "" : " transfer='no'";
  if(GetMappingCount(f->copies) == 0)
    result = StrAppend
      (NULL, "<file id='", f->name, "'", download, transfer, size, "/>", NULL);
  else {
    copies = HashToXML(f->copies, NULL, (XmlFn)&CopyToXML);
    result = StrAppend
      (NULL, "<file id='", f->name, "'", download, transfer, size, ">", copies,
       "</file>", NULL);
    FREE(copies);
  }
  if(*size != '\0')
    FREE(size);
  return result;

}


int
storage_shutdown(void) {
  char *c;
  unsigned i;
  for(i = 0; i < Size(pendingTransfers); i++) {
    if((c = StderrCaptureRetrieve(pendingTransfers[i]->stderrKey)) != NULL)
      FREE(c);
    FREE(pendingTransfers[i]);
  }
  VectorFree(pendingTransfers);
  return 1;
}
